[inherit('vglobal')]
module DPMIfile(output);

function bolTointeger(inBol : boolean):integer;
begin
  if inBol then bolTointeger:=1
  else if (not inBol) then bolTointeger:=0;
end;

function integerToBol(ininteger : integer):boolean;
begin
  if ininteger=1 then integerToBol:=true
  else if ininteger=0 then integerToBol:=false;
end;

function bolToReal(inBol : boolean):double;
begin
  if inBol then bolToReal:=1.0
  else if (not inBol) then bolToReal:=0.0;
end;

function realToBol(inWord : double):boolean;
begin
  if round(inWord)=1 then realToBol:=true
  else if round(inWord)=0 then realToBol:=false;
end;

  (*////////////////////////////////////////////////////////////////////////
     freeDvMemory dispose of all dvRec from memory, link^, cicLink^, and
     zMoment^ are also disposed of accordingly.
     inStart, inEnd are returned nil
   ///////////////////////////////////////////////////////////////////////*)
[global]
procedure freeDvMemory (var inStart, inEnd: dvPtr);
var
  tempDv,curPtr: dvPtr;
  j:integer;
begin
  curPtr:=inStart;
  while curPtr <> nil do with curPtr^ do begin
    tempDv:=next;
    if link<>nil then dispose(link);
    for j:=0 to numSubLevel do
     if zMoment[j]<>nil then dispose(zMoment[j]);
    dispose(curPtr);
    curPtr:=tempDv;
  end;
  inStart:=nil;
  inEnd:=nil;
end; {/// FreeDvMemory ///}

[global]
procedure getGlobalRhsMemory;
{called for only once from manager.p}
var
  matxSize,bdy,i : integer;
  tmpVector : PVector;
begin
  matxSize:=0;
  for i:=1 to numActiveBbody do matxSize:=matxSize+bodyAry[i]^.mSeg;

  for i:=1 to matxSize do begin
     {/// create pivot pointer ///}
     New(tmpVector);
     tmpVector^.RealNum:=0.0;
     tmpVector^.nextNum:=nil;
     if GlobalRhs = nil then
       GlobalRhs:= tmpVector
     else begin
       tmpVector^.nextNum:= globalRhs;
       globalRhs:= tmpVector;
     end;
   end;
end;

[global]
procedure getVectorMemory(var theVector: PVector; theSize:integer);
{no callers}
var
  tmpVector      : PVector;
  i: integer;
begin
  if theVector<>nil then freeVectorMemory(theVector);
  for i:=1 to theSize do begin
     {/// create pivot pointer ///}
     New(tmpVector);
     tmpVector^.RealNum:=0.0;
     tmpVector^.nextNum:=nil;
     if theVector = nil then
       theVector:= tmpVector
     else begin
       tmpVector^.nextNum:= theVector;
       theVector:= tmpVector;
     end;
   end;
end;

[global]
procedure freeVectorMemory(var theVector:PVector);
var
  i: integer;
  tmpVector : PVector;
begin
  tmpVector:=theVector;
  while tmpVector<>nil do begin
    theVector:=tmpVector^.nextNum;
    dispose(tmpVector);
    tmpVector:=theVector;
  end;
end;

[global]
procedure freeForceListMemory;
var
  curFceList : PFcList;
  tmpEntry,curEntry   : PfcEntry;
begin
  curFceList:=fcListStart;
  while curFceList<>nil do with curFceList^ do begin
     fcListStart:=curFceList^.dnFcList;
     curEntry:=fcEntry;
     while curEntry<>nil do begin
       tmpEntry:=curEntry^.nextFce;
       dispose(curEntry);
       curEntry:=tmpEntry;
     end;
     dispose(curFceList);
     curFceList:=fcListStart;
   end;
  fcListEnd:=nil;
end;

[global]
procedure freeProbeMemory;
var
  curList : PProbeList;
  tmpEntry,curEntry : PProbeData;
begin
  curList:=probeStart;
  while curList<>nil do with curList^ do begin
   probeStart:=curList^.dnPbList;
   curEntry:=probeData;
   while curEntry<>nil do begin
    tmpEntry:=curEntry^.nextSet;
    dispose(curEntry);
    curEntry:=tmpEntry;
   end;
   dispose(curList);
   curList:=probeStart;
   end;
  probeEnd:=nil;
end;


[global]
procedure freeMatrixMemory(var theMatrix:PCoeffRow);
var
  curCoefRow : PcoeffRow;
  curEntry: PVector;
begin
  curCoefRow:=theMatrix;
  while curCoefRow<>nil do with curCoefRow^ do begin
     theMatrix:=curCoefRow^.downRow;
     freeVectorMemory(RowStart);
     dispose(curCoefRow);
     curCoefRow:=theMatrix;
   end;
end;

[global]
procedure freeBodyAryMemory;
var
  j,i : integer;
begin
  for j:=1 to maxBody do
   if BodyAry[j]<>nil then with bodyAry[j]^ do begin
     for i:=1 to maxSeg do
      if seg[i]<>nil then begin
         dispose(seg[i]);
         seg[i]:=nil;
      end;
     if selfMatx<>nil then FreeMatrixMemory(selfMatx);
     for i:=1 to maxMutual do
      if mutlMatx[i]<>nil then FreeMatrixMemory(mutlMatx[i]);

     dispose(BodyAry[j]);
     BodyAry[j]:=nil;
    end; {with}
  {/// free global coeff matrix memory ///}
  if dynaMatx<>nil then freeMatrixMemory(dynaMatx);
  {/// free global right-hand-side memory ///}
  if globalRhs<>nil then freeVectorMemory(globalRhs);
  {/// free global list of lift/drag force memory ///}
  if fcListStart<>nil then freeForceListMemory;
  {/// free global list of probe sampling data ///}
  if probeStart<>nil then freeProbeMemory;
end;

[global]
procedure getGeoMemory(theBody:integer);
var
  i : integer;
begin
  for i := 1 to maxSeg do begin
    New(BodyAry[theBody]^.seg[i]);
  end;
end;

[global]
procedure freeChainOfPoint(var chainStart: PlinkPt);
var tmpPtr: PlinkPt;
begin
  while chainStart<>nil do begin
   tmpPtr:=chainStart^.downLink;
   dispose(chainStart);
   chainStart:=tmpPtr;
  end;
end;

[global]
procedure addPointToChain(var chainStart:PlinkPt;newPoint:zNumber);
var
  tmpLinkPt : PlinkPt;
begin
  New(tmpLinkPt);
  tmpLinkPt^.downLink:=nil;
  tmpLinkPt^.ptCoord.Re:=newPoint.Re;
  tmpLinkPt^.ptCoord.Im:=newPoint.Im;
  if ChainStart=nil then chainStart:=tmpLinkPt
  else begin
   tmpLinkPt^.downLink:=ChainStart;
   ChainStart:=tmpLinkPt;
  end;
end;

(*//////////////////////////////////////////////////////////////////////////
   dvPack 'dispose of ' those snuffed dv and render allocated memory to
   heap.  Dispose of also the pointers in the dvRec : link^, ciclink^ and
   zMoment^
///////////////////////////////////////////////////////////////////////////*)

procedure TrashBin(var aJunk: dvPtr);
var j,k : integer;
begin
  {if reUseMem then begin}
    for j:=0 to numSubLevel do begin
       if aJunk^.zMoment[j]<>nil then begin
         aJunk^.zMoment[j]^.ctcPter:=nil;
         for k:=1 to ctcOrderLimit do begin
           aJunk^.zMoment[j]^.zzj[k].Re:=0.0;
           aJunk^.zMoment[j]^.zzj[k].Im:=0.0;
         end;
       end;
    end;
    if Trash<>nil then begin
      aJunk^.next:=Trash;
      Trash:=aJunk;
    end else begin
      Trash:=aJunk;
      Trash^.next:=nil;
    end;
    aJunk:=nil;
    nTrash:=nTrash+1.0;
  {end else begin
    if aJunk^.link<>nil then dispose(aJunk^.link);
    for j:=0 to numSubLevel do
      if aJunk^.zMoment[j]<>nil then dispose(aJunk^.zMoment[j]);
    dispose(aJunk);
    aJunk:=nil;
  end;}
end;

procedure getNewdvPtr(var reNew: dvPtr);
var j:integer;
begin
  if (nTrash>1) and (Trash<>nil) then begin
   reNew:=Trash;
   Trash:=Trash^.next;
   nTrash:=nTrash-1.0;
  end else begin
   New(reNew);
   New(reNew^.link);
   for j:=0 to maxCellevel do reNew^.zMoment[j]:=nil;
  end;
end;

procedure clearTrash;
var tmp : dvPtr;
    j   : integer;
begin
  while Trash<>nil do begin
   tmp:=Trash;
   Trash:=Trash^.next;
   if tmp^.link<>nil then dispose(tmp^.link);
    for j:=0 to numSubLevel do
      if tmp^.zMoment[j]<>nil then dispose(tmp^.zMoment[j]);
    dispose(tmp);
    nTrash:=nTrash-1.0;
  end;
end;


[global]
procedure DvPack (Var ListStart, ListEnd : dvPtr; Var ListEmpty :boolean);
label 1;
var
  nextPtr,TempPtr,curPtr   : dvPtr;
  j: integer;
begin
  curPtr:= ListStart;
  while (curPtr<>nil) do begin
    nextPtr:=curPtr^.next;
    if curPtr^.Vortex.mag=0.0 then begin
      if (curPtr=ListStart) then begin
	TempPtr := ListStart;
        ListStart := ListStart^.Next;
        if ListStart<>nil then ListStart^.Last := nil
        else begin
          if listEnd<>nil then ListEnd := nil;
          TrashBin(tempPtr);
          {if tempPtr^.link<>nil then dispose(tempPtr^.link);
          for j:=0 to maxCellevel do
           if tempPtr^.zMoment[j]<>nil then dispose(tempPtr^.zMoment[j]);
          dispose(tempPtr);}
          ndv:=0;
          listEmpty:=true;
          goto 1;
        end;
      end
        else if curPtr=ListEnd then begin
	   {write(' ZZZZ ');}
	   TempPtr := ListEnd;
          ListEnd := ListEnd^.Last;
          if ListEnd<>nil then ListEnd^.Next := nil;
        end
        else begin
	     {write(' DEF ');}
	     tempPtr:=curPtr;
            if curPtr^.next<>nil then curPtr^.Next^.Last := curPtr^.Last;
            if curPtr^.last<>nil then curPtr^.Last^.Next := curPtr^.Next;
            if curPtr^.last<>nil then curPtr:=curPtr^.Last;     {/// move back pointer one record ///}
          end;

         {if (tempPtr <> nil) then begin
	     if tempPtr^.link<>nil then dispose(tempPtr^.link);
             for j:=0 to maxCellevel do
             if tempPtr^.zMoment[j]<>nil then dispose(tempPtr^.zMoment[j]);
             dispose(tempPtr);
          end;}
         TrashBin(tempPtr);

         ndv:=ndv-1;
       end;{if curPtr^.mag=dumpMag}
       curPtr := nextPtr;
     end;{while}
1:end; {/// EOProc. dvPack ///}


{///////////////////////////////////////////////////////////////////////
    Procedure dvAppend :

    This routine appends a record into the double-link pointer list which
    starts with ListStart and ends with ListEnd. The standard pointer
    allocation function GetMem is being used to make the code machine
    independent.

    Notes :

    Both ListStart and ListEnd must be initialised to nil before calling
    this routine for the first time.

    Pointers link^, cicLink^, and zMoment^ are initialised to nil
   ///////////////////////////////////////////////////////////////////////}

[global]
procedure dvAppend (Var ListStart, ListEnd : dvPtr;
                        NewRec : TVortex; var listLength: double);
var
  SearchPtr,TempPtr: dvPtr;
  j: integer;
begin
  {/// Create a new record and place the information in it \\\}
  {if (not Recycle(tempPtr)) then begin
   New(TempPtr);
   New(TempPtr^.link);
   for j:=0 to maxCellevel do TempPtr^.zMoment[j]:=nil;
  end;}
  getNewdvPtr(tempPtr);

  TempPtr^.Vortex:=NewRec;
  TempPtr^.Vortex.vel.Re:=0.0;
  TempPtr^.Vortex.vel.Im:=0.0;
  TempPtr^.Next:=nil;
  TempPtr^.Last:=nil;

  if (NewRec.nearBdy<>0) and (NewRec.nearSeg<>0) then
     tempPtr^.Link^.posi:=bodyAry[newRec.NearBdy]^.seg[NewRec.NearSeg]^.pvt
  else begin
     tempPtr^.Link^.posi.Re:=0.0;
     tempPtr^.Link^.posi.Im:=0.0;
  end;

  tempPtr^.Link^.veli.Re:=0.0;
  tempPtr^.Link^.veli.Im:=0.0;

  listLength:=listLength+1;
  SearchPtr := ListStart;
  {/// if the list is empty, make the new record the only one \\\}
  if SearchPtr = nil then begin
    ListStart := TempPtr;
    ListEnd := TempPtr;
    allGone:=false;
   end
  else begin
    TempPtr^.Last := ListEnd;
    ListEnd^.Next := TempPtr;
    ListEnd := TempPtr;
   end;
end; {/// Proc. dvAppend ///}


(*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
   SaveResult saves vortices data together with geometry, solution control
   data in a single file of REAL number. Integer/word numbers are therefore
   first change into Real number and then are stored up.

   The disk file in REAL format can be scanned through with PRNfile command
   at DOS.
 \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*)

[global]
procedure SaveResult(fileName: TString);
var
  curPtr   : dvPtr;
  destFile : file of double;
  j        : integer;
  readRec  : TVortex;
  ndvR,bfBdyR,offsetR,decayR,meshFactorR : double;
  codeR,NxR,NyR,NnoseR : double;
  Rmag,RtGen,Rx,Ry,Ru,Rv : double;
begin
  rewrite(destFile,fileName);
    {/// save bluff body info for later computation ///}
    codeR:=numActiveBbody*1.0;
    write(DestFile, codeR);
    for j:=1 to numActiveBbody do
     with ActiveBbody[j] do begin
       codeR:=shapeCode*1.0;
       NxR:=Nx*1.0;
       NyR:=Ny*1.0;
       NnoseR:=Nnose*1.0;

       write(DestFile, CodeR, BodyCX, BodyCY, BodyTilt);
       write(DestFile, xDim, yDim, zDim);
       write(DestFile, NxR, NyR, NnoseR);
     end;
    ndvR:=Ndv*1.0;
    bfBdyR:=0.0;{boltoReal(bfBdyActive);}
    offsetR:=0.0;{boltoReal(offsetActive);}
    decayR:=decayFactor;
    {leakR:=boltoReal(nuLeakCor);}
    meshFactorR:=basicMeshFactor;
    write(DestFile, ReyNum,alpha);
    write(DestFile, elapsed,dt,NdvR);
    write(DestFile, snuff);
    write(DestFile, bfBdyR, offsetR, decayR, meshFactorR);

    curPtr:=dvStart;
    while curPtr<>nil do with curPtr^.vortex do begin
        Rmag:=mag;
        RtGen:=tGen;
        Rx:=pos.Re;
        Ry:=pos.Im;
        write(DestFile,Rmag,RtGen,Rx,Ry);
        curPtr:=curPtr^.next;
    end;
    close(destfile);
end;

{\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\}
{  ReadDvRec procedure read data from disk and store with globalDvPlotPtr.  }
{  In this condition, data is ready to be plot out on screen with           }
{  Proc PlotVortexCloud which nulls pointer afterwards.                     }
{  Data have to be transferred to dyn. memory dvStart\dvEnd if they are     }
{  to be processed in Proc RestartSimulation                                }
{\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\}
[global]
procedure ReadDvforRestart (fileName: TString; var inStart, inEnd: dvPtr;
                    var itemRead:double;var inTime: timeRec);
var
  destFile : file of double;
  readRec : TVortex;
  ndvR,bfBdyR,offsetR,decayR,
  meshFactorR,numBodyR,codeR,
  NxR,NyR,NnoseR,Rmag,RtGen,
  Rx,Ry,Ru,Rv:double;
  j:integer;
begin
  itemRead:=0;

  {/// erase all existing data before retriving from disk ///}
  if inStart<>nil then freeDvMemory(inStart,inEnd);

  reset(destFile,fileName);
    {/// read data from disk ///}
    read(DestFile, numBodyR);
    numActiveBbody:=trunc(numBodyR);
    for j:=1 to numActiveBbody do with ActiveBbody[j] do begin
      read(DestFile, CodeR, BodyCX, BodyCY, BodyTilt);
      read(DestFile, xDim, yDim, zDim);
      read(DestFile, NxR, NyR, NnoseR);

      shapeCode:=Trunc(codeR);
      Nx:=trunc(NxR);
      Ny:=trunc(NyR);
      Nnose:=trunc(NnoseR);
    end;
    read(DestFile, ReyNum,alpha);
    read(DestFile, inTime.tLast,inTime.tStep,NdvR);
    read(DestFile, snuff);
    read(DestFile, bfBdyR, offsetR, decayR, MeshFactorR);

    if (decayR>-0.5) and (decayR<0.0) then
       decayFactor:=decayR;
    if MeshFactorR<>0 then basicMeshFactor:=MeshFactorR;

    while not Eof(destFile) do with readRec do begin
      {read(destFile,mag,tGen,x,y,u,v);}
      read(destFile,Rmag,RtGen,Rx,Ry);
      mag:=Rmag;
      tGen:=RtGen;
      pos.Re:=Rx;
      pos.Im:=Ry;
      vel.Re:=0.0;
      vel.Im:=0.0;
      nearBdy:=0;
      nearSeg:=0;
      dvAppend(inStart,inEnd,readRec,itemRead);
    end;
    close(destFile);
    {bfBdyActive:=realToBol(bfBdyR);}
    {offsetActive:=realToBol(offsetR);}
    {snuffActive:=false;}
    {nuLeakCor:=realToBol(LeakR);}
end; {/// Proc ReadDvForReStart ///}

  {\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
    (222)   Force record group
   \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\}

[global]
procedure FceListAppend (Var gbFceStart, gbFceEnd: PFcList; ListBegin: PFcEntry;
                        inTime: double);
var
  TempPtr    : PFcList;
begin
  {/// Create a new record and place the information in it \\\}
  New(TempPtr);
  tempPtr^.timeDue:=inTime;
  TempPtr^.fcEntry:=ListBegin;
  TempPtr^.upfcList:=nil;
  TempPtr^.dnfcList:=nil;
  fceListLength:=fceListLength+1;
  {/// if the list is empty, make the new record the only one \\\}
  if gbFceStart = nil then
   begin
    gbFceStart:= tempPtr;
    gbFceEnd:=tempPtr;
   end
  else
   begin
    tempPtr^.upFcList:= gbFceEnd;
    gbFceEnd^.dnFcList:=tempPtr;
    gbFceEnd:= tempPtr;
   end;
end; {/// Proc. fceListAppend ///}

[global]
procedure FcEntryAppend (Var oneListStart: PfcEntry; newDrag,newLift: double);
var
  SearchPtr, TempPtr    : PfcEntry;
begin
  {/// Create a new record and place the information in it \\\}
  New(TempPtr);
  TempPtr^.CDrag:=newDrag;
  TempPtr^.CLift:=newLift;
  TempPtr^.nextFce:=nil;

  {/// if the list is empty, make the new record the only one \\\}
  if oneListStart = nil then
    oneListStart := TempPtr
  else
   begin
    {/// locate end of list ///}
    searchPtr:=oneListStart;
     while searchPtr^.nextFce<>nil do searchPtr:=searchPtr^.nextFce;
    searchPtr^.nextFce:=tempPtr;
   end;
end; {/// Proc. FcEntryAppend ///}

[global]
procedure AppendForceList;
var
  curFceList : PFcList;
  curEntry   : PfcEntry;
  fileSave   : file of double;
  dumpR,Rmin,Rmax,Ravg  : double;
  bdy,i      : integer;
begin
  rewrite(fileSave,writeForceName);
  {/// generate new file of the same file name ///}
  dumpR:=numActiveBbody*1.0;
  write(fileSave,dumpR);
  for bdy:=1 to numActiveBbody do with BodyAry[bdy]^.PonBody do begin
   dumpR:=BodyAry[bdy]^.mSeg*1.0;
   write(fileSave,dumpR);
   dumpR:=numRecord*1.0;
   write(fileSave,dumpR);
   for i:=1 to BodyAry[bdy]^.mSeg do begin
    Rmin:=minPress[i];
    Rmax:=maxPress[i];
    Ravg:=AvgPress[i];
    write(fileSave,Rmin,Rmax,Ravg);
   end;
  end;
  curFceList:=fcListStart;
  while curFceList<>nil do with curFceList^ do begin
   write(fileSave, timeDue);
   curEntry:=fcEntry;
   while curEntry<>nil do begin
    write(fileSave, curEntry^.CDrag, curEntry^.CLift);
    curEntry:=curEntry^.nextFce;
   end;
   curFceList:=curFceList^.dnFcList;
  end;
  close(fileSave);
end;

[global]
procedure saveCP(filename:Tstring);
var
  fileSave             : file of double;
  dumpR,Rmin,Rmax,Ravg : double;
  bdy,i                : integer;
begin
  rewrite(fileSave,filename);
  dumpR:=numActiveBbody*1.0;
  write(fileSave,dumpR);    { save number of active bodies }
  dumpR:=elapsed*1.0;
  write(fileSave,dumpR);    { save time elapsed }
  dumpR:=BodyAry[1]^.PonBody.numRecord*1.0;
  write(fileSave,dumpR);  { save number of records averaged }

  for bdy:=1 to numActiveBbody do with BodyAry[bdy]^.PonBody do begin
   dumpR:=BodyAry[bdy]^.mSeg*1.0;
   write(fileSave,dumpR);  { save segment on body }
   for i:=1 to BodyAry[bdy]^.mSeg do begin
    Rmin:=minPress[i];
    Rmax:=maxPress[i];
    Ravg:=AvgPress[i];
    write(fileSave,Rmin,Rmax,Ravg);
   end;
  end;
  close(fileSave);
end;

[global]
procedure ReadForceList;
var
  thisListStart : PfcEntry;
  fileRead      : file of double;
  timme,tempdrag,templift : double;
  dumpR,Rmin,Rmax,Ravg    : double;
  i,bdy,numBody           : integer;
begin
  reset(fileRead,readForceName);
  read(fileRead,dumpR);
  numBody:=trunc(dumpR);
  for bdy:=1 to numBody do with BodyAry[bdy]^.PonBody do begin
   read(fileRead,dumpR); { read mSeg, dummy }
   read(fileRead,dumpR);
   numRecord:=trunc(dumpR);
   for i:=1 to BodyAry[bdy]^.mSeg do begin
    Read(fileRead,Rmin,Rmax,Ravg);
    minPress[i]:=Rmin;
    maxPress[i]:=Rmax;
    AvgPress[i]:=Ravg;
   end;
  end;
  while not Eof(fileRead) do begin
   read(fileRead, timme);
   thisListStart:=nil;
   for bdy:=1 to numBody do begin
    read(fileRead,tempDrag,tempLift);
    fcEntryAppend(thisListStart, tempDrag, tempLift);
   end;
   fceListAppend(fcListStart,fcListEnd,thisListStart,timme);
  end;
  close(fileRead);
end;

[global]
procedure probeListAppend (var theStart, theEnd: PProbeList; firstSet: PProbeData;
                          inTime: double);
var
  TempPtr    : PProbeList;
begin
  {/// Create a new record and place the information in it \\\}
  New(TempPtr);
  tempPtr^.sampleTime:=inTime;
  TempPtr^.probeData:=firstSet;
  TempPtr^.upPbList:=nil;
  TempPtr^.dnPbList:=nil;

  {probeLength:=probeLength+1;}
  {/// if the list is empty, make the new record the only one \\\}
  if theStart = nil then begin
    theStart:= tempPtr;
    theEnd:=tempPtr;
  end
  else begin
    tempPtr^.upPbList:=theEnd;
    theEnd^.dnPbList:=TempPtr;
    theEnd:= tempPtr;
  end;
end; {proc probeListAppend}

[global]
procedure probeAppend (var firstSet: PProbeData;
                       newTotP,newStatP:double;
                       newVel: zNumber);
var
  SearchPtr, TempPtr    : PProbeData;
begin
  {/// Create a new record and place the information in it \\\}
  New(TempPtr);
  TempPtr^.PTotal:=newTotP;
  TempPtr^.PStatic:=newStatP;
  TempPtr^.LocalVel:=newVel;
  TempPtr^.nextSet:=nil;

  {/// if the list is empty, make the new record the only one \\\}
  if firstSet = nil then
    firstSet:= TempPtr
  else begin
    {/// locate end of list ///}
   searchPtr:=firstSet;
   while searchPtr^.nextSet<>nil do searchPtr:=searchPtr^.nextSet;
   searchPtr^.nextSet:=tempPtr;
  end;
end; {/// Proc. FcEntryAppend ///}

[global]
procedure saveProbeList;
var
  curList    : PProbeList;
  curSet     : PProbeData;
  fileSave   : file of double;
  realA      : double;
  curPoint   : PlinkPt;
begin
  rewrite(fileSave,writeProbeName);
  {/// generate new file of the same file name ///}
  realA:=numProbePoint*1.0;
  write(fileSave,realA);

  { save coordinates of probe locations }
  curPoint:=probePoint;
  while curPoint<>nil do begin
   {realA:=curPoint^.ptCoord.Re*1.0;
   realB:=curPoint^.ptCoord.Im*1.0;}
   {write(fileSave,realA,realB);}
   write(fileSave,curPoint^.ptCoord.Re,curPoint^.ptCoord.Im);
   curPoint:=curPoint^.downLink;
  end;

  curList:=probeStart;
  while curList<>nil do with curList^ do begin
   realA:=sampleTime*1.0;
   write(fileSave,realA);

   curSet:=probeData;
   while curSet<>nil do begin
    write(fileSave, curSet^.PTotal, curSet^.PStatic,
                    curSet^.LocalVel.Re,curSet^.LocalVel.Im);
    curSet:=curSet^.nextSet;
   end;
   curList:=curList^.dnPbList;
  end;
  close(fileSave);
end;

procedure freeVelRec;
var
  curPtr: dvPtr;
begin
  curPtr:=dvStart;
  while curPtr<>nil do begin
   if curPtr^.link<>nil then dispose(curPtr^.Link);
   curPtr:=curPtr^.next;
  end;
end;

[global]
procedure createVelRec;
var
  curPtr: dvPtr;
  tmpVelRec: TvelRec;
begin
  tmpVelRec.veli.Re:=0.0;
  tmpVelRec.veli.Im:=0.0;
  tmpVelRec.posi.Re:=0.0;
  tmpVelRec.posi.Im:=0.0;

  curPtr:=dvStart;
  while curPtr<>nil do
   begin
    if curPtr^.link = nil then New(curPtr^.Link);
    tmpVelRec.posi.Re:=curPtr^.Vortex.pos.Re;
    tmpVelRec.posi.Im:=curPtr^.Vortex.pos.Im;
    curPtr^.link^:=tmpVelRec;
    curPtr:=curPtr^.next;
   end;
end;

procedure freePtrList;
begin
  if xOrder<>nil then dispose(xOrder);
  xOrder:=nil;
  if yOrder<>nil then dispose(yOrder);
  yOrder:=nil;
end;

[global]
procedure getXYptrList(ThisDvStart : dvPtr; var dvOnList: integer);
var
  index : integer;
  curPtr   : dvPtr;
begin
  { add memory checking here !!!!!!!!!! }

  if xOrder=nil then New(xOrder);
  if yOrder=nil then New(yOrder);

  index:=0;
  curPtr:=ThisDvStart;
  while curPtr<>nil do        {/// form the temporary lists ///}
   begin
    if index < maxDV then
     begin
      index:=index+1;
      xOrder^.theList[index]:=curPtr;
      yOrder^.theList[index]:=curPtr;
     end;
    curPtr:=curPtr^.next;
   end;
  dvOnList:=index;
end;

[global]
procedure getCellXYlist(chainStart : PChain; var NumDVinCell: integer);
var
  index:integer;
  curChain : PChain;
begin
  { add memory checking here !!!!!!!!!! }
  if xOrder=nil then New(xOrder);
  if yOrder=nil then New(yOrder);

  index:=0;
  curChain:=chainStart;
  while curChain<>nil do with curChain^ do begin
    if index < maxDV then begin
      index:=index+1;
      xOrder^.theList[index]:=dvElem;
      yOrder^.theList[index]:=dvElem;
     end;
    curChain:=curChain^.nextElem;
   end;
  NumDVinCell:=index;
end;

[global]
procedure freeMemoryAtEnd;
begin
  if dvStart<>nil then freeDvMemory(dvStart,dvEnd);
  if probePoint<>nil then freeChainOfPoint(probePoint);
  freeBodyAryMemory;
  QuitGridTTT;
  freeVelRec;
  freePtrList;
  clearTrash;
end;

end.
