[inherit('vglobal')]
program easyRun(input,output,logbook);

const
  validChar=['0'..'9','a'..'z','A'..'Z','.','/','\',':','+','-'];
  limiter=[',','(',')',';','='];

var
  logbook : text;

procedure ReadLogBook;
var
  curReal : double;
  hitEoln : boolean;
  Heading,CodeName,curString : Tstring;
  curInteger,IntError,RealError : integer;

procedure ReadHeading;
var
  OneChar: Char;
  TmpStg : Tstring;
begin
  TmpStg:='';
  repeat
   read(logbook,OneChar);
   if OneChar in ValidChar then
   TmpStg:=TmpStg+UpperCase(OneChar);
  until OneChar in Limiter;

  Heading:=TmpStg;
end;

Procedure readItem;
var
  OneChar: Char;
  tmpStg: Tstring;
  AReal:double;
begin
  TmpStg:='';
  repeat
  read(logbook,OneChar);
  if OneChar in validChar then
  TmpStg:=TmpStg+UpperCase(OneChar);
  until OneChar in Limiter;
  hitEoln:=OneChar=')';
  CodeName:=TmpStg;

  CurString:='';
  repeat
  read(logbook,OneChar);
  if OneChar in ValidChar then
   CurString:=CurString+UpperCase(OneChar);
  until (OneChar in Limiter);
  hitEoln:=OneChar=')';

  if CurString<>'' then Value(curString,curReal,RealError);
end;

procedure ReadSolution;
var dmp:integer;
begin
  Repeat
  ReadItem;
  if CodeName='RESTART' then
   begin
    if CurString='' then
     begin
      calculateMode:=calViscousFlow;
      vortexFlowMode:=calFreshRun;
     end
    else
     begin
      calculateMode:=calViscousFlow;
      vortexFlowMode:=calRestart;
      PathToReStartFile:=curString;
     end;
   end
  else
  if CodeName='FRESH' then
   begin
    if CurString='' then
     begin
      calculateMode:=calViscousFlow;
      vortexFlowMode:=calFreshRun;
      fileGrp:='NONAM';
      writeForceName:=fileGrp+'.FCE';
      readForceName:=writeForceName;
     end
    else
     begin
      calculateMode:=calViscousFlow;
      vortexFlowMode:=calFreshRun;
      splitString(CurString,fileGrp,dmp);
      writeForceName:=fileGrp+'.FCE';
      readForceName:=writeForceName;
     end;
   end
  else
  if CodeName='POTENT' then  calculateMode:=calPotential
  else
  if CodeName='REYN' then
   begin
    if RealError=0 then ReyNum:=abs(curReal);
   end
  else
  if CodeName='ANG' then
   begin
    if RealError=0 then alpha:=curReal;
   end
  else
  if CodeName='NSTEP' then
   begin
    if RealError=0 then NumTimeStep:=Trunc(abs(curReal));
   end
  else
  if CodeName='SSIZE' then
   begin
    if RealError=0 then dt:=abs(curReal);
   end
  else
  if CodeName='SSAVE' then
   begin
    if RealError=0 then SaveFileInt:=Trunc(abs(curReal));
   end;
  until hitEoln;
end;

procedure ReadBodyData;
{
    0 : Circle;
    1 : Ellipse;
    2 : Rect;
    3 : EPlate;
    4 : Wedge;
    5 : NACA4D input from Tvision window, not being used in logBook.txt
    6 : UserDefineShape(inBbody.UserInputFile);
    9 : NACA4D, XD chord,
                YD 4-digit code (add 9 before code i.e. 94412=NACA 4412

}
begin
  numActiveBbody:=numActiveBbody+1;
  with activeBbody[numActiveBbody] do
  begin
   Repeat
    ReadItem;

    if CodeName='CODE' then
     begin
      if RealError=0 then ShapeCode:=Trunc(curReal);
     end
    else
    if CodeName='CX' then
     begin
      if RealError=0 then BodyCX:=curReal;
     end
    else
    if CodeName='CY' then
     begin
      if RealError=0 then BodyCY:=curReal;
     end
    else
    if CodeName='CA' then
     begin
      if RealError=0 then BodyTilt:=curReal;
     end
    else
    if CodeName='XD' then
     begin
      if RealError=0 then xDim:=abs(curReal);
     end
    else
    if CodeName='YD' then
     begin
      if RealError=0 then yDim:=abs(curReal);
     end
    else
    if CodeName='ZD' then
     begin
      if RealError=0 then zDim:=abs(curReal);
     end
    else
    if CodeName='NX' then
     begin
      if RealError=0 then Nx:=trunc(abs(curReal));
     end
    else
    if CodeName='NY' then
     begin
      if RealError=0 then Ny:=trunc(abs(curReal));
     end
    else
    if CodeName='NZ' then
     begin
      if RealError=0 then Nnose:=trunc(abs(curReal));
     end
    else
    if CodeName='CIRC' then
    begin
     if RealError=0 then BoundCirc:=curReal;
    end;

   until hitEoln;
  end;
end;

procedure ReadOptions;
var
  dateOk,timeOk:boolean;
  dateString,timeString:Tstring;
  tmpYear,tmpMonth,tmpDay,tmpHour,tmpMin:integer;
begin
  dateOk:=false;
  timeOk:=false;
  Repeat
  ReadItem;
  if CodeName='DATE' then begin
    if CurString='' then checkTimeOn:=false
    else begin
     dateOk:=validDate(curString,tmpDay,tmpMonth,tmpYear);
     if dateOk then dateString:=curString else dateString:='';
    end;
   end
  else
  if CodeName='TIME' then begin
    if CurString='' then checkTimeOn:=false
    else begin
     timeOk:=validTime(curString,tmpHour,tmpMin);
     if timeOk then timeString:=curString else timeString:='';
    end;
   end
  else
  if CodeName='SCTC' then begin
   if RealError=0 then numSubLevel:=trunc(CurReal);
   end
  else
  {if CodeName='BUFFER' then
   begin
   if RealError=0 then bfBdyActive:=trunc(CurReal)=1;
   end
  else}
  if CodeName='REUSEMEM' then begin
   if RealError=0 then reUseMem:=trunc(CurReal)=1;
   end
  else
  if CodeName='TRACER' then begin
   if RealError=0 then tracerActive:=trunc(CurReal)=1;
   end
  else
  if CodeName='PRINT' then begin
   if RealError=0 then publish:=trunc(CurReal)=1;
   end
  else
   if CodeName='RSEED' then begin
    if RealError=0 then begin
      KRandSeed:=-trunc(abs(CurReal));
      K2BRandSeed:=-trunc(abs(CurReal));
      PortRandSeed:=-trunc(abs(CurReal));
     end;
   end
  else
  if CodeName='DOMNX' then begin
    if RealError=0 then DomainX:=abs(CurReal);
   end
  else
  if CodeName='DOMNY' then begin
    if RealError=0 then DomainY:=abs(CurReal);
   end
  else
  if CodeName='DECAY' then begin
    if RealError=0 then decayFactor:=-abs(CurReal);
   end
  else
  if CodeName='FLUCT' then begin
    if RealError=0 then fluctLevel:=CurReal;
  end;

  until hitEoln;
  if dateOk and timeOk then begin
    setAlarm(tmpYear,tmpMonth,tmpDay,tmpHour,tmpMin);
    checkTimeOn:=not timesUp;
    AlarmDate:=timeString+' '+dateString;
   end;
end;

procedure ReadStmLineOption;
begin
  Repeat
  ReadItem;
  if CodeName='DVFILE' then begin
    if CurString='' then begin
      calculateMode:=calStreamline;
      VortexStrmLine:=false;
    end
    else begin
      calculateMode:=calStreamline;
      VortexStrmLine:=true;
      PathToDvFile:=CurString;
    end;
  end
  else
  if CodeName='WRITETO' then begin
    if CurString='' then WriteSLName:='STREAM.SLF'
      else WriteSLName:=CurString+'.SLF';
  end
  else
  if CodeName='RBOUND' then begin
    if RealError=0 then strmRLimit:=CurReal;
  end
  else
  if CodeName='STMTIME' then begin
    if RealError=0 then strmTime:=CurReal;
  end;
  until hitEoln;
end;

procedure ReadLine;
var
  xone,xtwo,yone,ytwo : double;
  tmpPt,zprop         : zNumber;
  npnt,i              : integer;
begin
  Repeat
  ReadItem;
  if CodeName='XA' then begin
   if RealError=0 then xone:=CurReal;
  end
  else
  if CodeName='YA' then begin
   if RealError=0 then yone:=CurReal;
  end
  else
  if CodeName='XB' then begin
   if RealError=0 then xtwo:=CurReal
  end
  else
  if CodeName='YB' then begin
   if RealError=0 then ytwo:=CurReal
  end
  else
  if CodeName='NPT' then begin
   if RealError=0 then npnt:=round(CurReal);
  end;

  until hitEoln;

  numStrmLines:=numStrmLines+npnt;
  if npnt>1 then begin
   zProp.Re:=(xtwo-xone)/(npnt-1);
   zProp.Im:=(ytwo-yone)/(npnt-1);
   for i:=1 to npnt do begin
    tmpPt.Re:=Xone+zProp.Re*(i-1);
    tmpPt.Im:=Yone+zProp.Im*(i-1);
    addPointToChain(strmStartPts,tmpPt);
   end; { for }
  end
  else
   if npnt=1 then begin
    tmpPt.Re:=(Xone+Xtwo)*0.5;
    tmpPt.Im:=(Yone+Ytwo)*0.5;
    addPointToChain(strmStartPts,tmpPt);
   end; { for }
end;

procedure ReadAddPoint;
var
  zpt       : zNumber;
begin
  Repeat
  ReadItem;
  if CodeName='XA' then begin
    if RealError=0 then zPt.Re:=CurReal;
   end
  else
  if CodeName='YA' then begin
    if RealError=0 then zPt.Im:=CurReal;
   end;
  until hitEoln;

  numStrmLines:=numStrmLines+1;
  addPointToChain(strmStartPts,zpt);
end;

procedure ReadProbeOption;
begin
  Repeat
  ReadItem;
  if CodeName='FILE' then begin
    if CurString='' then WriteProbeName:='PROBE.PBE'
      else WriteProbeName:=CurString+'.PBE';
  end
  else
  if CodeName='ONTIME' then begin
    if RealError=0 then probeOn:=CurReal;
  end
  else
  if CodeName='OFFTIME' then begin
    if RealError=0 then probeOff:=CurReal;
  end;
  until hitEoln;
end;

procedure ReadProbePoint;
var
  zPt       : zNumber;
begin
  Repeat
  ReadItem;
  if CodeName='X' then begin
    if RealError=0 then zPt.Re:=CurReal;
  end
  else
  if CodeName='Y' then begin
    if RealError=0 then zPt.Im:=CurReal;
  end;
  until hitEoln;
  numProbePoint:=numProbePoint+1;
  addPointToChain(probePoint,zPt);
end;

begin
  reset(logbook);
  numActiveBbody:=0;
  numStrmLines:=0;
  numProbePoint:=0;
  Repeat

   ReadHeading;

   if Heading='SOLN' then ReadSolution
   else
   if Heading='BODY' then ReadBodyData
   else
   if Heading='OPTN' then ReadOptions
   else
   if Heading='STRM' then ReadStmLineOption
   else
   if Heading='LINE' then ReadLine
   else
   if Heading='ADDP' then ReadAddPoint
   else
   if Heading='GDATA' then ReadProbeOption
   else
   if Heading='PROBE' then ReadProbePoint;

  until Heading='END';
end;

begin
  ReadLogBook;
  GoEasyRun;
  freeMemoryAtEnd;
end.
