[inherit('vglobal')]
module GridTTT(output);

var
  McHeight,McWidth        : double;
  specialBox,ZData        : PzCoord;
  CGroup                  : array[0..maxCellevel] of PGridData;
  ctcCoeffA,ctcCoeffB     : ^CtcCoeff;
  ctcSumCoeff             : ^ctcArray;
  recentBase              : integer;

{//////////////////////////////////////////////////////////////////////////
  Procs. PchainBin, getNewPchain and ClearPcTrash deal with recycling of
  PChain dynamic memory
///////////////////////////////////////////////////////////////////////////}

procedure PchainBin(var aJunk: Pchain);
begin
  {if reUseMem then begin}
    if ChainTrash<>nil then begin
      aJunk^.nextElem:=ChainTrash;
      ChainTrash:=aJunk;
    end else begin
      ChainTrash:=aJunk;
      ChainTrash^.nextElem:=nil;
    end;
    nPCTrash:=nPCTrash+1.0;
    aJunk:=nil;
  {end else begin
    dispose(aJunk);
    aJunk:=nil;
  end;}
end;

procedure getNewPChain(var reNew: Pchain);
begin
  if (nPcTrash>1) and (ChainTrash<>nil) then begin
   reNew:=ChainTrash;
   ChainTrash:=ChainTrash^.nextElem;
   nPCTrash:=nPCTrash-1.0;
  end else begin
   new(reNew);
   {newChainMem:=newChainMem+1;}
  end;
end;

procedure clearPCTrash;
var tmp : Pchain;
begin
  while ChainTrash<>nil do begin
   tmp:=ChainTrash;
   ChainTrash:=ChainTrash^.nextElem;
   dispose(tmp);
   nPCTrash:=nPCTrash-1.0;
  end;
end;

{//////////////////////////////////////////////////////////////////////////
  Procs. PCellBin, getNewPCell and ClearCellTrash deal with recycling of
  PCell dynamic memory
///////////////////////////////////////////////////////////////////////////}

procedure PCellBin(var aJunk: PCell);
var j        : integer;
    tmpChain : PChain;
begin
  {if reUseMem then begin}
    if aJunk^.Ctoc<>nil then begin
     aJunk^.Ctoc^.Zm0:=0.0;   { sum zero moment in cell }
     for j:=1 to ctcOrderLimit do begin
      aJunk^.Ctoc^.Zmmt[j].Re:=0.0;
      aJunk^.Ctoc^.Zmmt[j].Im:=0.0;
     end;
    end;

    if PCellTrash<>nil then begin
      aJunk^.UpC:=PCellTrash;
      PCellTrash:=aJunk;
    end else begin
      PCellTrash:=aJunk;
      PCellTrash^.UpC:=nil;
    end;
    nPCellTrash:=nPCellTrash+1.0;
    aJunk:=nil;
  {end else begin
    if aJunk^.Ctoc<>nil then dispose(aJunk^.ctoc);
    dispose(aJunk);
    aJunk:=nil;
  end;}
end;

procedure getNewPCell(var reNew: PCell);
begin
  if (nPCellTrash>1) and (PCellTrash<>nil) then begin
   reNew:=PCellTrash;
   PCellTrash:=PCellTrash^.UpC;
   nPCellTrash:=nPCellTrash-1.0;
  end else begin
   new(reNew);
   reNew^.CtoC:=nil;
   {newCellMem:=newCellMem+1;}
  end;
  with reNew^ do begin
   Ctr.Re:=0.0; Ctr.Im:=0.0;
   Clear:=CellUnknown;
   Zdv:=0;
   dvInCell:=nil;
   NodeStart:=nil;
   NodeEnd:=nil;
   UpC:=nil;DnC:=nil;LfC:=nil;RtC:=nil;
   PData:=nil;
  end;
end;

procedure clearPCellTrash;
var tmp : PCell;
begin
  while PCellTrash<>nil do begin
   tmp:=PCellTrash;
   PCellTrash:=PCellTrash^.UpC;
   dispose(tmp);
   nPCellTrash:=nPCellTrash-1.0;
  end;
end;

{//////////////////////////////////////////////////////////////////////////
  Procs. PzCoordBin, getNewPzCoord and ClearPzCoordTrash deal with recycling
  of PzCoord dynamic memory
///////////////////////////////////////////////////////////////////////////}
procedure PzCoordBin(var aJunk: PzCoord);
var j: integer;
begin
  {if reUseMem then begin}
    for j:=1 to ctcOrderLimit do begin
      aJunk^.IZZab[j].Re:=0.0;
      aJunk^.IZZab[j].Im:=0.0;
    end;
    if PzCoordTrash<>nil then begin
      aJunk^.next:=PzCoordTrash;
      PzCoordTrash:=aJunk;
    end else begin
      PzCoordTrash:=aJunk;
      PzCoordTrash^.next:=nil;
    end;
    nPzCoordTrash:=nPzCoordTrash+1.0;
    aJunk:=nil;
  {end else begin
    dispose(aJunk);
    aJunk:=nil;
  end;}
end;

procedure getNewPzCoord(var reNew: PzCoord);
begin
  if (nPzCoordTrash>1) and (PzCoordTrash<>nil) then begin
   reNew:=PzCoordTrash;
   PzCoordTrash:=PzCoordTrash^.next;
   nPzCoordTrash:=nPzCoordTrash-1.0;
  end else begin
   new(reNew);
   {newzCoordMem:=newzCoordMem+1;}
  end;
end;

procedure clearPzCoordTrash;
var tmp : PzCoord;
begin
  while PzCoordTrash<>nil do begin
   tmp:=PzCoordTrash;
   PzCoordTrash:=PzCoordTrash^.next;
   dispose(tmp);
   nPzCoordTrash:=nPzCoordTrash-1.0;
  end;
end;

[global]
procedure CheckSmpCell(theStart:PSmpCell);
var
  curNode,curRow: PSmpCell;
begin
  curRow:=theStart;
  while curRow<>nil do begin
    curNode:=CurRow;
    while curNode<>nil do with curNode^ do begin
      freePt:=not IsPtInBody(ctr,noCheck);
      curNode:=curNode^.East;
     end;
    curRow:=curRow^.Nth;
  end;
end;

procedure resetSmpCell(whichSmp:PSmpCell;bin:integer);
var
  curNode,curRow: PSmpCell;
begin
  curRow:=whichSmp;
  while curRow<>nil do begin
    curNode:=CurRow;
    while curNode<>nil do begin
      curNode^.sample[bin]:=0.0;
      curNode:=curNode^.East;
    end;
    curRow:=curRow^.Nth;
  end;
end;

procedure freeSmpCell(var inStart, inEnd: PSmpCell);
var
  tempMesh,curMesh,tempRow,curRow : PSmpCell;
begin
  curRow:=inStart;
  while curRow<>nil do begin
   tempRow:=curRow^.East;
   curMesh:=curRow;
   while curMesh <> nil do with curMesh^ do begin
    tempMesh:=curMesh^.Nth;
    {/// free cell pointer ///}
    dispose(curMesh);
    curMesh:=tempMesh;
   end;
   curRow:=tempRow;
  end;
  inStart:=nil;
  inEnd:=nil;
end;

[global]
Procedure formSmpCell(var instart,inend :PsmpCell;
                          LoBound,UpBound :zNumber;
                          gridSize:double);
var
  nCol,nRow,iCol,jRow    : integer;
  tmpMesh                : TSmpCell;
  curNodeA,CurNodeB,
  NodeA,NodeB,
  tempRow,tempPtr   : PSmpCell;
begin
  nRow:=round(abs(UpBound.Im-LoBound.Im)/gridSize);
  nCol:=round(abs(UpBound.Re-LoBound.Re)/gridSize);
  {/// allocate global heap of required size ///}
  if inStart<>nil then freeSmpCell(inStart,inEnd);
  with tmpMesh do begin
   Ctr.Re:=0.0;
   Ctr.Im:=0.0;
   Nth:=nil;
   Sth:=nil;
   East:=nil;
   West:=nil;
   sample[vobin]:=0.0;
   sample[pobin]:=0.0;
   sample[psbin]:=0.0;
   freePt:=true;
  end;

  {/// calculate coordinate of node points ///}
  for iCol:=0 to nCol do begin
    tmpMesh.Ctr.Re:=LoBound.Re+iCol*gridSize;
    for jRow:=0 to nRow do begin
      new(tempPtr);
      tmpMesh.Ctr.Im:=LoBound.Im+jRow*gridSize;
      tempPtr^:=TmpMesh;

     {/// if the list is empty, make the new record the only one \\\}
     if inStart=nil then begin
       inStart:=TempPtr;
       inEnd:=TempPtr;
       tempRow:=TempPtr;
      end
     else begin
       TempPtr^.Sth:=inEnd;
       inEnd^.Nth:=TempPtr;
       inEnd:=TempPtr;
       if jRow=0 then begin
         tempPtr^.west:=tempRow;
         tempRow^.east:=tempPtr;
         tempRow:=tempPtr;
         tempRow^.Sth^.Nth:=nil;
         tempRow^.Sth:=nil;
       end;
      end;
    end;
  end;

  {/// form cross-row pointer link between pointers ///}
  NodeA:=inStart;
  NodeB:=NodeA^.East;
  while NodeB<>nil do begin
    CurNodeA:=NodeA^.Nth;
    CurNodeB:=NodeB^.Nth;
    while curNodeA<>nil do begin
      curNodeA^.East:=curNodeB;
      curNodeB^.West:=curNodeA;
      curNodeA:=curNodeA^.Nth;
      curNodeB:=curNodeB^.Nth;
    end;
    NodeA:=NodeB;
    NodeB:=NodeB^.East;
  end;
end;

[global]
procedure SmearDVonSmp( whichSmp:PSmpCell; celength:double; bin:integer);
var
  curPtr                   : dvPtr;
  xApart,yApart,
  iArea,iWidth,
  xdv,ydv,magdv            : double;
  loop,xincrt,yincrt,i     : integer;
  search                   : PSmpCell;
  outOfDomain,xDone,yDone  : boolean;
  weighUR,weighUL,weighLL,weighLR : double;
begin
  iArea:=1.0/sqr(celength);
  iWidth:=1.0/celength;
  curPtr:=dvStart;
  while curPtr<>nil do begin
    xdv:=curPtr^.vortex.pos.Re;
    ydv:=curPtr^.vortex.pos.Im;
    magdv:=curPtr^.vortex.mag;
    xDone:=false;
    yDone:=false;

    search:=whichSmp;
    {!!! search for nearest lower-left node !!!}
    loop:=0;
    outOfDomain:=false;
    repeat
      xApart:=(xdv-search^.Ctr.Re)/celength;
      yApart:=(ydv-search^.Ctr.Im)/celength;
      xincrt:=trunc(xApart);
      yincrt:=trunc(yApart);
      if xApart<-0.0001 then begin
        for i:=0 downto xincrt do if search^.west<>nil then
        search:=search^.west
       end
      else if xApart>1.0001 then begin
        for i:=1 to xincrt do if search^.east<>nil then
        search:=search^.east
       end
      else Xdone:=true;

      if yApart<-0.0001 then begin
        for i:=0 downto yincrt do if search^.sth<>nil then
        search:=search^.sth
       end
      else if yApart>1.0001 then begin
        for i:=1 to yincrt do if search^.nth<>nil then
        search:=search^.nth
       end
      else yDone:=true;

      {/// avoid node searching from being trapped in infinite loop ///}
      if loop>=2 then begin
        yDone:=true;xDone:=true;
        outOfDomain:=true;
       end;
       loop:=loop+1;
    until xDone and yDone;

    if not outOfDomain then begin
     {/// calculate aera weighting ///}
     xApart:=(xdv-search^.Ctr.Re);
     yApart:=(ydv-search^.Ctr.Im);

     weighUR:=xApart*yApart*iArea;
     weighUL:=(celength-xApart)*yApart*iArea;
     weighLL:=(celength-xApart)*(celength-yApart)*iArea;
     weighLR:=xApart*(celength-yApart)*iArea;

     {/// smear vorticity to the four surrounding nodes ///}
     search^.sample[bin]:=search^.sample[bin]+magdv*weighLL;
     if search^.nth<>nil then
       search^.nth^.sample[bin]:=search^.nth^.sample[bin]+magdv*weighUL;
     if search^.east<>nil then begin
       search^.east^.sample[bin]:=search^.east^.sample[bin]+magdv*weighLR;
       if search^.east^.nth<>nil then
       search^.east^.nth^.sample[bin]:=search^.east^.nth^.sample[bin]+magdv*weighUR;
     end;
    end;
    curPtr:=curPtr^.next;
  end;
end;

procedure CellClearance(CellStart:PCell);
var
  curRow,curNode: PCell;
begin
  curRow:=CellStart;
  while curRow<>nil do begin
    curNode:=curRow;
    while curNode<>nil do with curNode^ do begin
      if (Clear=CellUnKnown) then Clear:=chkCellinBody(curNode);
      if NodeStart<>nil then CellClearance(NodeStart);
      curNode:=curNode^.UpC;
     end;
    curRow:=curRow^.RtC;
   end;
end;

procedure GetCellCoverBody(GroupStart:PCell);
var
  CellFound: PCell;
  iSeg,bdy: integer;
begin
  for bdy:=1 to numActiveBbody do with bodyAry[bdy]^ do begin
    for iSeg:=1 to mSeg do begin
      CellFound:=aPtInCtocMesh(seg[iSeg]^.pvt,GroupStart);
      if CellFound<>nil then CellFound^.Clear:=CellNotClear;
      CellFound:=aPtInCtocMesh(seg[iSeg]^.datPt,GroupStart);
      if CellFound<>nil then CellFound^.Clear:=CellNotClear;
     end;
   end;
end;

procedure freeCellMemory (var inStart, inEnd: PCell);
var
  tmpChain: Pchain;
  tempMesh,curMesh,tempRow,curRow : PCell;
begin
  curRow:=inStart;
  while curRow<>nil do begin
    tempRow:=curRow^.RtC;
    curMesh:=curRow;
      while curMesh <> nil do with curMesh^ do begin
        tempMesh:=curMesh^.UpC;
        {/// free chain pointer first ///}
        while dvInCell<>nil do begin
          tmpChain:=dvInCell^.nextElem;
          {if dvInCell<>nil then dispose(dvInCell);}
          if dvInCell<>nil then PchainBin(dvInCell);
          dvInCell:=tmpChain;
        end;
        if NodeStart<>nil then freeCellMemory(NodeStart,NodeEnd);
        {if ctoc<>nil then dispose(ctoc);}
        {/// free cell pointer ///}
        {dispose(curMesh);}
        PCellBin(curMesh);
        curMesh:=tempMesh;
      end;
    curRow:=tempRow;
  end;
  inStart:=nil;
  inEnd:=nil;
end;

  (*//////////////////////////////////////////////////////////////////////
    The following procedure generates the grid system for cell-to-cell
    interaction method for calculation of vortex covective velocity.
    ////////////////////////////////////////////////////////////////////*)

procedure resetCtcDataBank(theCtcDataBank:PctcDataBank);
var
  i,j: integer;
begin
  if theCtcDataBank<>nil then begin
   for i:=-maxCtocData to maxCtocData do
    for j:=-maxCtocData to maxCtocData do
     if theCtcDataBank^[i,j]<> nil then begin
       PzCoordBin(theCtcDataBank^[i,j]);
       {dispose(theCtcDataBank^[i,j]);
       theCtcDataBank^[i,j]:=nil;}
     end;
  end;
end;

procedure freeGroupData(var theGroup:PGridData);
begin
  resetCtcDataBank(theGroup^.DataBank);
  if theGroup^.DataBank<>nil then dispose(theGroup^.DataBank);
  dispose(theGroup);
end;

procedure getPCtocMatx(var theCtcDataBank:PctcDataBank);
var i,j:integer;
begin
  if theCtcDataBank=nil then begin
     New(theCtcDataBank);
     {newDataBankMem:=newDataBankMem+1;}
     for i:=-maxCtocData to maxCtocData do
      for j:=-maxCtocData to maxCtocData do theCtcDataBank^[i,j]:=nil;
  end
  else resetCtcDataBank(theCtcDataBank);
end;

procedure getCtcCoeff;
var
  i,j : integer;
  theValue: double;
begin
  if ctcCoeffA=nil then New(ctcCoeffA);
  if ctcCoeffB=nil then New(ctcCoeffB);
  if ctcSumCoeff=nil then New(ctcSumCoeff);
  for i:=0 to ctcOrderLimit do begin
    for j:=0 to ctcOrderLimit do begin
      if (i+j)<=ctcOrderLimit then
       theValue:=factorial(i+j)/(factorial(i)*factorial(j))
      else theValue:=0.0;
      ctcCoeffA^[i,j]:=theValue;
      ctcCoeffB^[i,j]:=sin((2*(i+j)+1)*pi*0.5)*theValue;
     end;
    ctcSumCoeff^[i]:=sin((2*i+1)*pi*0.5);
   end;
end;

procedure ClearDvMemoryCtoc;
var
  j: integer;
  curPtr: dvPtr;
begin
  curPtr:=dvStart;
  while curPtr<>nil do begin
   for j:=0 to numSubLevel do
    if curPtr^.zMoment[j]<>nil then curPtr^.zMoment[j]^.ctcPter:=nil;
   curPtr:=curPtr^.next;
  end;
end;

procedure getCtcNodes(var theStart,theEnd:PCell;LLeft,URight:zNumber;
                          theLevel:integer);
{calculate coordinates of nodes in CTC mesh at assigned Level
callers  Proc. formSubCell;
         Proc. FormCtcMesh;
}

var
  tmpCell: TCell;
  iCol,jRow: integer;
  TotalCol,TotalRow: integer;
  curNodeA,CurNodeB,NodeA,NodeB,tempCRow,tempPtr: PCell;
  xx,yy,cellWidth,cellHeight: double;
begin
  {writeln(' before getCtcNodes : ',nPCellTrash:4:0);}
  cellWidth:=CGroup[theLevel]^.xWidth;
  cellHeight:=CGroup[theLevel]^.yHeight;
  TotalRow:=Round(abs(URight.Im-LLeft.Im)/cellHeight);
  TotalCol:=Round(abs(URight.Re-LLeft.Re)/cellWidth);
  if theLevel=0 then begin
   if TotalRow=0 then TotalRow:=1;
   if TotalCol=0 then TotalCol:=1;
  end;
  if theStart<>nil then freeCellMemory(theStart,theEnd);

  {/// calculate coordinate of node points ///}
  { columns are arranged in ascending order of x-coordinate }

  for iCol:=1 to TotalCol do begin
   xx:=LLeft.Re+(iCol-0.50)*cellWidth;
   for jRow:=1 to TotalRow do begin
   { rows are arranged in ascending order of y-coordinate }
    yy:=LLeft.Im+(jRow-0.50)*cellHeight;

    getNewPCell(tempPtr);
    tempPtr^.PData:=CGroup[theLevel];
    tempPtr^.Ctr.Re:=xx;
    tempPtr^.Ctr.Im:=yy;

    {/// if the list is empty, make the new record the only one \\\}
    if theStart=nil then begin
     theStart:=TempPtr;
     theEnd:=TempPtr;
     tempCRow:=TempPtr;
    end
    else begin
     TempPtr^.DnC:=theEnd;
     theEnd^.UpC:=TempPtr;
     theEnd:=TempPtr;
     if jRow=1 then begin
      tempPtr^.LfC:=tempCRow;
      tempCRow^.RtC:=tempPtr;
      tempCRow:=tempPtr;
      tempCRow^.DnC^.UpC:=nil;
      tempCRow^.DnC:=nil;
     end;
    end;
   end;{for jRow}
  end;{for iCol}

   {/// form cross-row pointer link between pointers ///}
  NodeA:=theStart;
  NodeB:=NodeA^.RtC;
  while NodeB<>nil do begin
   CurNodeA:=NodeA^.UpC;
   CurNodeB:=NodeB^.UpC;
   while curNodeA<>nil do begin
    curNodeA^.RtC:=curNodeB;
    curNodeB^.LfC:=curNodeA;
    curNodeA:=curNodeA^.UpC;
    curNodeB:=curNodeB^.UpC;
   end;
   NodeA:=NodeB;
   NodeB:=NodeB^.RtC;
  end;
end;

procedure formSubCell(whichNode:PCell);
var
  LLCorner,URCorner:zNumber;
begin
  with whichNode^ do begin
   LLCorner.Re:=ctr.Re-CGroup[PData^.Level]^.xWidth*0.5;
   LLCorner.Im:=ctr.Im-CGroup[PData^.Level]^.yHeight*0.5;
   URCorner.Re:=ctr.Re+CGroup[PData^.Level]^.xWidth*0.5;
   URCorner.Im:=ctr.Im+CGroup[PData^.Level]^.yHeight*0.5;
   getCtcNodes(NodeStart,NodeEnd,LLCorner,URCorner,PData^.Level+1);
  end;
end;

procedure formAllSubCell(CellStart:PCell);
var
  curChain: PChain;
  curRow,curNode,searchNode: PCell;
begin
  if CellStart^.PData^.Level<numSubLevel then begin
   curRow:=CellStart;
   while curRow<>nil do begin
    curNode:=curRow;
    while curNode<>nil do begin
      if (curNode^.NodeStart=nil) then formSubCell(curNode);
      formAllSubCell(curNode^.NodeStart);
      curNode:=curNode^.UpC;
    end;
    curRow:=curRow^.RtC;
   end;
  end;
end;

procedure formCtoCMesh;
var i: integer;
begin
  {/// form the main mesh system ///}
  for i:=0 to numSubLevel do begin
   if CGroup[i]=nil then begin
     New(CGroup[i]);
     {newCGroupMem:=newCGroupMem+1;}
     with CGroup[i]^ do begin
      level:=0;
      xWidth:=0.0;
      yHeight:=0.0;
      InvWidth:=0.0;
      InvHeight:=0.0;
      DataBank:=nil;
     end;
   end;
   with CGroup[i]^ do begin
    if i=0 then
     xWidth:=exp(numSubLevel*ln(NumSubCell))*basicMeshFactor*cutOff*ln(ReyNum)
    else xWidth:=CGroup[i-1]^.xWidth/numSubCell;
    yHeight:=xWidth;
    level:=i;
    InvWidth:=1.0/xWidth;
    InvHeight:=1.0/yHeight;

    if i=0 then getCtcNodes(ctocStart,ctocEnd,LLFlowField,URFlowField,level);
    getPCtocMatx(DataBank);
   end;{with}
  end;
  if (ctcCoeffA=nil) or (ctcCoeffB=nil) or (ctcSumCoeff=nil) then getCtcCoeff;
  ClearDvMemoryCtoc;
  if reformCtoc then reformCtoc:=false;
  getCellCoverBody(CtocStart);
  cellClearance(CtocStart);
end;

procedure resetTCtoc(inPCtoc:PCtoc);
var i:integer;
begin
  with inPCtoc^ do begin
   Zm0:=0.0;
   for i:=1 to ctcOrderLimit do begin
     Zmmt[i].Re:=0.0;
     Zmmt[i].Im:=0.0;
   end;
  end;
end;

procedure resetCtocNode(theStart,theEnd:PCell);
var
  curNode, curCRow: PCell;
  curChain: PChain;
begin
  curCRow:=theStart;
  while curCRow<>nil do begin
    curNode:=curCRow;
    while curNode<>nil do with curNode^ do begin { do if cell is NOT empty }
      Zdv:=0;
      if dvInCell<>nil then begin
        {// MUST reset ctoc because calculation is accumulative ///}
        if ctoc<>nil then ResetTCtoc(ctoc);
        while dvInCell<>nil do begin { free previously allocated dv in cell }
          curChain:=dvInCell^.nextElem;
          {dispose(dvInCell);}
          PchainBin(dvInCell);
          dvInCell:=curChain;
        end;
        resetCtocNode(NodeStart,NodeEnd);
       end; { completed a cell }
      curNode:=UpC; { go on to next cell down the row }
     end;   { completed one row }
    curCRow:=curCRow^.RtC;
   end;
end;

procedure findNode(theDv:dvPtr;theStart:PCell);
const
  halfDimTolce=0.500001;
var
  tmpPtr: Pchain;
  searchNode: PCell;
  outOfDomain,xDone,yDone: boolean;
  theLevel,nOrder,xincrt,yincrt,i,loop: integer;
  widthFactor,heightFactor,thisMag,xApart,yApart: double;
begin
  theLevel:=theStart^.PData^.Level;
  with theDv^ do begin
  if zMoment[theLevel]=nil then begin
     New(zMoment[theLevel]);
     {newzMmtMem:=newzMmtMem+1;}
     zMoment[theLevel]^.CtcPter:=nil;
     searchNode:=theStart;
  end
  else begin
   if (theLevel=0) and (zMoment[theLevel]^.CtcPter<>nil) then
     searchNode:=zMoment[theLevel]^.CtcPter
   else searchNode:=theStart;
  end;

  widthFactor:=CGroup[theLevel]^.InvWidth;
  heightFactor:=CGroup[theLevel]^.InvHeight;

  outOfDomain:=false;
  xDone:=false;
  yDone:=false;
  loop:=0;
  repeat
    xApart:=(vortex.pos.Re-searchNode^.Ctr.Re)*widthFactor;
    yApart:=(vortex.pos.Im-searchNode^.Ctr.Im)*heightFactor;
    xincrt:=round(xApart);
    yincrt:=round(yApart);
    if xApart<-halfDimTolce then begin
      for i:=-1 downto xincrt do if searchNode^.LfC<>nil then
       searchNode:=searchNode^.LfC;
     end
    else
     if xApart>halfDimTolce then begin
       for i:=1 to xincrt do if searchNode^.RtC<>nil then
        searchNode:=searchNode^.RtC;
      end
    else Xdone:=true;

    if yApart<-halfDimTolce then begin
      for i:=-1 downto yincrt do if searchNode^.DnC<>nil then
       searchNode:=searchNode^.DnC;
     end
    else
    if yApart>halfDimTolce then begin
      for i:=1 to yincrt do if searchNode^.UpC<>nil then
       searchNode:=searchNode^.UpC;
     end
    else yDone:=true;

    {/// avoid node searching from being trapped in infinite loop ///}
    if loop>=2 then begin
      yDone:=true;xDone:=true;
      outOfDomain:=true;
     end;
    loop:=loop+1;
  until xDone and yDone;

  end; { with theDv^ }

  if not outOfDomain then begin

    getNewPchain(tmpPtr);
    tmpPtr^.dvElem:=theDv;
    tmpPtr^.nextElem:=nil;

    { store searchNode for as starting point in next searching }
    theDv^.zMoment[theLevel]^.ctcPter:=searchNode;
    searchNode^.Zdv:=searchNode^.Zdv+1;
    if searchNode^.ctoc=nil then begin
      New(searchNode^.ctoc);
      {newCtocMem:=newCtocMem+1;}
      resetTCtoc(searchNode^.ctoc); {/// MUST initialise data ///}
    end;
    {/// append dv to element-list of searchNode ///}
    if searchNode^.dvInCell = nil then searchNode^.dvInCell:=tmpPtr
    else begin
      TmpPtr^.nextElem:=searchNode^.dvInCell;
      searchNode^.dvInCell:=TmpPtr;
    end;

    {/// calculate moments about cell center ///}
    with theDv^.zMoment[theLevel]^ do begin
      zzj[1].Re:=theDv^.vortex.pos.Re - searchNode^.Ctr.Re;
      zzj[1].Im:=theDv^.vortex.pos.Im - searchNode^.Ctr.Im;
      {thisMag:=theDv^.vortex.mag;}
      {/// effect vorticity decay ///}
      thisMag:=vortDecay(theDv);

      searchNode^.ctoc^.Zm0:= searchNode^.ctoc^.Zm0 + thisMag;
      {/// first moment ///}
      searchNode^.ctoc^.Zmmt[1].Re:= searchNode^.ctoc^.Zmmt[1].Re
                                 + thisMag*zzj[1].Re;
      searchNode^.ctoc^.Zmmt[1].Im:= searchNode^.ctoc^.Zmmt[1].Im
                                 + thisMag*zzj[1].Im;
      for nOrder:=2 to ctcOrderLimit do begin
        zzj[nOrder].Re:=zzj[1].Re*zzj[nOrder-1].Re-zzj[1].Im*zzj[nOrder-1].Im;
        zzj[nOrder].Im:=zzj[1].Re*zzj[nOrder-1].Im+zzj[1].Im*zzj[nOrder-1].Re;
        searchNode^.ctoc^.Zmmt[nOrder].Re:= searchNode^.ctoc^.Zmmt[nOrder].Re
                                            + thisMag*zzj[nOrder].Re;
        searchNode^.ctoc^.Zmmt[nOrder].Im:= searchNode^.ctoc^.Zmmt[nOrder].Im
                                            + thisMag*zzj[nOrder].Im;
      end;
    end; {with}
  end;
end;

[global]
function aPtInCtocMesh(thePt:zNumber;theStart:PCell):PCell;
const
  halfDimTolce=0.500001;
var
  widthFactor,heightFactor,
  xApart,yApart: double;
  xincrt,yincrt,i,loop: integer;
  outOfDomain,xDone,yDone: boolean;
  theLevel: integer;
  searchNode:PCell;
begin
   theLevel:=theStart^.PData^.Level;
   searchNode:=theStart;
   widthFactor:=CGroup[theLevel]^.InvWidth;
   heightFactor:=CGroup[theLevel]^.InvHeight;

   outOfDomain:=false;
   xDone:=false;
   yDone:=false;
   loop:=0;
   repeat
    xApart:=(thePt.Re-searchNode^.Ctr.Re)*widthFactor;
    yApart:=(thePt.Im-searchNode^.Ctr.Im)*heightFactor;
    xincrt:=round(xApart);
    yincrt:=round(yApart);
    if xApart<-halfDimTolce then begin
      for i:=-1 downto xincrt do if searchNode^.LfC<>nil then
       searchNode:=searchNode^.LfC;
     end
    else
     if xApart>halfDimTolce then begin
       for i:=1 to xincrt do if searchNode^.RtC<>nil then
        searchNode:=searchNode^.RtC;
      end
    else Xdone:=true;

    if yApart<-halfDimTolce then begin
      for i:=-1 downto yincrt do if searchNode^.DnC<>nil then
       searchNode:=searchNode^.DnC;
     end
    else
    if yApart>halfDimTolce then begin
      for i:=1 to yincrt do if searchNode^.UpC<>nil then
       searchNode:=searchNode^.UpC;
     end
    else yDone:=true;

    {/// avoid node searching from being trapped in infinite loop ///}
    if loop>=2 then begin
      yDone:=true;xDone:=true;
      outOfDomain:=true;
     end;
    loop:=loop+1;
   until xDone and yDone;
  if OutOfDomain then aPtInCtocMesh:=nil
  else if searchNode^.NodeStart=nil then aPtInCtocMesh:=searchNode
  else aPtInCtocMesh:=aPtInCtocMesh(thePt,searchNode^.NodeStart);
end;

procedure assignDVsubLevel(CellStart:PCell);
var
  curChain: PChain;
  curRow,curNode,searchNode: PCell;
begin
  if CellStart^.PData^.Level<numSubLevel then begin
  curRow:=CellStart;
  while curRow<>nil do begin
    curNode:=curRow;
    while curNode<>nil do begin
      if (curNode^.dvInCell<>nil) then begin
        if (curNode^.NodeStart=nil) then begin
          formSubCell(curNode);
          GetCellCoverBody(curNode^.NodeStart);
          CellClearance(curNode^.NodeStart);
         end;
        curChain:=curNode^.dvInCell;
        while curChain<>nil do with curChain^ do begin
          FindNode(dvElem,curNode^.NodeStart);
          curChain:=curChain^.nextElem;
        end;
        assignDVsubLevel(curNode^.NodeStart);
       end;
      curNode:=curNode^.UpC;
     end;
    curRow:=curRow^.RtC;
   end;
  end;
end;

[global]
procedure assignDVtoCtoc;
var
  curPtr: dvPtr;
begin
  if (CtoCStart=nil) or ReformCtoc then formCtoCMesh;
  resetCtocNode(ctocStart,ctocEnd);     { clear previous cell data }
  curPtr:=dvStart;
  while curPtr<>nil do begin
    {!!! search for nearest lower-left node !!!}
    FindNode(curPtr,ctocStart);
    curPtr:=curPtr^.next;
  end; { while }
  assignDVsubLevel(CtocStart);
  getCtocStatistic;
end;

procedure EasyFindNode(theDv:dvPtr;theStart:PCell);
const
  halfDimTolce=0.500001;
var
  widthFactor,heightFactor,thisMag,xApart,yApart : double;
  theLevel,nOrder,xincrt,yincrt,i,loop           : integer;
  outOfDomain,xDone,yDone                        : boolean;
  tmpPtr                                         : Pchain;
  searchNode                                     : PCell;
  tmpZmmt                                        : TzMoment;
begin
  theLevel:=theStart^.PData^.Level;
  searchNode:=theStart;
  widthFactor:=CGroup[theLevel]^.InvWidth;
  heightFactor:=CGroup[theLevel]^.InvHeight;

  outOfDomain:=false;
  xDone:=false;
  yDone:=false;
  loop:=0;
  repeat
    xApart:=(theDv^.vortex.pos.Re-searchNode^.Ctr.Re)*widthFactor;
    yApart:=(theDV^.vortex.pos.Im-searchNode^.Ctr.Im)*heightFactor;
    xincrt:=round(xApart);
    yincrt:=round(yApart);
    if xApart<-halfDimTolce then begin
     for i:=-1 downto xincrt do if searchNode^.LfC<>nil then
      searchNode:=searchNode^.LfC;
    end
    else
    if xApart>halfDimTolce then begin
     for i:=1 to xincrt do if searchNode^.RtC<>nil then
      searchNode:=searchNode^.RtC;
    end
    else Xdone:=true;

    if yApart<-halfDimTolce then begin
     for i:=-1 downto yincrt do if searchNode^.DnC<>nil then
      searchNode:=searchNode^.DnC;
    end
    else
    if yApart>halfDimTolce then begin
     for i:=1 to yincrt do if searchNode^.UpC<>nil then
      searchNode:=searchNode^.UpC;
    end
    else yDone:=true;
    {/// avoid node searching from being trapped in infinite loop ///}
    if loop>=2 then begin
     yDone:=true;xDone:=true;
     outOfDomain:=true;
    end;
    loop:=loop+1;
  until xDone and yDone;


  if not outOfDomain then begin

   getNewPchain(tmpPtr);
   tmpPtr^.dvElem:=theDv;
   tmpPtr^.nextElem:=nil;

   searchNode^.Zdv:=searchNode^.Zdv+1;
   if searchNode^.ctoc=nil then begin
    New(searchNode^.ctoc);
    {newCtocMem:=newCtocMem+1;}
    resetTCtoc(searchNode^.ctoc); {/// MUST initialise data ///}
   end;
    {/// append dv to element-list of searchNode ///}
   if searchNode^.dvInCell = nil then searchNode^.dvInCell:=tmpPtr
   else begin
    TmpPtr^.nextElem:=searchNode^.dvInCell;
    searchNode^.dvInCell:=TmpPtr;
   end;

   {/// calculate moments about cell center ///}
   with tmpZmmt do begin
    zzj[1].Re:=theDv^.vortex.pos.Re - searchNode^.Ctr.Re;
    zzj[1].Im:=theDv^.vortex.pos.Im - searchNode^.Ctr.Im;
    if theDv^.vortex.nearBdy<>0 then thisMag:=theDv^.vortex.mag
     else thisMag:=vortDecay(theDv);
    {thisMag:=vortDecay(theDv);}
    searchNode^.ctoc^.Zm0:= searchNode^.ctoc^.Zm0 + thisMag;
    {/// first moment ///}
    searchNode^.ctoc^.Zmmt[1].Re:= searchNode^.ctoc^.Zmmt[1].Re
                                 + thisMag*zzj[1].Re;
    searchNode^.ctoc^.Zmmt[1].Im:= searchNode^.ctoc^.Zmmt[1].Im
                                 + thisMag*zzj[1].Im;
    for nOrder:=2 to ctcOrderLimit do begin
     zzj[nOrder].Re:=zzj[1].Re*zzj[nOrder-1].Re-zzj[1].Im*zzj[nOrder-1].Im;
     zzj[nOrder].Im:=zzj[1].Re*zzj[nOrder-1].Im+zzj[1].Im*zzj[nOrder-1].Re;
     searchNode^.ctoc^.Zmmt[nOrder].Re:= searchNode^.ctoc^.Zmmt[nOrder].Re
                                            + thisMag*zzj[nOrder].Re;
     searchNode^.ctoc^.Zmmt[nOrder].Im:= searchNode^.ctoc^.Zmmt[nOrder].Im
                                            + thisMag*zzj[nOrder].Im;
    end;
   end; {with}
   end;
  {dispose(tmpZmmt);}
end;

procedure EasyAssignSub(CellStart:PCell);
var
  curChain: PChain;
  curRow,curNode,searchNode: PCell;
begin
  if CellStart^.PData^.Level<numSubLevel then begin
  curRow:=CellStart;
  while curRow<>nil do begin
   curNode:=curRow;
   while curNode<>nil do begin
    if (curNode^.dvInCell<>nil) then begin
     {to form submesh if not done so }
     if (curNode^.NodeStart=nil) then begin
      formSubCell(curNode);
      GetCellCoverBody(curNode^.NodeStart);
      CellClearance(curNode^.NodeStart);
     end;
     { assign all dv in one cell to submesh }
     curChain:=curNode^.dvInCell;
     while curChain<>nil do with curChain^ do begin
      EasyFindNode(dvElem,curNode^.NodeStart);
      curChain:=curChain^.nextElem;
     end;
     {go into further submesh level, if any }
     EasyAssignSub(curNode^.NodeStart);
    end;
    curNode:=curNode^.UpC;
   end;
   curRow:=curRow^.RtC;
  end;
  end;
end;

[global]
procedure easyAssign;
var
  curPtr: dvPtr;
begin
  if (CtoCStart=nil) or ReformCtoc then formCtoCMesh;
  resetCtocNode(ctocStart,ctocEnd);     { clear previous cell data }
  curPtr:=dvStart;
  while curPtr<>nil do begin
   {// find for nearest lower-left node, it must be done for CTC method //}
   EasyFindNode(curPtr,ctocStart);
   curPtr:=curPtr^.next;
  end; { while }
  EasyAssignSub(CtocStart);
  getCtocStatistic;
end;

{/// calculate vortex-to-vortex interact velocities, with 2*pi term incl'd ///}
procedure vortexInteract(mPtr,nPtr: dvPtr);
var
  Mmag,Nmag,
  sinCore,cosCore,rsq,oneOver2rsq,Xmn,Ymn: double;
begin
  Xmn:=mPtr^.Vortex.pos.Re-nPtr^.vortex.pos.Re;
  Ymn:=mPtr^.vortex.pos.Im-nPtr^.vortex.pos.Im;

  {/// effect vorticity decay ///}
  Mmag:=vortDecay(mPtr);
  Nmag:=vortDecay(nPtr);
  rsq:=sqr(Xmn)+sqr(Ymn);
  oneOver2rsq:=0.5/(pi*rsq);
  if rsq>cutOffsq then begin
    mPtr^.Vortex.vel.Re:=mPtr^.Vortex.vel.Re+Nmag*Ymn*oneOver2rsq;
    mPtr^.Vortex.vel.Im:=mPtr^.Vortex.vel.Im-Nmag*Xmn*oneOver2rsq;
    nPtr^.Vortex.vel.Re:=nPtr^.Vortex.vel.Re-Mmag*Ymn*oneOver2rsq;
    nPtr^.Vortex.vel.Im:=nPtr^.Vortex.vel.Im+Mmag*Xmn*oneOver2rsq;
  end
  else begin
    {/// using Rankine vortex model ///}
    mPtr^.Vortex.vel.Re:=mPtr^.Vortex.vel.Re+Nmag*Ymn*oneOver2piCSq;
    mPtr^.Vortex.vel.Im:=mPtr^.Vortex.vel.Im-Nmag*Xmn*oneOver2piCSq;
    nPtr^.Vortex.vel.Re:=nPtr^.Vortex.vel.Re-Mmag*Ymn*oneOver2piCSq;
    nPtr^.Vortex.vel.Im:=nPtr^.Vortex.vel.Im+Mmag*Xmn*oneOver2piCSq;
  end;
end;

{/// calcualte CTC interaction of all dv in cellA to cellB ///}
procedure CtocGroupInteract(cellA,cellB:PCell;zLevel:integer);
var
  i,j: integer;
  jChain: Pchain;
  Am0,Bm0: double;
  tmpSum,dump: zNumber;
  sumZmmtAA,sumZmmtBB: array[0..ctcOrderLimit] of zNumber;
begin
  Am0:=cellA^.ctoc^.Zm0;
  Bm0:=cellB^.ctoc^.Zm0;
  sumZmmtAA[0].Re:=Am0;
  sumZmmtAA[0].Im:=0.0;
  sumZmmtBB[0].Re:=Bm0;
  sumZmmtBB[0].Im:=0.0;

  with ZData^ do begin
    { evaluate cell A's moment terms }
    for i:=1 to OrderTerm do begin
      dump.Re:= cellA^.ctoc^.Zmmt[i].Re*IZZab[i].Re
                - cellA^.ctoc^.Zmmt[i].Im*IZZab[i].Im;
      dump.Im:= cellA^.ctoc^.Zmmt[i].Re*IZZab[i].Im
                + cellA^.ctoc^.Zmmt[i].Im*IZZab[i].Re;
      sumZmmtAA[0].Re:=sumZmmtAA[0].Re+dump.Re;
      sumZmmtAA[0].Im:=sumZmmtAA[0].Im+dump.Im;
     end;

    for i:=1 to OrderTerm do begin
      sumZmmtAA[i].Re:=Am0*IZZab[i].Re;
      sumZmmtAA[i].Im:=Am0*IZZab[i].Im;
      for j:=1 to OrderTerm-i do begin
        dump.Re:= cellA^.ctoc^.Zmmt[j].Re*IZZab[j+i].Re
                  - cellA^.ctoc^.Zmmt[j].Im*IZZab[j+i].Im;
        dump.Im:= cellA^.ctoc^.Zmmt[j].Re*IZZab[j+i].Im
                  + cellA^.ctoc^.Zmmt[j].Im*IZZab[j+i].Re;
        sumZmmtAA[i].Re:=sumZmmtAA[i].Re+ctcCoeffA^[i,j]*dump.Re;
        sumZmmtAA[i].Im:=sumZmmtAA[i].Im+ctcCoeffA^[i,j]*dump.Im;
       end;
     end;

    { calculate cell B's moment terms }
    for i:=1 to OrderTerm do begin
      dump.Re:= cellB^.ctoc^.Zmmt[i].Re*IZZab[i].Re
                - cellB^.ctoc^.Zmmt[i].Im*IZZab[i].Im;
      dump.Im:= cellB^.ctoc^.Zmmt[i].Re*IZZab[i].Im
                + cellB^.ctoc^.Zmmt[i].Im*IZZab[i].Re;
      sumZmmtBB[0].Re:=sumZmmtBB[0].Re+ctcCoeffB^[0,i]*dump.Re;
      sumZmmtBB[0].Im:=sumZmmtBB[0].Im+ctcCoeffB^[0,i]*dump.Im;
     end;

    for i:=1 to OrderTerm do begin
      sumZmmtBB[i].Re:=ctcCoeffB^[i,0]*Bm0*IZZab[i].Re;
      sumZmmtBB[i].Im:=ctcCoeffB^[i,0]*Bm0*IZZab[i].Im;
      for j:=1 to OrderTerm-i do begin
        dump.Re:= cellB^.ctoc^.Zmmt[j].Re*IZZab[j+i].Re
                  - cellB^.ctoc^.Zmmt[j].Im*IZZab[j+i].Im;
        dump.Im:= cellB^.ctoc^.Zmmt[j].Re*IZZab[j+i].Im
                  + cellB^.ctoc^.Zmmt[j].Im*IZZab[j+i].Re;
        sumZmmtBB[i].Re:=sumZmmtBB[i].Re+ctcCoeffB^[i,j]*dump.Re;
        sumZmmtBB[i].Im:=sumZmmtBB[i].Im+ctcCoeffB^[i,j]*dump.Im;
       end;
     end;
  end; { with ZData }

  { evaluate velocities of dv in cell B due to cell A }
  jChain:=cellB^.dvInCell;
  while jChain<>nil do with jChain^.dvElem^ do begin
     tmpSum.Re:=sumZmmtAA[0].Re;
     tmpSum.Im:=sumZmmtAA[0].Im;
     for i:=1 to ZData^.orderTerm do begin
       dump.Re:= zMoment[zLevel]^.zzj[i].Re*sumZmmtAA[i].Re
                 - zMoment[zLevel]^.zzj[i].Im*sumZmmtAA[i].Im;
       dump.Im:= zMoment[zLevel]^.zzj[i].Re*sumZmmtAA[i].Im
                 + zMoment[zLevel]^.zzj[i].Im*sumZmmtAA[i].Re;
       tmpSum.Re:=tmpSum.Re+ctcSumCoeff^[i]*dump.Re;
       tmpSum.Im:=tmpSum.Im+ctcSumCoeff^[i]*dump.Im;
      end;
     dump.Re:=ZData^.velCoeff.Re*tmpSum.Re-ZData^.velCoeff.Im*tmpSum.Im;
     dump.Im:=ZData^.velCoeff.Re*tmpSum.Im+ZData^.velCoeff.Im*tmpSum.Re;
     vortex.vel.Re:=vortex.vel.Re + dump.Re;
     vortex.vel.Im:=vortex.vel.Im - dump.Im;
     jChain:=jChain^.nextElem;
  end;{ while/with}

  { calculate velocities of dv in cell A due to cell B }
  jChain:=cellA^.dvInCell;
  while jChain<>nil do with jChain^.dvElem^ do begin
     tmpSum.Re:=sumZmmtBB[0].Re;
     tmpSum.Im:=sumZmmtBB[0].Im;
     for i:=1 to ZData^.orderTerm do begin
       dump.Re:= zMoment[zLevel]^.zzj[i].Re*sumZmmtBB[i].Re
                 - zMoment[zLevel]^.zzj[i].Im*sumZmmtBB[i].Im;
       dump.Im:= zMoment[zLevel]^.zzj[i].Re*sumZmmtBB[i].Im
                 + zMoment[zLevel]^.zzj[i].Im*sumZmmtBB[i].Re;
       tmpSum.Re:=tmpSum.Re+ctcSumCoeff^[i]*dump.Re;
       tmpSum.Im:=tmpSum.Im+ctcSumCoeff^[i]*dump.Im;
      end;
     dump.Re:=ZData^.velCoeff.Re*tmpSum.Re-ZData^.velCoeff.Im*tmpSum.Im;
     dump.Im:=ZData^.velCoeff.Re*tmpSum.Im+ZData^.velCoeff.Im*tmpSum.Re;
     vortex.vel.Re:=vortex.vel.Re - dump.Re;
     vortex.vel.Im:=vortex.vel.Im + dump.Im;
     jChain:=jChain^.nextElem;
  end;{ while/with }
end;  { Proc. CtocGroupInteract }

{// evaluate the velocity at a point induced by vortices in a CTC cell ///}
procedure CtocGroupToPt(cellA:PCell; var indVel: zNumber);
var
  i: integer;
  sumZmmt: zNumber;
begin
  sumZmmt.Re:=cellA^.ctoc^.Zm0;
  sumZmmt.Im:=0.0;

  with ZData^ do begin
   for i:=1 to OrderTerm do begin
     sumZmmt.Re:= sumZmmt.Re+(cellA^.ctoc^.Zmmt[i].Re*IZZab[i].Re
                 - cellA^.ctoc^.Zmmt[i].Im*IZZab[i].Im);
     sumZmmt.Im:= sumZmmt.Im+(cellA^.ctoc^.Zmmt[i].Re*IZZab[i].Im
                 + cellA^.ctoc^.Zmmt[i].Im*IZZab[i].Re);
   end;
  indVel.Re:=indVel.Re + (ZData^.velCoeff.Re*sumZmmt.Re
                       - ZData^.velCoeff.Im*sumZmmt.Im);
  indVel.Im:=indVel.Im - (ZData^.velCoeff.Re*sumZmmt.Im
                       + ZData^.velCoeff.Im*sumZmmt.Re);
  end;
end;  { Proc. CtocGroupToPt }

{/// evaluate the velocity of a group of dv induced by a single dv ///}
procedure CtocPtToGroup(cellB:PCell;zLevel:integer);
var
  i,j: integer;
  jChain: Pchain;
  tmpSum,dump: zNumber;
begin
  { evaluate velocities of dv in cell B due to cell A }
  jChain:=cellB^.dvInCell;
  while jChain<>nil do with jChain^.dvElem^ do begin
    tmpSum.Re:=1.0;
    tmpSum.Im:=0.0;
    for i:=1 to ZData^.orderTerm do begin
      dump.Re:= zMoment[zLevel]^.zzj[i].Re*ZData^.IZZab[i].Re
              - zMoment[zLevel]^.zzj[i].Im*ZData^.IZZab[i].Im;
      dump.Im:= zMoment[zLevel]^.zzj[i].Re*ZData^.IZZab[i].Im
              + zMoment[zLevel]^.zzj[i].Im*ZData^.IZZab[i].Re;
      tmpSum.Re:=tmpSum.Re+ctcSumCoeff^[i]*dump.Re;
      tmpSum.Im:=tmpSum.Im+ctcSumCoeff^[i]*dump.Im;
     end;
    dump.Re:=ZData^.velCoeff.Re*tmpSum.Re-ZData^.velCoeff.Im*tmpSum.Im;
    dump.Im:=ZData^.velCoeff.Re*tmpSum.Im+ZData^.velCoeff.Im*tmpSum.Re;
    vortex.vel.Re:=vortex.vel.Re + dump.Re;
    vortex.vel.Im:=vortex.vel.Im - dump.Im;
    jChain:=jChain^.nextElem;
  end;{ while/with}
end;  { Proc. CtocGroupInteract }

function evaluateAccy(base,compare: zNumber):boolean;
var
   errRe,ErrIm : double;
begin
   if base.Re<>0.0 then
     errRe:=abs((compare.Re-base.Re)/base.Re)
   else
     errRe:=abs(compare.Re-base.Re);
   if base.Im<>0.0 then
     errIm:=abs((compare.Im-base.Im)/base.Im)
   else
     errIm:=abs(compare.Im-base.Im);
   if (errRe < errLimit) and (errIm<errLimit) then
    evaluateAccy:=true
   else evaluateAccy:=false;
end;

function getOptimalTerm(theGrid:PGridData;ZZab:zNumber):integer;
{ find the number of terms req'd to achieve desired accuracy in the cell-to-
  cell approximation. }
label 1;
var
  done: boolean;
  Orders: integer;
  Exact,Zij,zzi,zzj,Zimj,SerSum,RHS,orderOne,dump: zNumber;
begin
  { set complex coordinate of the two points relative to cell centers }
  zzi.Re:=-0.5*theGrid^.xWidth;
  zzi.Im:=0.5*theGrid^.yHeight;
  zzj.Re:=0.5*theGrid^.xWidth;
  zzj.Im:=-0.5*theGrid^.yHeight;

  { calculate zi-zj which is used in summation series }
  Zimj.Re:=zzi.Re-zzj.Re;
  Zimj.Im:=zzi.Im-zzj.Im;

  { set complex coord. between the two points }
  Zij.Re:=ZZab.Re+zzj.Re-zzi.Re;
  Zij.Im:=ZZab.Im+zzj.Im-zzi.Im;

  { calculate exact LHS, 1/Zij }
  zInv(Zij,Exact);
  Orders:=0;
  SerSum.Re:=1.0;
  serSum.Im:=0.0;
  zDiv(SerSum,ZZab,RHS);

  orders:=1;
  zDiv(Zimj,ZZab,orderOne);              { calc (zi-zj)/Zab }
  SerSum.Re:=serSum.Re+orderOne.Re;
  serSum.Im:=serSum.Im+orderOne.Im;

  zDiv(SerSum,ZZab,RHS);
  done:=evaluateAccy(Exact,RHS);
  if done then
   begin
     getOptimalTerm:=Orders;             { orders=1 }
     goto 1;
   end;
  dump.Re:=orderOne.Re;
  dump.Im:=orderOne.Im;
  repeat
    { calculate series sum }
    orders:=orders+1;
    zProd(orderOne,dump,RHS);
    dump.Re:=RHS.Re;
    dump.Im:=RHS.Im;
    serSum.Re:=serSum.Re+RHS.Re;
    serSum.Im:=serSum.Im+RHS.Im;
    zDiv(SerSum,ZZab,RHS);
    done:=evaluateAccy(Exact,RHS);
  until Done or (Orders>ctcOrderLimit);
  getOptimalTerm:=orders;
1:end;

(*
                                x  cell B
                              /
             --|----|----|--/-|----|----|----|---
               |    |    |/   |    |    |    |
            ---|----|---/|----|----|----|----|---
               |    | /  |    |    |    |    |
             --|----/----|----|----|----|----|----
               |  / |    |    |    |    |    |
             --|----|----|----|----|----|----|--
               | cell A  |                   |

    Zab = Zb - Za, is always pointing left (i.e. Zab.Re >=0.0);
*)

procedure QCtocmatxData(theGrid:PGridData;ctcX,ctcY:integer);
var
  i: integer;
  ImodZab: double;
  Zab,dump: znumber;
begin
  if ((ctcX>=-maxCtocData) and (ctcX<=maxCtocData))
   and ((ctcY>=-maxCtocData) and (ctcY<=maxCtocData)) then begin
    if theGrid^.DataBank^[ctcX,ctcY]<>nil then
      ZData:=theGrid^.DataBank^[ctcX,ctcY]
    else begin {/// if data not yet available, calculate once and for all ///}
      {New(theGrid^.DataBank^[ctcX,ctcY]);}
      getNewPzCoord(theGrid^.DataBank^[ctcX,ctcY]);
      with theGrid^.DataBank^[ctcX,ctcY]^ do begin
        Zab.Re:=ctcX*theGrid^.xWidth;
        Zab.Im:=ctcY*theGrid^.yHeight;
        orderTerm:=getOptimalTerm(theGrid,Zab);           { value for storage }
        {/// calculate i/(2*pi*Zab) ///}
        IModZab:=1.0/(sqr(Zab.Re)+sqr(Zab.Im));
        VelCoeff.Re:=Zab.Im*ImodZab*0.5/pi;
        VelCoeff.Im:=Zab.Re*ImodZab*0.5/pi;

        IZZab[1].Re:=Zab.Re*ImodZab;
        IZZab[1].Im:=-Zab.Im*ImodZab;
        for i:=2 to ctcOrderLimit do
         zProd(IZZab[1],IZZab[i-1],IZZab[i]);

        { assign value for output }
        ZData:=theGrid^.DataBank^[ctcX,ctcY];
       end; { with }
     end;
   end { if xIndex & yIndex <=maxCallData }
  else begin { take care of out-of-range indexes }
     if specialBox=nil then New(specialBox);{getMem(specialBox, sizeOf(TzCoord));}
     with specialBox^ do begin
      orderTerm:=1;
      Zab.Re:=ctcX*theGrid^.xWidth;
      Zab.Im:=ctcY*theGrid^.yHeight;
      IModZab:=1.0/(sqr(Zab.Re)+sqr(Zab.Im));
      IZZab[1].Re:=Zab.Re*ImodZab;
      IZZab[1].Im:=-Zab.Im*ImodZab;
      VelCoeff.Re:=Zab.Im*ImodZab*0.5/pi;
      VelCoeff.Im:=Zab.Re*ImodZab*0.5/pi;
     end;
     ZData:=specialBox;
   end;
end;

procedure GroupToPtData(theCell:PCell;thePoint:zNumber);
var
  i: integer;
  ImodZab: double;
  Zab,dump: znumber;
begin
  if specialBox=nil then New(specialBox);{getMem(specialBox, sizeOf(TzCoord));}
  with specialBox^ do begin
    Zab.Re:=thePoint.Re-theCell^.Ctr.Re;
    Zab.Im:=thePoint.Im-theCell^.Ctr.Im;

    orderTerm:=getOptimalTerm(theCell^.PData,Zab);
    {/// calculate i/(2*pi*Zab) ///}
    IModZab:=1.0/(sqr(Zab.Re)+sqr(Zab.Im));
    VelCoeff.Re:=Zab.Im*ImodZab*0.5/pi;
    VelCoeff.Im:=Zab.Re*ImodZab*0.5/pi;

    IZZab[1].Re:=Zab.Re*ImodZab;
    IZZab[1].Im:=-Zab.Im*ImodZab;
    for i:=2 to CtcOrderLimit do
     zProd(IZZab[1],IZZab[i-1],IZZab[i]);
   end;
  ZData:=specialBox;
end;

procedure dvToGpData(theDv:dvPtr;theCell:PCell);
var
  i: integer;
  curMag,
  ImodZab: double;
  Zab,dump: znumber;
begin
  if specialBox=nil then New(specialBox);{getMem(specialBox, sizeOf(TzCoord));}
  with specialBox^ do begin
    Zab.Re:=theCell^.Ctr.Re-theDv^.Vortex.pos.Re;
    Zab.Im:=theCell^.Ctr.Im-theDv^.Vortex.pos.Im;

    orderTerm:=getOptimalTerm(theCell^.PData,Zab);
    {/// calculate i/(2*pi*Zab) ///}
    IModZab:=1.0/(sqr(Zab.Re)+sqr(Zab.Im));
    {curMag:=theDv^.vortex.mag;}
    {/// effect vorticity decay ///}
    curMag:=vortDecay(theDv);

    VelCoeff.Re:=curMag*Zab.Im*ImodZab*0.5/pi;
    VelCoeff.Im:=curMag*Zab.Re*ImodZab*0.5/pi;

    IZZab[1].Re:=Zab.Re*ImodZab;
    IZZab[1].Im:=-Zab.Im*ImodZab;
    for i:=2 to orderTerm do
     zProd(IZZab[1],IZZab[i-1],IZZab[i]);
   end;
  ZData:=specialBox;
end;

procedure UVwithinCell(theCell:PCell);
{/// calculate mutual interaction of dv within same cell ///}
var
  mElem,nElem: Pchain;
begin
  mElem:=theCell^.dvInCell;
  while mElem<>nil do begin
    nElem:=mElem^.nextElem;
    while nElem<>nil do begin
      vortexInteract(mElem^.dvElem,nElem^.dvElem);
      nElem:=nElem^.nextElem;
     end;
    mElem:=mElem^.nextElem;
   end;{while mElem}
end;

procedure UVbetweenCell(theCellA,theCellB:PCell);
var
  mElem,nElem: Pchain;
begin
  mElem:=theCellA^.dvInCell;
  while mElem<>nil do begin
    nElem:=theCellB^.dvInCell;
    while nElem<>nil do begin
      vortexInteract(mElem^.dvElem, nElem^.dvElem);
      nElem:=nElem^.nextElem;
     end;
    mElem:=mElem^.nextElem;
   end;{while mElem}
end;

procedure xyuvFromCell(thePoint:zNumber;theCellA:PCell;var indVel:zNumber);
var
  thisMag,sinCore,cosCore,rsq,oneOver2rsq,Xmn,Ymn: double;
  mElem: Pchain;
begin
  mElem:=theCellA^.dvInCell;
  while mElem<>nil do with mElem^.dvElem^.Vortex do begin
    {thisMag:=mElem^.dvElem^.vortex.mag;}
    {/// effect vorticity decay ///}
    thisMag:=vortDecay(mElem^.dvElem);
    Xmn:=thePoint.Re-pos.Re;
    Ymn:=thePoint.Im-pos.Im;
    rsq:=sqr(Xmn)+sqr(Ymn);
    oneOver2rsq:=0.5/(pi*rsq);
    if rsq>cutOffsq then begin
     indVel.Re:=indVel.Re + thisMag*Ymn*oneOver2rsq;
     indVel.Im:=indVel.Im - thisMag*Xmn*oneOver2rsq;
    end
    else begin
     {/// Rankine vortex model ///}
     indVel.Re:=indVel.Re + thisMag*Ymn*oneOver2piCSq;
     indVel.Im:=indVel.Im - thisMag*Xmn*oneOver2piCSq;

     {/// Chorin vortex model ///}
     {sinCore:=oneOver2piC*Ymn/sqrt(rsq);
     cosCore:=oneOver2piC*Xmn/sqrt(rsq);
     indVel.Re:=indVel.Re + thisMag*sinCore;
     indVel.Im:=indVel.Im - thisMag*cosCore;}
    end;
    mElem:=mElem^.nextElem;
   end;
end;

procedure xyuvToCell(thePoint:dvPtr;theCellA:PCell);
var
  thisMag,
  sinCore,cosCore,rsq,oneOver2rsq,Xmn,Ymn: double;
  mElem: Pchain;
begin
  mElem:=theCellA^.dvInCell;
  while mElem<>nil do with mElem^.dvElem^.Vortex do begin
    {thisMag:=thePoint^.Vortex.mag;}
    {/// effect vorticity decay ///}
    thisMag:=vortDecay(thePoint);

    Xmn:=pos.Re-thePoint^.Vortex.pos.Re;
    Ymn:=pos.Im-thePoint^.Vortex.pos.Im;
    rsq:=sqr(Xmn)+sqr(Ymn);
    oneOver2rsq:=0.5/(pi*rsq);
    if rsq>cutOffsq then begin
      vel.Re:=vel.Re + thisMag*Ymn*oneOver2rsq;
      vel.Im:=vel.Im - thisMag*Xmn*oneOver2rsq;
     end
    else begin
      {/// Rankine vortex model ///}
      vel.Re:=vel.Re + thisMag*Ymn*oneOver2piCSq;
      vel.Im:=vel.Im - thisMag*Xmn*oneOver2piCSq;

      {/// Chorin vortex model ///}
      {sinCore:=oneOver2piC*Ymn/sqrt(rsq);
      cosCore:=oneOver2piC*Xmn/sqrt(rsq);
      vel.Re:=vel.Re + thePoint^.Vortex.mag*sinCore;
      vel.Im:=vel.Im - thePoint^.Vortex.mag*cosCore;}
     end;
    mElem:=mElem^.nextElem;
   end;
end;

procedure twoCellGroup(GroupA,GroupB:PCell);
var
  curCtocA,curCRowA,curCtocB,curCRowB: PCell;
  xxx,yyy: integer;
begin
  curCRowA:=GroupA;
  while curCRowA<>nil do begin
   curCtocA:=curCRowA;
   while curCtocA<>nil do begin
    if curCtocA^.dvInCell<>nil then begin
     curCRowB:=GroupB;
     while curCRowB<>nil do begin
      curCtocB:=curCRowB;
      while curCtocB<>nil do begin
       if curCtocB^.dvInCell<>nil then begin
        {/// determine the number of terms required ///}
        xxx:=round((curCtocB^.Ctr.Re-curCtocA^.Ctr.Re)*GroupA^.PData^.InvWidth);
        yyy:=round((curCtocB^.Ctr.Im-curCtocA^.Ctr.Im)*GroupA^.PData^.InvHeight);
        QCtocMatxData(GroupA^.PData,XXX,YYY);
        if ZData^.orderTerm>ctcOrderLimit then begin
         if ((curCtocA^.Zdv>minDvPerCell) and (curCtocB^.Zdv>minDvPerCell)) and
            ((curCtocA^.nodeStart<>nil) and (curCtocB^.nodeStart<>nil)) then
           twoCellGroup(curCtocA^.NodeStart,curCtocB^.NodeStart)
         else begin
           UVbetweenCell(curCtocA,curCtocB);
         end;
        end
        else begin
          CtocGroupInteract(curCtocA,curCtocB,GroupA^.PData^.level);
        end;
       end;{/// if curCtocB^.dvInCtoc<>nil and curCtocB<>curCtocA ///}
       curCtocB:=curCtocB^.UpC;
      end;{while curCtocB}
      curCRowB:=curCRowB^.RtC;
     end;{while curCRowB}
    end;{if curCtocA^.dvIncell<>nil}
    curCtocA:=curCtocA^.UpC;
   end;
   curCRowA:=curCRowA^.RtC;
  end;
end;

[global]
procedure evaluateCtocUV(CellGroup:PCell);
var
  curCtocA,curCRowA,curCtocB,curCRowB: PCell;
  XXX,YYY: integer;
begin
  { curCRowA and curCRowB always point to the first cell of a row }
  curCRowA:=CellGroup;
  while curCRowA<>nil do begin
    { curCtocA is active mover, which goes through the whole domain and
      accumulates induced velocity from other cells. Once it gone through
      complete domain, it is set idle for that time step. Its effect on
      other cells is calculated as double-linked vortex pairs method. }
   curCtocA:=curCRowA;
   while curCtocA<>nil do begin
    if curCtocA^.dvInCell<>nil then begin
     {/// calculate induced velocity among dv within same cell ///}
     if (curCtocA^.Zdv>minDvPerCell) and (curCtocA^.nodeStart<>nil) then
      evaluateCtocUV(curCtocA^.NodeStart)
     else UVwithinCell(curCtocA);


     {/// calculate induced velocity among dv from other cells ///}
     curCRowB:=curCRowA; { start out at the same column }
     while curCRowB<>nil do begin
      {/// assign curCtocB; if curCRowB points to same row as
           curCRowB does, then assign it next cell of cellA   ///}
      if curCRowB<>curCRowA then curCtocB:=curCRowB
      else curCtocB:=curCtocA;
      while curCtocB<>nil do begin
       if (curCtocB^.dvInCell<>nil) and (curCtocB<>curCtocA) then begin
        {/// determine the number of terms required ///}
        XXX:=round((curCtocB^.Ctr.Re -
                    curCtocA^.Ctr.Re)*cellGroup^.PData^.InvWidth);
        YYY:=round((curCtocB^.Ctr.Im -
                    curCtocA^.Ctr.Im)*cellGroup^.PData^.InvHeight);
        QCtocMatxData(CellGroup^.PData,XXX,YYY);
        if ZData^.orderTerm>ctcOrderLimit then begin
         if ((curCtocA^.Zdv>minDvPerCell) and (curCtocB^.Zdv>minDvPerCell)) and
            ((curCtocA^.nodeStart<>nil) and (curCtocB^.nodeStart<>nil)) then
          twoCellGroup(curCtocA^.NodeStart,curCtocB^.NodeStart)
         else begin
           UVbetweenCell(curCtocA,curCtocB);
          end;
        end else begin
         CtocGroupInteract(curCtocA,curCtocB,CellGroup^.PData^.level);
        end
       end;{/// if curCtocB^.dvInCtoc<>nil and curCtocB<>curCtocA ///}
       curCtocB:=curCtocB^.UpC;
      end;{while curCtocB}
      curCRowB:=curCRowB^.RtC;
     end;{while curCRowB}
    end;{if curCtocA^.dvIncell<>nil}
    curCtocA:=curCtocA^.UpC;
   end;
   curCRowA:=curCRowA^.RtC;
  end;
end;

[global]
procedure CtocXYUV(GroupStart:PCell;aPt:zNumber;var inducedVel:zNumber);
{
  Calculate the induced velocity inducedVel at a given point aPt
  due to all the dv in a cell group GroupStart.

  CtocXYUV is to add calculated value on top of the velocity variable,
  reset variable to zero before calling this procedure
}
var
  curCtoc,curCRow: PCell;
begin
  curCRow:=GroupStart;
  while curCRow<>nil do begin
   curCtoc:=curCRow;
   while curCtoc<>nil do begin
    if curCtoc^.dvInCell<>nil then begin
     {/// calculate induced velocity among dv within same cell ///}
     GroupToPtData(curCtoc,aPt);
     if ZData^.orderTerm>ctcOrderLimit then
       if (curCtoc^.Zdv>minDvPerCell) and (curCtoc^.nodeStart<>nil) then
        CtocXYUV(curCtoc^.NodeStart,aPt,inducedVel)
       else xyuvFromCell(aPt,curCtoc,inducedVel)
     else
     CtocGroupToPt(curCtoc,inducedVel);
    end;
   curCtoc:=curCtoc^.UpC;
   end;
   curCRow:=curCRow^.RtC;
  end;
end;

procedure CellCtocXYUV(theCell:PCell;Point:zNumber;var inducedVel:zNumber);
begin
  {/// calculate induced velocity among dv within same cell ///}
  GroupToPtData(theCell,Point);
  if ZData^.orderTerm>ctcOrderLimit then
    if (theCell^.Zdv>minDvPerCell) and (theCell^.nodeStart<>nil) then
     CtocXYUV(theCell^.NodeStart,Point,inducedVel)
    else xyuvFromCell(point,theCell,inducedVel)
  else CtocGroupToPt(theCell,inducedVel);
end;

procedure dvRhsOneCell(theCell:PCell);
var
  thisMag:double;
  bdy,Sgt:integer;
  mElem: Pchain;
begin
  mElem:=theCell^.dvInCell;
  while mElem<>nil do begin
    {thisMag:=mElem^.dvElem^.vortex.mag;}

    {/// effect vorticity decay ///}
    thisMag:=vortDecay(mElem^.dvElem);
    getOneInduced(mElem^.dvElem);
    if mElem^.dvElem^.vortex.mag<>0.0 then begin
     for bdy:=1 to numActiveBbody do with bodyAry[bdy]^ do begin
       for sgt:=1 to mSeg do
        fullRhs[sgt]:=fullRhs[sgt]+dvRhs[sgt]*thisMag;
      end;
     sumActiveDv:=sumActiveDv+thisMag;
    end;
    {/// sum magitude of all active vortices ///}
   mElem:=mElem^.nextElem;
  end;{while}
end;

[global]
procedure RoundtheWorld(GroupStart:PCell);
var
  iSgt,bdy,i: integer;
  inVel,dump: zNumber;
  circul: double;
  curCtoc,curCRow: PCell;
begin
  curCRow:=GroupStart;
  while curCRow<>nil do begin
   curCtoc:=curCRow;
   while curCtoc<>nil do begin
    if curCtoc^.dvInCell<>nil then begin
     if curCtoc^.Clear=CellNotClear then begin
       if (curCtoc^.Zdv>minDvPerCell) and (curCtoc^.NodeStart<>nil) then
         RoundtheWorld(curCtoc^.NodeStart)
       else dvRhsOneCell(curCtoc);
      end
     else begin
       for bdy:=1 to numActiveBbody do with bodyAry[bdy]^ do begin
        {fillChar(adhoc,sizeOf(vectorM),0);}
        circul:=0.0;
        for i:=1 to mSeg do begin
         {/// CellCtocXYUV is accumulative so have to set inVel zero ///}
         inVel.Re:=0.0;inVel.Im:=0.0;
         CellCtocXYUV(curCtoc,seg[i]^.pvt,inVel);
         adhoc[i]:=sqr(inVel.Re)+sqr(inVel.Im);
         dvRhs[i]:=(inVel.Re*seg[i]^.cosine+inVel.Im*seg[i]^.sine);
         circul:=circul+dvRhs[i]*seg[i]^.ds;
        end;
        if abs(circul)>zeroLimit then begin
          iSgt:=minTagOfList(adhoc,mSeg);
          dvRhs[iSgt]:=-Circul/seg[iSgt]^.ds+dvRhs[iSgt];
         end;
        for i:=1 to mSeg do fullRhs[i]:=fullRhs[i]+dvRhs[i];
       end;
       sumActiveDv:=sumActiveDv+curCtoc^.ctoc^.Zm0;
      end;
    end;
    curCtoc:=curCtoc^.UpC;
   end;
   curCRow:=curCRow^.RtC;
  end;
end;

procedure CtocDvToGpUV(point:dvPtr;CellGroup:PCell);
{/// calculate the vel of all dv in a cellGroup induced by vortex at a
     given point.
///}
var
  curCtocA,curCRowA: PCell;
begin
  curCRowA:=CellGroup;
  while curCRowA<>nil do begin
   curCtocA:=curCRowA;
   while curCtocA<>nil do begin
    if curCtocA^.dvInCell<>nil then begin
     {/// calculate induced velocity among dv within same cell ///}
     dvToGpData(Point,curCtocA);
     if ZData^.orderTerm>ctcOrderLimit then begin
       if (curCtocA^.nodeStart<>nil) then
        CtocDvToGpUV(point,curCtocA^.NodeStart)
       else xyuvToCell(point,curCtocA);
      end
     else CtocPtToGroup(curCtocA,curCtocA^.PData^.level);
    end;
   curCtocA:=curCtocA^.UpC;
   end;
   curCRowA:=curCRowA^.RtC;
  end;
end;

procedure formMergMesh(baseNo:integer);
var
  McCol,McRow,iCol,jRow   : integer;
  tmpCell                 : TCell;
  curNodeA,CurNodeB,
  NodeA,NodeB,
  curMCRow,tempPtr        : PCell;
  xxSpan,yySpan,
  xx,yy,dmp,McLowX,McLowY : double;
begin
  {McWidth:=abs(LLFlowField.Re-URFlowField.Re)/(0.85*McCol);
  McHeight:=abs(LLFlowField.Im-URFlowField.Im)/(0.85*McRow);}

  if baseNo=0 then baseNo:=29; { safe guard division by zero }
  yySpan:=abs(LLFlowField.Im-URFlowField.Im);
  xxSpan:=abs(LLFlowField.Re-URFlowField.Re);
  dmp:=baseNo*yySpan/xxSpan;

  McWidth:=1.25*xxSpan/baseNo;
  McHeight:=1.25*yySpan/dmp;

  {// prepare square mesh //}
  if McWidth>McHeight then dmp:=McWidth
  else dmp:=McHeight;
  McWidth:=dmp;
  McHeight:=dmp;
  McCol:=round(xxSpan/dmp);
  McRow:=round(yySpan/dmp);
  McLowX:=LLFlowField.Re-0.5*(McCol*dmp-xxSpan);
  McLowY:=LLFlowField.Im-0.5*(McRow*dmp-yySpan);


  if MCStart<>nil then freeCellMemory(MCStart, MCEnd);

  {/// calculate coordinate of node points ///}
  { columns are arranged in ascending order of x-coordinate }
  for iCol:=1 to McCol do begin
   xx:=McLowX+(iCol-1)*McWidth;
   for jRow:=1 to McRow do begin
   { rows are arranged in ascending order of y-coordinate }
    yy:=McLowY+(jRow-1)*McHeight;
    getNewPCell(tempPtr);
    tempPtr^.Ctr.Re:=xx;
    tempPtr^.Ctr.Im:=yy;

    {/// if the list is empty, make the new record the only one \\\}
    if MCStart = nil then begin
     MCStart:=TempPtr;
     MCEnd:=TempPtr;
     curMCRow:=TempPtr;
    end
    else begin
     TempPtr^.DnC:=MCEnd;
     MCEnd^.UpC:=TempPtr;
     MCEnd:=TempPtr;
     if jRow=1 then begin
      tempPtr^.LfC:=curMCRow;
      curMCRow^.RtC:=tempPtr;
      curMCRow:=tempPtr;
      curMCRow^.DnC^.UpC:=nil;
      curMCRow^.DnC:=nil;
     end;
    end;
   end;{for jRow}
  end;{for iCol}

  {/// form cross-row pointer link between pointers ///}
  NodeA:=MCStart;
  NodeB:=NodeA^.RtC;
  while NodeB<>nil do begin
   CurNodeA:=NodeA^.UpC;
   CurNodeB:=NodeB^.UpC;
   while curNodeA<>nil do begin
    curNodeA^.RtC:=curNodeB;
    curNodeB^.LfC:=curNodeA;
    curNodeA:=curNodeA^.UpC;
    curNodeB:=curNodeB^.UpC;
   end;
   NodeA:=NodeB;
   NodeB:=NodeB^.RtC;
  end;
  if ReformMerg then ReformMerg:=false;
end; { Proc formMergCell }

procedure resetMergCell;
var
  curNode, curCRow: PCell;
  curPchain,tmpPchain: Pchain;
  i: integer;
begin
  curCRow:=MCStart;
  while curCRow<>nil do begin
    curNode:=curCRow;
    while curNode<>nil do with curNode^ do begin { do if cell is NOT empty }
      while dvInCell<>nil do begin { free previously allocated dv in cell }
        tmpPchain:=dvInCell^.nextElem;
        {dispose(dvInCell);}
        PchainBin(dvInCell);
        dvInCell:=tmpPchain;
       end;
     curNode:=UpC; { go on to next cell down the row }
     end;   { completed one row }
    curCRow:=curCRow^.RtC;
   end;
   {/// reset dvOutofRange record ///}
   while dvOffMerg<>nil do begin
     tmpPchain:=dvOffMerg^.nextElem;
     {dispose(dvOffMerg);}
     PchainBin(dvOffMerg);

     dvOffMerg:=tmpPchain;
    end;
end; {Proc. ResetCell}

[global]
procedure AssignMergCell;
var
  mergNo,xincrt,yincrt,i,loop     : integer;
  aRnd,xdv,ydv,
  WidthInv,HeightInv,
  xApart,yApart            : double;
  searchNode               : PCell;
  tempChain                : Tchain;
  tmpPtr                   : Pchain;
  curPtr                   : dvPtr;
  xDone,yDone,outOfDomain  : boolean;
begin
  if (istep mod mergReform)=0 then begin
   reformMerg:=true;
   repeat
    aRnd:=KnuthRan;
    if (aRnd<0.5) then mergNo:=31
    else mergNo:=29;
   until mergNo<>recentBase;
   recentBase:=mergNo;
  end;
  if (MCStart=Nil) or reformMerg then formMergMesh(recentBase);

  resetMergCell; { initialise cell for this new calculation }
  WidthInv:=1.0/McWidth;
  HeightInv:=1.0/McHeight;
  curPtr:=dvStart;
  while curPtr<>nil do begin
    xdv:=curPtr^.Vortex.pos.Re;
    ydv:=curPtr^.Vortex.pos.Im;
    xDone:=false;
    yDone:=false;
    {!!! allocate zMemory if nil !!!}
    searchNode:=MCStart;

    {!!! search for nearest lower-left node !!!}
    outOfDomain:=false;
    loop:=0;
    repeat
      xApart:=(xdv-searchNode^.Ctr.Re)*WidthInv;
      yApart:=(ydv-searchNode^.Ctr.Im)*HeightInv;
      xincrt:=round(xApart);
      yincrt:=round(yApart);
      if xApart<-0.5001 then begin
          for i:=-1 downto xincrt do if searchNode^.LfC<>nil then
          searchNode:=searchNode^.LfC;
        end
      else
       if xApart>0.5001 then begin
           for i:=1 to xincrt do if searchNode^.RtC<>nil then
           searchNode:=searchNode^.RtC;
         end
       else Xdone:=true;

      if yApart<-0.5001 then begin
         for i:=-1 downto yincrt do if searchNode^.DnC<>nil then
          searchNode:=searchNode^.DnC;
        end
      else
       if yApart>0.5001 then begin
          for i:=1 to yincrt do if searchNode^.UpC<>nil then
          searchNode:=searchNode^.UpC;
        end
       else yDone:=true;

      {/// avoid node searching from being trapped in infinite loop ///}
      if loop>=2 then begin
        yDone:=true;xDone:=true;
        outOfDomain:=true;
       end;
      loop:=loop+1;
    until xDone and yDone;

    tempChain.dvElem:=curPtr;
    tempChain.nextElem:=nil;

    getNewPchain(tmpPtr);
    tmpPtr^:=tempChain;

    if not outOfDomain then begin
      {/// append dv to element-list of searchNode ///}
      if searchNode^.dvInCell = nil then searchNode^.dvInCell:=tmpPtr
      else begin
        TmpPtr^.nextElem:=searchNode^.dvInCell;
        searchNode^.dvInCell:=TmpPtr;
       end;
     end
    else begin {if outofRange}
      { should dv fall out of active domain, add it to dvOutOfRange list }
      { store searchNode for as starting point in next searching }
      if dvOffMerg = nil then dvOffMerg:=tmpPtr
       else begin
         TmpPtr^.nextElem:=dvOffMerg;
         dvOffMerg:=TmpPtr;
        end;
     end;
   curPtr:=curPtr^.next;
  end; { while }
end; { AssignMergCell }

[global]
procedure QuitGridTTT;
var j:integer;
begin
  { free memory from pointers of cell associated data }
  if CtocStart<>nil then freeCellMemory(CtocStart,CtocEnd);
  if MCStart<>nil then freeCellMemory(MCStart,MCEnd);
  if SmpStart<>nil then freeSmpCell(smpStart,smpEnd);
  if ctcCoeffA<>nil then dispose(ctcCoeffA);
  if ctcCoeffB<>nil then dispose(ctcCoeffB);
  if ctcSumCoeff<>nil then dispose(ctcSumCoeff);
  for j:=0 to numSubLevel do
   if CGroup[j]<>nil then freeGroupData(CGroup[j]);
  if specialBox<>nil then dispose(specialBox);
end;

[global]
procedure initGridData;
var j:integer;
begin
  dvOffMerg:=nil;
  MCStart:=nil;
  MCEnd:=nil;
  CtocStart:=nil;
  CtocEnd:=nil;
  ReformCtoc:=false;
  ReformMerg:=false;
  recentBase:=29;
  ctcCoeffA:=nil;
  ctcCoeffB:=nil;
  ctcSumCoeff:=nil;
  specialBox:=nil;
  for j:=0 to numSubLevel do
   CGroup[j]:=nil;
end;

to begin do
  initGridData;
end.
