[inherit('vglobal')]
Module Manager(output);

Var
  CurNx,CurNy,CurNnose : integer;
  CurXDim,CurYDim,CurZDim,
  curXOrigin,curYOrigin,Beta : double;

procedure inputCircle(bdyTag:integer);
var
  fi,dfi,
  angm2,
  radius  : double;
  n       : integer;
begin
  {/// init seg[] ptr ///}
  with BodyAry[bdyTag]^ do begin
    mSeg:=CurNx;
    getGeoMemory(bdyTag);
    AeroFoil:=false;
    dfi:=2.0*pi/mSeg;
    angm2:=pi/mSeg;
    radius:=0.5*curXDim;
    for n:=1 to mSeg{+1} do begin
       fi:=(n-1)*dfi;
       seg[n]^.datPt.Re:=curXOrigin-radius*cos(fi);
       seg[n]^.datPt.Im:=curYOrigin+radius*sin(fi);
    end;
  end; {with}
end;

procedure inputRect(bdyTag:integer);
const limitCoeff = 0.15;
var
  coeff,propX,
  xstep,ystep,
  txform,fi,angm2    : double;
  n,NYYX,NYYXX       : integer;
  tmpXdata,tmpYdata  : vectorM;

procedure roundCorner;
var
  ii                 : integer;
  dmpPtA,dmpPtB,
  dmpPtC,dmpPtD      : zNumber;
begin
  dmpPtA.Re:= tmpXData[curNx+curNy]+
                0.75*(tmpXData[curNx+curNy+1]-tmpXData[curNx+curNy]);
  dmpPtA.Im:= tmpYData[curNx+curNy]+
                0.75*(tmpYData[curNx+curNy+1]-tmpYData[curNx+curNy]);
  dmpPtB.Re:= tmpXData[curNx+curNy+1]+
                0.25*(tmpXData[curNx+curNy+2]-tmpXData[curNx+curNy+1]);
  dmpPtB.Im:= tmpYData[curNx+curNy+1]+
                0.25*(tmpYData[curNx+curNy+2]-tmpYData[curNx+curNy+1]);

  dmpPtC.Re:= tmpXData[curNx+2*curNy]+
                0.75*(tmpXData[curNx+2*curNy+1]-tmpXData[curNx+2*curNy]);
  dmpPtC.Im:= tmpYData[curNx+2*curNy]+
                0.75*(tmpYData[curNx+2*curNy+1]-tmpYData[curNx+2*curNy]);
  dmpPtD.Re:= tmpXData[curNx+2*curNy+1]+
                0.25*(tmpXData[curNx+2*curNy+2]-tmpXData[curNx+2*curNy+1]);
  dmpPtD.Im:= tmpYData[curNx+2*curNy+1]+
                0.25*(tmpYData[curNx+2*curNy+2]-tmpYData[curNx+2*curNy+1]);

  BodyAry[bdyTag]^.mSeg:=BodyAry[bdyTag]^.mSeg+1;
  tmpXData[curNx+curNy+1]:=dmpPtA.Re;
  tmpYData[curNx+curNy+1]:=dmpPtA.Im;
    {/// shift point one segment backkward ///}
  for ii:=BodyAry[bdyTag]^.mSeg downto curNx+curNy+3 do begin
    tmpXData[ii]:=tmpXData[ii-1];
    tmpYData[ii]:=tmpYData[ii-1];
  end;
  tmpXData[curNx+curNy+2]:=dmpPtB.Re;
  tmpYData[curNx+curNy+2]:=dmpPtB.Im;

  BodyAry[bdyTag]^.mSeg:=BodyAry[bdyTag]^.mSeg+1;
  tmpXData[curNx+2*curNy+2]:=dmpPtC.Re;
  tmpYData[curNx+2*curNy+2]:=dmpPtC.Im;
    {/// shift point one segment backkward ///}
  for ii:=BodyAry[bdyTag]^.mSeg downto curNx+2*curNy+4 do begin
    tmpXData[ii]:=tmpXData[ii-1];
    tmpYData[ii]:=tmpYData[ii-1];
  end;
  tmpXData[curNx+2*curNy+3]:=dmpPtD.Re;
  tmpYData[curNx+2*curNy+3]:=dmpPtD.Im;
end;

begin

(*
            y axis
          |                  xDim
          |             _____________________
          |            |                    |
          |            |                    |
          |            |                    |   yDim
          |            |                    |
          |            |____________________|
          |            ^center
      ____|________________________________________
          |                                     x axis


*)
  {/// init seg[] ptr ///}
  with BodyAry[bdyTag]^ do begin
    mSeg:=2*(curNx+curNy);
    getGeoMemory(bdyTag);
    AeroFoil:=false;
    txform:=beta*pi/180;
    xstep:=curXDim/curNx;
    ystep:=curYDim/curNy;
    {/// input square on 1st quar., with vertice at origin ///}
    NYYX:=2*curNy+curNx+2;
    NYYXX:=2*(curNy+curNx)+2;
    coeff:=0.0;{0.05;}  {/// 0 : uniform segment length <coeff<0.15 ////}
    for n := 1 to curNy+1 do begin
       propX:=(n-1)/curNy;
       tmpxdata[n]:=0.0;                    {/// 1st side, x=0.0    ///}
       tmpxdata[NYYX-n]:=curXDim;    {/// 3rd side, x=length ///}
       tmpydata[n]:=(propX-coeff*sin(propX*2.0*pi))*curYDim;
       tmpydata[NYYX-n]:=tmpydata[n];
    end;

    for n:=2 to curNX do begin
       propX:=(n-1)/curNx;
       tmpYdata[curNY+n]:=curYDim;
       tmpYdata[NYYXX-n]:=0.0;
       tmpXdata[curNY+n]:=(propX-coeff*sin(propX*2.0*pi))*curXDim;
       tmpXdata[NYYXX-n]:=tmpXdata[curNY+n];
    end;
    {roundCorner;}

    {/// rotate square to required orientation, and move from origin to
       its body center ///}

    for n:=1 to mSeg do begin
       seg[n]^.datPt.Re:=tmpXdata[n]*cos(txform)+tmpYdata[n]*sin(txform)+curXOrigin;
       seg[n]^.datPt.Im:=tmpYdata[n]*cos(txform)-tmpXdata[n]*sin(txform)+curYOrigin;
    end;
  end;
end;

procedure inputEllipse(bdyTag:integer);
const
  limitCoeff = 0.06;
var
  n                     : integer;
  coeff,x2yRatio,
  prop,txform,fi,dfi : double;
  tmpXdata,tmpYdata     : vectorM;
begin
  with BodyAry[bdyTag]^ do begin
    mSeg:=CurNx;
    getGeoMemory(bdyTag);
    AeroFoil:=false;
    {elpratio:=curyDim/curxDim;}
    dfi:=2.0*pi/mSeg;
    txform:=beta*pi/180.0;
    {x2yRatio:=curXDim/curYDim;
    if x2yRatio>1 then
     coeff:=(x2yRatio-1.0)/x2yRatio*limitCoeff
    else if x2yRatio<1 then
     coeff:=-(1.0-x2yRatio)*limitCoeff
    else coeff:=0.0;}
    coeff:=0.0;
    {write( ' coeff  ellipse ',coeff);wait;}
    {\\\ generate body with center at origin \\\}
    for n := 1 to mSeg{+1} do begin
       prop:=(n-1)/mSeg;
       fi:=(prop+coeff*sin(prop*2.0*pi))*2.0*pi;
       {fi:=(n-1)*dfi;}
       tmpXdata[n]:= 0.0 - curXDim*0.5*cos(fi);
       tmpYdata[n]:= curYDim*0.5*sin(fi);
    end;
    {/// rotate body and move to its body center ///}
    for n:=1 to mSeg{+1} do begin
      seg[n]^.datPt.Re:=curXOrigin+tmpXdata[n]*cos(txform)+tmpYdata[n]*sin(txform);
      seg[n]^.datPt.Im:=curYOrigin+tmpYdata[n]*cos(txform)-tmpXdata[n]*sin(txform);
    end;
  end;{with}
end;

procedure inputWedge(bdyTag:integer);
var
  n                             : integer;
  txform,zDimRad,xTop,yTop      : double;
  tmpXdata,tmpYdata             : vectorM;
begin

(*         y-axis
        T
        |                   /\
        |          yDim   /    \
        |               /        \
        |             /  zDim      \
        |           /_)______________\
        |          ^        xDim
        |          center
      --|------------------------------------->
        |                                 x-axis

    Notes :  xDim, yDim : lengths, zDim : included angle
             center at front tip;

*)
  with BodyAry[bdyTag]^ do begin
    mSeg:=curNx+curNy+curNnose;
    getGeoMemory(bdyTag);
    AeroFoil:=false;
    txform:=beta*pi/180.0;
    zDimRad:=curZDim*pi/180.0;
    xTop:=curYDim*cos(zDimRad);
    yTop:=curYDim*sin(zDimRad);

    for n:=1 to curNY+1 do begin
       tmpXdata[n]:=xTop*(n-1)/curNY;
       tmpYdata[n]:=yTop*(n-1)/curNY;
    end;

    for n:=1 to curNnose do begin
       tmpXdata[curNY+1+n]:=n/curNnose*(curXDim-xTop)+xTop;
       tmpYdata[curNY+1+n]:=yTop-n/curNnose*yTop;
    end;

    for n:=1 to curNx do begin
       tmpXdata[curNY+curNnose+1+n]:=curXDim-n/curNX*curXDim;
       tmpYdata[curNY+curNnose+1+n]:=0.0;
    end;

    for n:=1 to mSeg{+1} do begin
       seg[n]^.datPt.Re:=curXOrigin+tmpXdata[n]*cos(txform)+tmpYdata[n]*sin(txform);
       seg[n]^.datPt.Im:=curYOrigin+tmpYdata[n]*cos(txform)-tmpXdata[n]*sin(txform);
    end;

  end;{with}
end;

procedure inputEPlate(bdyTag:integer);
var
  i                                   : integer;
  curFactor,xPoint,avgLength,upSideY,
  lowSideY,Rratio,xb,xstep,ystep,
  txform,fi,dfi,lastSeg,last2Seg      : double;
  tmpXdata,tmpYdata                   : vectorM;
  loopCounter                         : integer;
  satisfied                           : boolean;

(*

                        xDim
                 |-----------------------|
                 |    ___________________    _________
                    /                    |          |
            ______/______________________|______    | yDim
                 ^\                      |          |
    elliptic nose   \ ___________________|  ________|___



    Notes :  zDim defines ratio of axes of the elliptic nose;
             ^ center at nose tip;

*)
  procedure findNoseLength;
  var
    dAngle,UpX,UpY,tmpX,tmpY,Length : double;
    subk,angleSeg                   : integer;
  begin
    Length:=0.0;
    UpX:=0.0;UpY:=0.0;
    angleSeg:=2000;
    dAngle:=0.5*pi/angleSeg;
    for subk:=1 to AngleSeg do
     begin
       fi:=subk*dAngle;
       tmpX:=xb*(1.0-cos(fi));
       tmpY:=curYDim*0.5*sin(fi);
       Length:=Length+sqrt(sqr(tmpX-UpX)+sqr(tmpY-UpY));
       UpX:=tmpX;UpY:=tmpY;
     end;
    avgLength:=(Length+curXDim-xb)/curNx;
  end;

  procedure getEPdata(xWise: double);
  begin
    if xWise<xb then UpSideY:=0.5*curYDim*sqrt(1-sqr((xWise-xb)/xb))
    else if xWise<=curXDim then UpSideY:=curYDim*0.5;
    LowSideY:=-UpSideY;
  end;

  procedure findSectionLength(var movePoint : double);
  var
    targetL,dAngle,startX,
    dxStep,tmpX,tmpY,theLength : double;
    subk,AngleSeg              : integer;
    meetLength                 : boolean;
  begin
    meetLength:=false;
    theLength:=0.0;
    startX:=movePoint;
    getEPdata(movePoint);
    tmpX:=movePoint;
    tmpY:=upsideY;
    targetL:=(1.0-0.5*sqr((startX-xb)/curXDim))*curFactor*avgLength;
    dxStep:=0.002*curXDim/curNx;
    subk:=0;
    repeat
     movePoint:=startX+subk*dxStep;
     getEPdata(movePoint);
     theLength:=theLength+sqrt(sqr(movePoint-tmpX)+sqr(UpsideY-tmpY));
     tmpX:=movePoint;tmpY:=UpsideY;
     subk:=subk+1;
     meetLength:=abs((theLength-targetL)/targetL)*curNx<curXDim;
    until meetLength or (movePoint>=curXDim);
  end;

begin
  {/// init seg[i] ptr ///}
  with BodyAry[bdyTag]^ do
  begin
    mSeg:=2*curNx+curNy;
    getGeoMemory(bdyTag);
    AeroFoil:=false;
    xb:= curZDim*curYDim*0.5;
    xstep:= (curXDim-xb)/curNx;
    yStep:= curYDim/curNy;
    dfi:= 0.5*pi/curNx;
    txform:= beta*pi/180.0;
    {/// first generate body with center at origin ///}
    findNoseLength;
    curFactor:=1.0;
    loopCounter:=0;
    repeat
       satisfied:=false;
       loopCounter:=loopCounter+1;
      xPoint:=0.0;
      tmpXdata[1]:=0.0;
      tmpYdata[1]:=0.0;
      for i:=1 to curNx do begin
         findSectionLength(xPoint);
         if i=curNx then xPoint:=curXDim;
         getEPdata(xPoint);
         tmpXdata[i+1]:=xPoint;
         tmpYdata[i+1]:=UpSideY;
         tmpXdata[mSeg+1-i]:=xPoint;
         tmpYdata[mSeg+1-i]:=LowSideY;
       end;
      lastSeg:=sqrt( sqr(tmpXdata[curNx+1]-tmpXdata[curNx])+
                     sqr(tmpYdata[curNx+1]-tmpYdata[curNx]));
      last2Seg:=sqrt( sqr(tmpXdata[curNx]-tmpXdata[curNx-1])+
                      sqr(tmpYdata[curNx]-tmpYdata[curNx-1]));
      if (lastSeg<>0) and (last2Seg<>0) then begin
        if lastSeg>1.1*last2Seg then
          curFactor:= curFactor*(1+abs(lastSeg-last2Seg)/(avgLength*curNx))
        else if lastSeg<0.9*last2Seg then
          curFactor:= curFactor*(1-abs(lastSeg-last2Seg)/(avgLength*curNx))
        else satisfied:=true;
      end
      else curFactor:=curFactor*(1-1/CurNx);
    until satisfied or (loopCounter>10);

    for i:=1 to curNy do begin
       tmpXdata[curNx+1+i]:=curXDim;
       tmpYdata[curNx+1+i]:=curYDim*0.5-i*yStep;
    end;

    {/// rotate body and move to its body center ///}
    for i:=1 to mSeg do begin
       seg[i]^.datPt.Re:=curXOrigin+tmpXdata[i]*cos(txform)+tmpYdata[i]*sin(txform);
       seg[i]^.datPt.Im:=curYOrigin+tmpYdata[i]*cos(txform)-tmpXdata[i]*sin(txform);
    end;
  end;{with}
end;
procedure inputUserDefineShape(bdyTag:integer;fileName: TString);
var
  i                                   : integer;
  tmpXdata,tmpYdata                   : vectorM;
  fileRead                            : text;
  txform                              : double;
  closed                              : boolean;
begin
  reset(fileRead,fileName);

    closed:=false;
    with BodyAry[bdyTag]^ do begin
     getGeoMemory(bdyTag);
     AeroFoil:=false;
     readln(fileRead,tmpXdata[1],tmpYdata[1]);
     i:=2;
     while (not Eof(fileRead)) and (not closed) and (i<maxSeg) do begin
       readln(fileRead,tmpXdata[i],tmpYdata[i]);
       if (tmpXdata[i]=tmpXdata[1]) and (tmpYdata[i]=tmpYdata[1]) then begin
         closed:=true;
         mSeg:=i;
       end;
       i:=i+1;
     end;
     if not closed then begin
       mSeg:=i;
       tmpXdata[mSeg]:=tmpXdata[1];
       tmpYdata[mSeg]:=tmpYdata[1];
     end;

     {/// move and rotate body ///}
     txform:= beta*pi/180.0;
     for i:=1 to mSeg do begin
       seg[i]^.datPt.Re:=curXOrigin+tmpXdata[i]*cos(txform)+tmpYdata[i]*sin(txform);
       seg[i]^.datPt.Im:=curYOrigin+tmpYdata[i]*cos(txform)-tmpXdata[i]*sin(txform);
     end;
    end; { with }
    close(fileRead);

end;

procedure inputNACA4D(bdyTag:integer);
var
  i,loopCounter         : integer;
  lastSeg,last2Seg,
  curFactor,
  Avglength,
  Afm,Afp,Aft,
  txform,dfi,xPoint,
  yMean,mThick,mSlope,
  coeffA,coeffB,coeffC  : double;
  tmpXdata,tmpYdata     : vectorM;
  satisfied,dumpBol     : boolean;

  {** aerofoil data based on NACA 4-digit wing section
      the mean line is defined by the equation

      y = m/p^2 (2px - x^2)  forward of maximum ordinate

  and
      y = m/(1-p)^2 [(1-2p) + 2px - x^2 ] aft of maximum ordinate

     thickness distribution is given by

      yt = 5t (0.2969 sqrt(x) - 0.1281x - 0.3516x^2 + 0.2843x^3 - 0.1015x^4 )

   }

  procedure getNACAdata(xWise : double);
  var
    xSQR : double;
  begin
    xSQR:=sqr(xWise);
    if xWise<Afp then begin
       yMean:=coeffA*( 2.0*Afp*xWise - xSQR );
       mSlope:=2.0*coeffA*( Afp - xWise );
     end
    else begin
        yMean:=coeffB*( coeffC + 2.0*Afp*xWise - xSQR );
        mSlope:=2.0*coeffB*( Afp - xWise );
    end;
    mThick:=Aft*(0.29690*sqrt(xWise)-0.12810*xWise-0.35160*xSQR
                 +0.28430*xWise*xSQR-0.10150*xSQR*xSQR);
  end;

  procedure findCurveLength;
  var
    dAngle,
    UpX,UpY,
    LowX,LowY,
    CurUpX,CurUpY,
    CurLowX,CurLowY,
    UpLength,LowLength : double;
    subk,AngleSeg           : integer;
  begin
    UpLength:=0.0;
    LowLength:=0.0;
    UpX:=0.0;UpY:=0.0;
    LowX:=0.0;LowY:=0.0;
    AngleSeg:=2000;
    dAngle:=pi/AngleSeg;
    for subk:=1 to AngleSeg do begin
      xPoint:=0.5*(1-cos(subk*dAngle));
      getNACAdata(xPoint);
      CurUpX:=(xPoint - mThick*sin(mSlope));
      CurUpY:=(yMean + mThick*cos(mSlope));
      CurLowX:=(xPoint + mThick*sin(mSlope));
      CurLowY:=(yMean - mThick*cos(mSlope));
      UpLength:=UpLength+sqrt(sqr(CurUpX-UpX)+sqr(CurUpY-UpY));
      LowLength:=LowLength+sqrt(sqr(CurLowX-LowX)+sqr(CurLowY-LowY));
      UpX:=CurUpX;UpY:=CurUpY;
      LowX:=CurLowX;LowY:=CurLowY;
     end;
    avgLength:=0.5*(UpLength+LowLength)/curNx;
    {writeln(' Upper side length ',UpLength);
    writeln(' Lower side length ',LowLength);}
  end;

  procedure findSectionLength(var theXpoint:double);
  var
    targetL,dAngle,startX,UpX,UpY,LowX,LowY,
    CurUpX,CurUpY,CurLowX,CurLowY,UpLength,LowLength : double;
    subk,AngleSeg : integer;
    meetLength : boolean;
  begin
    meetLength:=false;
    UpLength:=0.0;
    LowLength:=0.0;
    startX:=theXpoint;
    targetL:=(1.375-3.0*sqr(startX-0.5))*curFactor*avgLength;
    getNACAdata(startX);
    UpX:=(startX - mThick*sin(mSlope));
    UpY:=(yMean + mThick*cos(mSlope));
    LowX:=(startX + mThick*sin(mSlope));
    LowY:=(yMean - mThick*cos(mSlope));
    dAngle:=pi/(150*curNx);
    subk:=1;
    repeat
     theXpoint:=startX + 0.5*(1-cos(subk*dAngle));
     getNACAdata(theXpoint);
     CurUpX:=(theXpoint - mThick*sin(mSlope));
     CurUpY:=(yMean + mThick*cos(mSlope));
     CurLowX:=(theXpoint + mThick*sin(mSlope));
     CurLowY:=(yMean - mThick*cos(mSlope));
     UpLength:=UpLength+sqrt(sqr(CurUpX-UpX)+sqr(CurUpY-UpY));
     LowLength:=LowLength+sqrt(sqr(CurLowX-LowX)+sqr(CurLowY-LowY));
     UpX:=CurUpX;UpY:=CurUpY;
     LowX:=CurLowX;LowY:=CurLowY;
     subk:=subk+1;
     meetLength:=abs(((Uplength+LowLength)*0.5-targetL)/targetL)*curNx<1.0;
    until meetLength or (theXpoint>=1.0);
  end;

  procedure simpleNACA;
  var ij:integer;
  begin
     dfi:=pi/curNx;
     tmpXdata[1]:=0.0;
     tmpYdata[1]:=0.0;
     for ij:=1 to curNx do
      begin
        xPoint:=0.5*(1-cos(ij*dfi));
        getNACAdata(xPoint);
        tmpXdata[ij+1]:=(xPoint - mThick*sin(mSlope))*curXDim;
        tmpYdata[ij+1]:=(yMean + mThick*cos(mSlope))*curXDim;

        tmpXdata[2*curNx+1-ij]:=(xPoint + mThick*sin(mSlope))*curXDim;
        tmpYdata[2*curNx+1-ij]:=(yMean - mThick*cos(mSlope))*curXDim;
      end;
  end;

begin
  {/// init seg[i] ptr ///}
  with BodyAry[bdyTag]^ do
  begin
   if deNumberNACA4D(curYDim,Afm,Afp,Aft) then
    begin
      mSeg:=2*curNx;
      getGeoMemory(bdyTag);
      AeroFoil:=true;
      SegTe:=curNx;
      Afm:=Afm*0.01;
      Afp:=Afp*0.1;
      Aft:=Aft*0.05;
      if Afp>0.0 then coeffA:=Afm/sqr(Afp)
      else coeffA:=0.0;
      if Afp<1.0 then coeffB:=Afm/sqr(1-Afp)
      else coeffB:=0.0;
      coeffC:=1-2.0*Afp;
      txform:=beta*pi/180.0;

      if true then
       begin
         simpleNACA;
         {geoForPlotting:=false;}
       end
      else
       begin
         findCurveLength;
         curFactor:=1.0;
         loopCounter:=0;
         repeat
           satisfied:=false;
	   loopCounter:=loopCounter+1;
           xPoint:=0.0;
           tmpXdata[1]:=0.0;
           tmpYdata[1]:=0.0;
           for i:=1 to curNx do
            begin
              findSectionLength(xPoint);
              if i=curNx then xPoint:=1.00;
              getNACAdata(xPoint);
              tmpXdata[i+1]:=(xPoint - mThick*sin(mSlope))*curXDim;
              tmpYdata[i+1]:=(yMean + mThick*cos(mSlope))*curXDim;

              tmpXdata[mSeg+1-i]:=(xPoint + mThick*sin(mSlope))*curXDim;
              tmpYdata[mSeg+1-i]:=(yMean - mThick*cos(mSlope))*curXDim;
            end;
           lastSeg:=sqrt( sqr(tmpXdata[curNx+1]-tmpXdata[curNx])+
                     sqr(tmpYdata[curNx+1]-tmpYdata[curNx]));
           last2Seg:=sqrt( sqr(tmpXdata[curNx]-tmpXdata[curNx-1])+
                     sqr(tmpYdata[curNx]-tmpYdata[curNx-1]));
           if (lastSeg<>0) and (last2Seg<>0) then
             begin
              if lastSeg>1.1*last2Seg then
               curFactor:= curFactor*(1+abs(lastSeg-last2Seg)
                           /(avgLength*curNx))
              else if lastSeg<0.9*last2Seg then
               curFactor:= curFactor*(1-abs(lastSeg-last2Seg)
                           /(avgLength*curNx))
              else satisfied:=true;
             end
           else curFactor:=curFactor*(1-1/CurNx);
         until satisfied or (loopCounter>10);
         if not satisfied then simpleNACA;
       end;

      {/// rotate body and move to its body center ///}
      for i:=1 to mSeg{+1} do
       begin
         seg[i]^.datPt.Re:= curXOrigin + tmpXdata[i]*cos(txform)
                                       + tmpYdata[i]*sin(txform);
         seg[i]^.datPt.Im:= curYOrigin + tmpYdata[i]*cos(txform)
                                       - tmpXdata[i]*sin(txform);
       end;

    end;{if}
  end;{with}
end;

procedure inputOneGeo(bdyTag:integer);
begin
  with ActiveBbody[bdyTag] do begin
    curXOrigin:= BodyCX;
    curYOrigin:= BodyCY;
    beta:= BodyTilt;
    curXDim:= xDim;
    curYDim:= yDim;
    curZDim:= zDim;
    curNx:= Nx;
    curNy:= Ny;
    curNnose:= Nnose;

    case shapeCode of

    0      : inputCircle(bdyTag);
    1      : inputEllipse(bdyTag);
    2      : inputRect(bdyTag);
    3      : inputEPlate(bdyTag);
    4      : inputWedge(bdyTag);
    5      : inputUserDefineShape(bdyTag,UserInputFile);
    NACA4D : inputNACA4D(bdyTag);

    end; {case}
   end; {with}
end;

procedure inputGeo;
var
  j,i,kSeg : integer;
begin {/// SOProc. InputGeo ///}
  { free existing memory before generating new data, including DynaMatx,
    globalRhs }
  freeBodyAryMemory;
  for j:=1 to numActiveBbody do begin
     {/// get memory for current body ///}
     New(BodyAry[j]);
     with BodyAry[j]^ do begin
       selfMatx:=nil;
       for i:=1 to maxMutual do mutlMatx[i]:=nil;
       PonBody.NumRecord:=0;
       for kSeg:=1 to maxSeg do begin
         PonBody.MaxPress[kSeg]:=0.0;
         PonBody.MinPress[kSeg]:=0.0;
         PonBody.AvgPress[kSeg]:=0.0;
        end;
       { temporary set 0.0, change later for modelling of body with
         prescribed circulation, consider carefully the vorticity
         conservation equation }
       prescribeCirc:=0.0;
       {prescribeCirc:=activeBbody[j].BoundCirc;}
       ctrlVolCirc:=0.0;
     end;{with}
     inputOneGeo(j);
   end;
  getGlobalRhsMemory; { generate global RHS list }
end; {Proc InputGeo}

  { calculate detail information of segment on center/radius of curvature,
    concave/convex, sharp corner ... }
procedure SegmentInfo;
var
  i,j : integer;
  nextSlope,lastSlope,nextCofSlope,
  lastCofSlope,thisCofSlope,inclAng,length,normDir: double;
  flat,sameCofSlope: boolean;
begin
  for j:=1 to numActiveBbody do with bodyAry[j]^ do begin
    for i:=1 to mSeg do begin
      thisCofSlope:=getCofSlope(j,i);

      if abs(thisCofSlope)>tooSmall then begin
        nextCofSlope:=getCofSlope(j,i+1);
        lastCofSlope:=getCofSlope(j,i-1);
        nextSlope:=getSlope2Seg(j,i,i+1);
        lastSlope:=getSlope2Seg(j,i-1,i);
        with seg[i]^ do begin
          if ((abs(thisCofSlope)-abs(nextCofSlope))<tooSmall) or
            ((abs(thisCofSlope)-abs(lastCofSlope))<tooSmall)
          then sameCofSlope:=true else sameCofSlope:=false;

          if (abs(nextSlope)<tooSmall) or (abs(lastSlope)<tooSmall) then
            flat:=true else flat:=false;

          if flat and sameCofSlope then isSharp:=true
          else isSharp:=false;

          if AeroFoil and ((i=SegTe-1) or (i=SegTe)
                      or   (i=SegTe+1) or (i=SegTe+2)) then isSharp:=true;
          {/// determine radius of curvature of segment ///}
          radOfCurv:=0.50*ds/sin(0.25*abs(thisCofSlope));

          {/// determine included angle of segment ///}
          inclAng:=0.500*abs(thisCofSlope);

          {/// determine x,y coordinate of the center of curvature ///}
          length:=radOfCurv*sin(0.50*(pi-inclAng));
          normDir:=slope+0.5*pi;
          CenRad.Re:=pvt.Re-length*cos(normDir);
          CenRad.Im:=pvt.Im-length*sin(normDir);

          {/// determine if segment is convex ///}
          if thisCofSlope<0.0 then begin
            isConvex:=true;
            CenRad.Re:=pvt.Re-length*cos(normDir);
            CenRad.Im:=pvt.Im-length*sin(normDir);
          end
          else begin
            isConvex:=false;
            CenRad.Re:=pvt.Re+length*cos(normDir);
            CenRad.Im:=pvt.Im+length*sin(normDir);
          end;
        end;{with}
      end
      else begin
        seg[i]^.IsSharp:=false;
        seg[i]^.radOfCurv:=0.0;
      end;
    end;{for i mSeg}
    {seg[mSeg+1]^:=seg[1]^;}
  end;{for j}
end;

procedure getXYspan;
var
  i,j               : integer;
  tempXmin,tempXmax,
  tempYmin,tempYmax : double;
begin
  tempXmin:=tooBig;
  tempXmax:=-tooBig;
  tempYmin:=tooBig;
  tempYmax:=-tooBig;
  for j:=1 to numActiveBbody do
  with bodyAry[j]^ do begin
    for i:=1 to mSeg do
    with seg[i]^ do begin
      if datPt.Re<tempXmin then tempXmin:=datPt.Re;
      if datPt.Re>tempXmax then tempXmax:=datPt.Re;
      if datPt.Im<tempYmin then tempYmin:=datPt.Im;
      if datPt.Im>tempYmax then tempYmax:=datPt.Im;
    end;
  end;
  Xspan:=abs(tempXmax-tempXmin);
  Yspan:=abs(tempYmax-tempYmin);
  BodyLowerX:=tempXmin;
  BodyLowerY:=tempYmin;
  LLFlowField.Re:=tempXmin-0.2*xSpan;
  LLFlowField.Im:=tempYmin-0.5*DomainY*ySpan;
  URFlowField.Re:=tempXmax+DomainX*xSpan;
  URFlowField.Im:=tempYmax+0.5*DomainY*ySpan;
end;

procedure getArea;
var
  xx1,yy1,areaWh,SlantRatio : double;
  p,k,bdy                   : integer;

function SectorArea(i,j:integer):double;
var
  xe1,ye1,xe2,ye2,pla,nga,aba      : double;
begin
  with bodyAry[bdy]^ do begin
    xe1:=seg[i]^.datPt.Re;
    ye1:=seg[i]^.datPt.Im;
    xe2:=seg[j]^.datPt.Re;
    ye2:=seg[j]^.datPt.Im;
  end;
  pla:=(xe1*ye2+xe2*yy1+xx1*ye1);
  nga:=(xe1*yy1+xx1*ye2+xe2*ye1);
  sectorArea:=0.5*abs(pla-nga);
end;

begin
  for bdy:=1 to numActiveBbody do with bodyAry[bdy]^ do begin
    xx1:=seg[1]^.datPt.Re;
    yy1:=seg[1]^.datPt.Im;
    areaWh:=0.0;
    for p:=2 to mSeg-1 do areaWh:=areaWh+SectorArea(p,p+1);
    slantRatio:=sqrt(areaWh*pi)/dsSum;
    if slantRatio<=0.48 then slant:=true else slant:=false;
  end;
end;

procedure BasicGeometry;
  { prepare the basic data of geometry for plotting geometry, BUT not
    sufficient to do any calculation on it.}
var
  abscos,sincos,t,x1,y1,x2,y2    : double;
  j,n,nTwo                       : integer;
begin
  inputGeo;   { input data points on bodies }
  {/// determine slopes, control points ///}
  for j:=1 to numActiveBbody do with bodyAry[j]^ do begin
    {curBody:=j;}
    x1:=seg[1]^.datPt.Re;
    y1:=seg[1]^.datPt.Im;

    dsSum:=0.0; { keep count of total sum of segment length }
    for n:=1 to mSeg do
    with seg[n]^ do begin
      {/// angles are always measured at +ccw, range from 0 to 2*pi ///}
      nTwo:=getSegmentTag(j,n+1);
      x2:=seg[nTwo]^.datPt.Re; y2:=seg[nTwo]^.datPt.Im;
      ds:=sqrt(sqr(x2-x1)+sqr(y2-y1));
      {InvDs:=1.0/ds;}
      sine:=(y2-y1)/ds;cosine:=(x2-x1)/ds;
      abscos:=abs(cosine);
      sincos:=sine*cosine;
      if abscos>tooSmall then t:=arctan(sine/cosine);
      if (abscos<tooSmall) or (abscos=tooSmall)
       then slope:=sine/abs(sine)*pi*0.50;
      if cosine>tooSmall then slope:=t;
      if cosine<(-tooSmall) then slope:=t-pi;

      {/// check angle for +ccw direction ///}
      {/// if slope[n]<(-tooSmall) then slope[n]:=2.0*pi+slope[n] ///}
      pvt.Re:=(x1+x2)*0.50;
      pvt.Im:=(y1+y2)*0.50;
      x1:=x2;y1:=y2;
      dsSum:=dsSum+ds;
     end;{for/with mSeg}

    {/// assign data for the (m+1)th segment ///}
    {seg[mSeg+1]^:=seg[1]^;}
    {/// calc various parameters ///}
    dsAvg:=dsSum/mSeg;
  end;{for/with numActiveBbody}
  getXYspan;
end;

function DetailGeometry:boolean;
  { perform the basic geometry data input, check if bodies overlapping
    each other and prepare detail geometry data for later calculation }
begin
  basicGeometry;
  if not BodyLapped then begin
    segmentInfo;
    getArea;
    DetailGeometry:=true;
    DataError:=noError;
   end
  else begin
    DetailGeometry:=false;
    DataError:=OverLapping;
   end;
end;

(*/////////////////////////////////////////////////////////////////////////
    procedure getFsRhs

    the projection of free stream onto each segment is constant throughout,
    this values are evaluated first and are stored in FsRhs.
 /////////////////////////////////////////////////////////////////////////*)

procedure getFsRhs;
var
  i,j          : integer;
begin
  for j:=1 to numActiveBbody do with bodyAry[j]^ do
   for i:=1 to mSeg do
    fsRhs[i]:=-(UxMean*seg[i]^.cosine+VyMean*seg[i]^.sine);
end;

[global]
function prepareData:boolean;
var
  segMean,dsMean : double;
  j              : integer;
begin {Proc prepareData}
  {/// convert angle of attack from degree to radian ///}
  alphaRad:=alpha*pi/180.0;
  {/// calculate x,y component of main stream velocities ///}
  if ReyNum<>0.0 then begin
    UxMean:=cos(alphaRad);
    VyMean:=sin(alphaRad);
    uxCurt:=UxMean;
    vyCurt:=VyMean;
   end
  else begin
    UxMean:=0.0;
    VyMean:=0.0;
    uxCurt:=UxMean;
    vyCurt:=VyMean;
   end;
  DataError:=noError;
  if detailGeometry then begin
     getfsRhs;
     {getBoundaryVel;}
  (*/////////////////////////////////////////////////////////////////////////
    determine time step size for iteration, the size chosen is that dv will
    past one segment in a single time step.  User can also input desired
    value via solution control window. Should user leave the value zero at
    input window, it is to be calculated as following code
  /////////////////////////////////////////////////////////////////////////*)

    dsMean:=0.0;      { mean segment length }
    segMean:=0.0;     { mean number of segment }
    for j:=1 to numActiveBbody do with bodyAry[j]^ do begin
      dsMean:=dsMean+dsAvg;
      segMean:=segMean+mSeg*1.0;
    end;
    dsMean:=dsMean/numActiveBbody;
    segMean:=segMean/numActiveBbody;

    (*////////////////////////////////////////////////////////////////////
      set the cut-off radius, and its values of cutoff^2, 2*pi*cutOff^2
     ////////////////////////////////////////////////////////////////////*)
    if ReyNum<>0 then begin
      cutOff:=dsMean/(2.0*ln(100));
      ShedOffset:=1.66*sqrt(abs(dt)/ReyNum);
      rwCoeffND:=sqrt(4.0*abs(dt)/ReyNum);
      bufferRad:=2.0/sqrt(ReyNum);
     end
    else begin
      cutOff:=0.25*dsMean/(pi*pi);
      ShedOffset:=cutOff;
      rwCoeffND:=0.0;
    end;

    cutOffSq:=sqr(cutOff);
    oneOver2CSq:=0.5/sqr(cutOff);
    oneOver2piCSq:=0.50/(pi*sqr(cutOff));
    oneOver2piC:=0.50/(pi*cutOff);
    Rcri:=2.0*cutOff;
    fluctFlow:=fluctLevel>0.0;
    doProbing:=((probeOff-probeOn)>0.0) and (probeOff>0.0);

    {/// set the boundary value of d.v. strength for vortice merging ///}
    if dt=0.0 then begin
       prepareData:=false;
       DataError:=invalidTimeStep;
    end
    else prepareData:=true;
  end
  else prepareData:=false;
end; {/// EOProc PrepareDatas ///}

function prepareForceFile:boolean;
begin
  writeForceName:=fileGrp+'.FCE';
  readForceName:=writeForceName;
  readForceList;
  prepareForceFile:=true;
end;

function PrepareRestart: boolean;
var
  reTime      : timeRec;
  dumpStr     : TString;
begin
  PrepareRestart:=false;
  ReadDvforRestart(pathToRestartFile,dvStart,dvEnd,ndv,reTime);
  dumpStr:=pathToReStartFile;
  splitString(dumpStr,fileGrp,isave); { find the file group of restart file}
  firstAccess:=true;
  if ndv > 0 then begin{/// check if all body in ActiveBbody[j] are contained in stream record ///}
    dt:=reTime.tStep;
    prepareData;
    elapsed:=reTime.tLast;
    prepareRestart:=true;
  end
  else DataError:=nullNdv;
end;

procedure RK4xxx(thePt:zNumber;dydx,hsize:double;var outPt:zNumber);
{  Runge-Kutta 4 step method :  based on the xy coordinate system;
   given initial point thePt and derviate dydx at thePt,
   the stepsize hsize. RK4xy advances the solution over
   an interval hsize and return as outPt.
}
const
  OneSixth = 0.1666666;
var
  medPt,velA : zNumber;
  insideBdy : integer;
  hh,h6,dyt,dym : double;
begin
  hh:=hsize*0.5;
  h6:=hsize*OneSixth;

  medPt.Re:=thePt.Re+hh;           { second step }
  medPt.Im:=thePt.Im+dydx*hh;
  xyuvQuick(medPt,velA,insideBdy,noCheck);
  dyt:=velA.Im/velA.Re;

  medPt.Im:=thePt.Im+dyt*hh;  { third step }
  xyuvQuick(medPt,velA,insideBdy,noCheck);
  dym:=velA.Im/velA.Re;

  outPt.Re:=thePt.Re+hsize;        { forth step }
  outPt.Im:=thePt.Im+dym*hsize;
  dym:=dym+dyt;
  xyuvQuick(outPt,velA,insideBdy,noCheck);
  dyt:=velA.Im/velA.Re;
  outPt.Im:=thePt.Im+h6*(dydx+dyt+2.0*(dym));
end;

procedure RK4yyy(thePt:zNumber;dxdy,ksize:double;var outPt:zNumber);
{  Runge-Kutta 4 step method : based on the Y-X coordinate system;
   given initial point thePt and derviate dxdy at thePt,
   the stepsize hsize. RK4xy advances the solution over
   an interval hsize and return as outPt.
}
const
  OneSixth = 0.1666666;
var
  medPt,velA : zNumber;
  insideBdy : integer;
  kk,k6,dxt,dxm : double;
begin
  kk:=ksize*0.5;
  k6:=ksize*OneSixth;
  medPt.Im:=thePt.Im+ksize;
  medPt.Re:=thePt.Re+dxdy*kk;
  xyuvQuick(medPt,velA,insideBdy,noCheck);
  dxt:=velA.Re/velA.Im;
  medPt.Re:=thePt.Re+dxt*kk;
  xyuvQuick(medPt,velA,insideBdy,noCheck);
  dxm:=velA.Re/velA.Im;
  outPt.Im:=thePt.Im+ksize;
  outPt.Re:=thePt.Re+dxm*ksize;
  dxm:=dxm+dxt;
  xyuvQuick(outPt,velA,insideBdy,noCheck);
  dxt:=velA.Re/velA.Im;
  outPt.Re:=thePt.Re+k6*(dxdy+dxt+2.0*(dxm));
end;

procedure preStrmRK4;
var
  nos,i,j,k                        : integer;
  diskFile                         : file of double;
  delt,xCoordR,yCoordR,
  codeR,NxR,NyR,NnoseR,NLineR,NptR : double;
  curPoint                         : PlinkPt;
  LinePtr                          : POneLine;

  procedure evaluateLine(fromPt:zNumber;incrmt:double);
  label 1;
  var
    reversed  : boolean;
    MedPt,velA : zNumber;
    insideBdy : integer;
    h,dydx,dxdy : double;
  begin
    with LinePtr^ do begin
     j:=1;
     theLine[1].Re:=fromPt.Re;
     theLine[1].Im:=fromPt.Im;
     writeln(fromPt.Re,' ',fromPt.Im);
     reversed:=false;
     while (j<maxStrmPt) and (theLine[j].Re<strmRlimit) do begin
      xyuvQuick(theLine[j],velA,insideBdy,yesCheck);
      if insideBdy=0 then begin
       dydx:=velA.Im/velA.Re;
       if abs(dydx)<2.0 then begin
        {evaluate value based on x/y coordinates }
        h:=velA.Re*incrmt;
        RK4xxx(theLine[j],dydx,h,theLine[j+1]);
       end
       else begin
        h:=velA.Im*incrmt;
        dydx:=velA.Re/velA.Im;
        RK4yyy(theLine[j],dydx,h,theLine[j+1]);
       end;
       {check for looping }
       if (j>10) and (abs(theLine[j].Re-fromPt.Re)<abs(h)) and
          (abs(theLine[j].Im-fromPt.Im)<abs(h)) then goto 1;
      end
      else goto 1;

     {if (j<100) and (not reversed) then begin
       j:=0;
       incrmt:=-1.0*delt;
       reversed:=true;
      end
      else goto 1;}
      j:=j+1;
     end;
    1: write(' number points : ',j,' ',theLine[j].Re);
    end
  end; {evaluateline}

begin
  rewrite(diskFile, WriteSLName);
  codeR:=numActiveBbody*1.0;
  write(diskFile,codeR);
  for j:=1 to numActiveBbody do with ActiveBbody[j] do begin
   codeR:=shapeCode*1.0;
   NxR:=Nx*1.0;
   NyR:=Ny*1.0;
   NnoseR:=Nnose*1.0;
   write(diskFile, CodeR, BodyCX, BodyCY, BodyTilt);
   write(diskFile, xDim, yDim, zDim);
   write(diskFile, NxR, NyR, NnoseR);
  end;

  NLineR:=NumStrmLines*1.0;
  write(diskFile, NLineR);
  delt:=strmTime;

  writeln(' R limit ',strmRlimit);
  writeln(' L limit ',strmLlimit);
  writeln(' dt      ',delt);

  curPoint:=StrmStartPts;
  new(LinePtr);
  with LinePtr^ do
   for nos:=1 to NumStrmlines do begin
    writeln(' doing line : ',nos);
    evaluateLine(curPoint^.ptCoord,delt);
    {/// store complete line to disk ///}
    NptR:=(j-1)*1.00;         { cast j-1 to real and save to disk }
    write(diskFile,NptR);
     for i:=1 to j-1 do begin
      xCoordR:=theLine[i].Re;
      yCoordR:=theLine[i].Im;
      write(diskFile,xCoordR,yCoordR);
     end;
      {/// next start point down the queue ///}
     curPoint:=curPoint^.downLink;
   end;
  close(diskFile);
  dispose(LinePtr);
end;

procedure ComputeStrmLine;
var
  reTime      : timeRec;
begin
  if VortexStrmLine then begin
   { reset grid system for new setup to be used with xyuvQuick }
   QuitGridTTT;
   initGridData;
   { read dv data from disk file }
   ReadDvforRestart(pathToDvFile,dvStart,dvEnd,ndv,reTime);
   firstAccess:=true;
   dt:=reTime.tStep;
   prepareData;
   { determine suitable domain size to cover all dv, override data
     determined in PrepareData }
   SearchHiLoDV(dvStart,LLFlowField,URFlowField);
   elapsed:=reTime.tLast;
   { generate surface vorticity with existing vorticity field }
   ViscousFlow;
   {findStagnation;}
   preStrmRK4;
   end
  else begin
   if prepareData then begin
    potentialFlow;
    {findStagnation;}
    PreStrmRK4;
    {PreStrmAB4;}
   end;
  end;
  freeChainOfPoint(StrmStartPts);
end;

[global]
function GoEasyRun:boolean;
begin
  GoEasyRun:=false;
  case calculateMode of

  calPotential :
   begin
    if prepareData then
     begin
      potentialFlow;
      GoEasyRun:=true;
     end;
   end;

  calViscousFlow :
   begin
    case vortexFlowMode of

      calFreshRun :
       begin
        if prepareData then
         {if prepareForceFile then}
          {if prepareTracerFile then}
           begin
            FreshStart;
            GoEasyRun:=true;
            DataError:=noError;
           end;
       end;

      calRestart :
       begin
          if prepareRestart then
           if prepareData then
            if prepareForceFile then
             {if prepareTracerFile then}
              begin
               reStart;
               GoEasyRun:=true;
               DataError:=noError;
              end;
       end;
      end; {case vortexFlowMode }
   end; { case of calVortexFlow }

   calStreamline : ComputeStrmLine;

  end; { case }
end;

end.

