[inherit('vglobal')]
Module Merging;

[global]
procedure DVMerging;
var
  i,k,NumSearch     : integer;
  thisSnuff,xSpanInv,
  CurMergeRad,
  x2bodyRatio           : double;
  AllowCombine,
  found1,found2         : boolean;
  bingo                 : array[1..2] of dvPtr;
  appro                 : array[1..2] of double;
  DVsign                : array[1..2] of boolean;
  idPtr                 : dvPtr;
  curNode, curRow       : PCell;
  tmpChain              : Pchain;

function item(input:integer):integer;
begin
 if input<1 then item:=1
  else
   if input>NumSearch then item:=NumSearch
    else item:=input;
end;  {/// EOFunc Item ///}

  {/////////////////////////////////////////////////////////////////////
    The putBingo procedure checks for 'being merged' of a found vortex.
    If the say vortex has been merged, its dv[i]^.mag:=0.0 was set,
    putBingo checks this value and prevents any merged vortex being
    merged again in the same merging process.

    bingo[1..2] of dvPtr records DV pointers of the two closest DV's
    appro[1..2] of double records distance away of the two closest DV's
    appro[1], and bingo[1] correspond to the CLOSEST DV
   \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\}

procedure putBingo(inputDV:dvPtr);
var
  dxx,dyy : double;
begin
  if (inputDV^.Vortex.mag<>0.0) and (inputDV<>idPtr) then
   begin
    if not found1 then
     begin
      dxx:=idPtr^.Vortex.pos.Re - inputDV^.Vortex.pos.Re;
      dyy:=idPtr^.Vortex.pos.Im - inputDV^.Vortex.pos.Im;
      found1:=true;
      bingo[1]:=inputDV;
      appro[1]:=sqrt(sqr(dxx)+sqr(dyy));
      DVsign[1]:=sameSign(idPtr^.Vortex.mag,inputDV^.Vortex.mag);
     end
    else if inputDV<>Bingo[1] then
     begin
      dxx:=idPtr^.Vortex.pos.Re - inputDV^.Vortex.pos.Re;
      dyy:=idPtr^.Vortex.pos.Im - inputDV^.Vortex.pos.Im;
      found2:=true;
      bingo[2]:=inputDV;
      appro[2]:=sqrt(sqr(dxx)+sqr(dyy));
      DVsign[2]:=sameSign(idPtr^.Vortex.mag,inputDV^.Vortex.mag);
     end;
   end;
  {/// by comparing distance, determine closest DV ///}
  {/// bingo[1] is always the closest DV           ///}
   if found2 and (appro[1]>appro[2]) then
    begin
     swapReal(appro[1],appro[2]);
     swapPtr(bingo[1],bingo[2]);
     swapBool(DVsign[1],DVsign[2]);
    end;
end;  {/// EOProc putBingo ///}

procedure mergeTwo (dvPtrOne,dvPtrTwo : dvPtr);
var
  MagOne,MagTwo,
  absT1,absT2,SumAbsTT,
  sumTT,
  InvAbsT1T2        : double;
  tmpDvRec          : TVortex;
  ssign             : boolean;

procedure combineDVs;
begin
  tmpDvRec.mag:=sumTT;
  tmpDvRec.tGen:=elapsed;
  tmpDvRec.pos.Re:=(dvPtrOne^.Vortex.pos.Re*absT1+dvPtrTwo^.Vortex.pos.Re*absT2)*InvAbsT1T2;
  tmpDvRec.pos.Im:=(dvPtrOne^.Vortex.pos.Im*absT1+dvPtrTwo^.Vortex.pos.Im*absT2)*InvAbsT1T2;
  tmpDvRec.nearBdy:=dvPtrOne^.Vortex.nearBdy;
  tmpDvRec.nearSeg:=dvPtrOne^.Vortex.nearSeg;
  dvAppend(dvStart,dvEnd,tmpDvRec,ndv);

  dvPtrOne^.Vortex.mag:=0.0;        {/// set dv[]^.mag:=0.0 to flag merged ///}
  dvPtrTwo^.Vortex.mag:=0.0;
end;

begin  {/// SOProc MergeTwo ///}
  MagOne:=VortDecay(dvPtrOne);
  MagTwo:=VortDecay(dvPtrTwo);
  ssign:=sameSign(dvPtrOne^.vortex.mag,dvPtrTwo^.Vortex.mag);
  absT1:=abs(MagOne);
  absT2:=abs(MagTwo);
  SumAbsTT:=absT1+absT2;
  sumTT:=MagOne+MagTwo;
  InvAbsT1T2:=1.0/(absT1+absT2);
  if not sSign then combineDVs
   else if allowCombine then combineDVs;
end; {/// EOProc MergeTwo ///}

procedure mergeThree (dvPtrOne,dvPtrTwo,dvPtrTee : dvPtr);
var
  MagOne,MagTwo,MagTee,
  absT1,absT2,absT3,InvAbsSum : double;
  tmpDvRec          : TVortex;
begin  {/// SOProc MergeTwo ///}
  {write(' M3 ');}
  MagOne:=VortDecay(dvPtrOne);
  MagTwo:=VortDecay(dvPtrTwo);
  MagTee:=VortDecay(dvPtrTee);
  {ssign:=sameSign(dvPtrOne^.vortex.mag,dvPtrTwo^.Vortex.mag);}
  absT1:=abs(MagOne);
  absT2:=abs(MagTwo);
  absT3:=abs(MagTee);
  InvAbsSum:=1.0/(absT1+absT2+absT3);
  with tmpDvRec do begin
   mag:=MagOne+MagTwo+MagTee;
   tGen:=elapsed;
   pos.Re:=( dvPtrOne^.Vortex.pos.Re*absT1+dvPtrTwo^.Vortex.pos.Re*absT2+
             dvPtrTee^.Vortex.pos.Re*absT3)*InvAbsSum;
   pos.Im:=( dvPtrOne^.Vortex.pos.Im*absT1+dvPtrTwo^.Vortex.pos.Im*absT2+
             dvPtrTee^.Vortex.pos.Im*absT3)*InvAbsSum;
   nearBdy:=dvPtrOne^.Vortex.nearBdy;
   nearSeg:=dvPtrOne^.Vortex.nearSeg;
  end;
  dvAppend(dvStart,dvEnd,tmpDvRec,ndv);

  dvPtrOne^.Vortex.mag:=0.0; {/// set dv[]^.mag:=0.0 to flag having merged ///}
  dvPtrTwo^.Vortex.mag:=0.0;
  dvPtrTee^.Vortex.mag:=0.0;
end; {/// EOProc MergeTwo ///}

procedure MergeInAction;
var
  offset,onX,onY   : integer;

procedure xTOy;
var
  j,xup,xdn,yup,ydn : integer;
begin
  j:=1;
  xup:=item(onX+offset);
  xdn:=item(onX-offset);
  repeat
   yup:=item(onY+j);
   ydn:=item(onY-j);
   if (xOrder^.theList[xup]=yOrder^.theList[yup]) or
      (xOrder^.theList[xup]=yOrder^.theList[ydn]) then
    putBingo(xOrder^.theList[xup]);
   if (xOrder^.theList[xdn]=yOrder^.theList[yup]) or
      (xOrder^.theList[xdn]=yOrder^.theList[ydn]) then
    putBingo(xOrder^.theList[xdn]);
   j:=j+1
  until found2 or (j>offset);
end;

procedure yTOx;
var
  j,xup,xdn,yup,ydn : integer;
begin
  j:=1;
  yup:=item(onY+offset);
  ydn:=item(onY-offset);
  repeat
   xup:=item(onX+j);
   xdn:=item(onX-j);
   if (yOrder^.theList[yup]=xOrder^.theList[xup]) or
      (yOrder^.theList[yup]=xOrder^.theList[xdn]) then
    putBingo(yOrder^.theList[yup]);
   if (yOrder^.theList[ydn]=xOrder^.theList[xup]) or
      (yOrder^.theList[ydn]=xOrder^.theList[xdn]) then
    putBingo(yOrder^.theList[ydn]);
   j:=j+1
  until found2 or (j>=offset);
end;

begin
  for onX:=1 to NumSearch-2 do
  if ValidDV(xOrder^.theList[onX]) then begin
    {/// find closest DV from every each other ///}
    found1:=false;found2:=false;
    idPtr:=xOrder^.theList[onX];
    {/// compare pointers on X, Y list to locate the same DV on yPtr list ///}
    onY:=1;
    while yOrder^.theList[onY]<>idPtr do onY:=onY+1;
    offset:=1;
    xTOy;
    while (not found2) and (offset<NumSearch) do begin
      offset:=offset+1;
      xTOy;
      yTOx;
    end;
    {/// determine Rcri regarding x-position of DV ///}
    x2bodyRatio:=(idPtr^.Vortex.pos.Re-bodyLowerX)*xSpanInv;

    if x2bodyRatio<10.0 then begin
      AllowCombine:=false;
      CurMergeRad:=Rcri;
      {CurMergeRad:=Rcri*(0.02*x2bodyRatio+0.8);}
      {CurMagLimit:=magCri;}
    end
    else begin
      AllowCombine:=false;
      CurMergeRad:=Rcri;
      {CurMagLimit:=((x2bodyRatio-10.0)/25.0+1)*magCri;}
    end;

    {if (not DVsign[1]) and (appro[1]<CurMergeRad) then
      mergeTwo(idPtr,bingo[1])
    else if found2 and (not DVsign[2]) and (appro[2]<CurMergeRad) then
      mergeTwo(idPtr,bingo[2])
    else if (appro[1]<CurMergeRad) then
      mergeTwo(idPtr,bingo[1])
    else if found2 and (appro[2]<CurMergeRad) then
      mergeTwo(idPtr,bingo[2]);}

    if found2 then begin
     if ((not DVsign[1]) and (appro[1]<CurMergeRad)) then
      if ((not DVsign[2]) and (appro[2]<CurMergeRad)) then
       mergeThree(idPtr,bingo[1],bingo[2])
      else mergeTwo(idPtr,bingo[1])
     else
      if ((not DVsign[2]) and (appro[2]<CurMergeRad)) then
       mergeTwo(idPtr,bingo[2])
      else if (appro[1]<CurMergeRad) then mergeTwo(idPtr,bingo[1])
    end
    else
     if (appro[1]<CurMergeRad) then mergeTwo(idPtr,bingo[1]);

  end
  else begin
    thisSnuff:=thisSnuff+VortDecay(xOrder^.theList[onX]);{^.Vortex.mag;}
    xOrder^.theList[onX]^.Vortex.mag:=0.0;
  end;
end;

begin    {/// SOProc DVmerging ///}
  AssignMergCell;
  thisSnuff:=0.0;        { initiate strength of snuffed dv in this process }
  xSpanInv:=1/xSpan;

  curRow:=MCStart;
  while curRow<>nil do begin
    curNode:=CurRow;
    while curNode<>nil do begin
      if curNode^.dvInCell<>nil then begin
        getCellXYlist(curNode^.dvInCell,NumSearch);
        if NumSearch >4 then begin
          xQsort(xOrder^.theList,1,NumSearch);
          yQsort(yOrder^.theList,1,NumSearch);
          MergeInAction;
         end;
       end;
       curNode:=curNode^.UpC;
     end;
    curRow:=curRow^.RtC;
   end;

  if dvOffMerg<>nil then begin
   getCellXYlist(dvOffMerg,NumSearch);
   if NumSearch>4 then begin
    xQsort(xOrder^.theList,1,NumSearch);
    yQsort(yOrder^.theList,1,NumSearch);
    MergeInAction;
   end;

   tmpChain:=dvOffMerg;
   while tmpChain<>nil do begin
    if not ValidDV(tmpChain^.dvElem) then begin
     thisSnuff:=thisSnuff+VortDecay(tmpChain^.dvElem);{^.Vortex.Mag;}
     tmpChain^.dvElem^.Vortex.Mag:=0.0;
    end;
    tmpChain:=tmpChain^.nextElem;
   end;

   end;
   snuff:=snuff+thisSnuff;
   dvPack(dvStart,dvEnd,allGone);
end;  {/// EOProc DVMerging ///}
end.

