[inherit('vglobal')]
module Ptvortex(output);

var
  goKutta    : boolean;

procedure bodystm( XYpt:zNumber; var stmvalue:double; var insideBdy:integer;
                    doCheck:boolean);
label 99;
const closeSurface=5.0e-6;
var
  i,j : integer;
  tmpStm,
  Xjm,Yjm,r,oneOver2Rsq,rsq,segVortex : double;

procedure xysubelem;
var
  nextTag,isub,nsubs            : integer;
  xo,yo,xn,yn,xsub,ysub,acstm : double;
begin
  acstm:=0.0;
  with bodyAry[j]^ do begin
   if r>tooSmall then begin
    nsubs:=24*(1+round(2*seg[i]^.ds/r));
    if abs(nsubs)>200 then nsubs:=200
   end
   else nsubs:=200;
  nextTag:=getSegmentTag(j,i+1);
  xo:=seg[i]^.datPt.Re;
  yo:=seg[i]^.datPt.Im;
  xn:=seg[nextTag]^.datPt.Re;
  yn:=seg[nextTag]^.datPt.Im;
  for isub:=1 to nsubs do begin
   xsub:=xo+(xn-xo)/nsubs*(isub-0.5);
   ysub:=yo+(yn-yo)/nsubs*(isub-0.5);
   Xjm:=XYpt.Re-xsub;
   Yjm:=XYpt.Im-ysub;
   rsq:=sqr(Xjm)+sqr(Yjm);
   if rsq>closeSurface then acstm:=acstm+segVortex*ln(sqrt(rsq));
  end;
  end;{with}
  tmpStm:=tmpStm+acstm/nsubs;
end; {subelement}

begin
  insideBdy:=0;
  tmpStm:=0.0;
  for j:=1 to numActiveBbody do with bodyAry[j]^ do begin
   for i:=1 to mSeg do with seg[i]^ do begin
    segVortex:=survor[i]*ds;
    Xjm:=XYpt.Re-pvt.Re;
    Yjm:=XYpt.Im-pvt.Im;
    rsq:=sqr(Xjm)+sqr(Yjm);
    r:=sqrt(rsq);
    if (r/ds>2.0) then begin { r/ds changed from 1.0 to 1.5 for trial }
     tmpStm:=tmpStm+segVortex*ln(r);
    end
    else begin
     XYsubelem;
     {/// check if point XYpt lies within body j ///}
     if doCheck and PtInOneBody(XYpt,j,noCheck) then begin
      insideBdy:=j;
      goto 99;
     end;
    end;
   end; {for/with}
  end;
  stmValue:=stmValue+0.5*tmpStm/pi;
99:end;

Procedure xyStmSum(Point:zNumber;var inStmfn:double);
var
  theMag,Xjm,Yjm,
  tmpStm,rsq  : double;
  curPtr           : dvPtr;
begin {/// SOProc xyuv ///}
  tmpStm:=0.0;
  curPtr:=dvStart;
  while curPtr<>nil do begin
   if curPtr^.vortex.nearBdy<>0 then theMag:=curPtr^.vortex.mag
   else theMag:=vortDecay(curPtr);
   Xjm:=Point.Re-curPtr^.vortex.pos.Re;
   Yjm:=Point.Im-curPtr^.vortex.pos.Im;
   rsq:=sqr(Xjm)+sqr(Yjm);
   if rsq>=cutOffsq then tmpStm:=tmpStm+theMag*ln(sqrt(rsq));
   curPtr:=curPtr^.next;
  end; {while/with}
  inStmfn:=inStmfn + tmpStm*0.5/pi;
end; {/// Proc xyuvSum ///}

[global]
procedure xystm( xypt:zNumber; var stmValue:double;
                var insideBdy:integer; doCheck:boolean);
begin
  stmValue:=uxCurt*xypt.Im-vyCurt*xypt.Re;
  bodystm(xypt,stmValue,insideBdy,doCheck);
  xyStmSum(xypt,stmValue);
end; {/// Proc xyuv ///}

procedure bodyxyuv( XYpt:zNumber; var indVel:zNumber; var insideBdy:integer;
                    doCheck:boolean);
label 99;
const closeSurface=5.0e-6;
var
  SelfCorr,tmpVel : zNumber;
  i,j : integer;
  Xjm,Yjm,r,oneOver2Rsq,rsq,segVortex : double;

procedure xysubelem;
var
  nextTag,isub,nsubs            : integer;
  xo,yo,xn,yn,xsub,ysub,acu,acv : double;
begin
  acu:=0.0;acv:=0.0;
  with bodyAry[j]^ do begin
   if r>tooSmall then begin
    nsubs:=24*(1+round(2*seg[i]^.ds/r));
    if abs(nsubs)>200 then nsubs:=200
   end
   else nsubs:=200;
  nextTag:=getSegmentTag(j,i+1);
  xo:=seg[i]^.datPt.Re;
  yo:=seg[i]^.datPt.Im;
  xn:=seg[nextTag]^.datPt.Re;
  yn:=seg[nextTag]^.datPt.Im;
  for isub:=1 to nsubs do begin
   xsub:=xo+(xn-xo)/nsubs*(isub-0.5);
   ysub:=yo+(yn-yo)/nsubs*(isub-0.5);
   Xjm:=XYpt.Re-xsub;
   Yjm:=XYpt.Im-ysub;
   rsq:=sqr(Xjm)+sqr(Yjm);
   if rsq>closeSurface then begin
    oneOver2Rsq:=0.5/rsq;
    acu:=acu+segVortex*Yjm*oneOver2Rsq;
    acv:=acv-segVortex*Xjm*oneOver2Rsq;
   end
   else begin
    selfCorr.Re:=segVortex/seg[i]^.ds*seg[i]^.selfCoup*cos(seg[i]^.slope);
    selfCorr.Im:=segVortex/seg[i]^.ds*seg[i]^.selfCoup*sin(seg[i]^.slope);
   end;
  end;
  end;{with}
  tmpVel.Re:=tmpVel.Re+acu/nsubs;
  tmpVel.Im:=tmpVel.Im+acv/nsubs;
end; {subelement}

begin {bodyxyuv}
  insideBdy:=0;
  tmpVel.Re:=0.0; tmpVel.Im:=0.0;
  selfCorr.Re:=0.0; selfCorr.Im:=0.0;
  for j:=1 to numActiveBbody do with bodyAry[j]^ do begin
   for i:=1 to mSeg do with seg[i]^ do begin
    segVortex:=survor[i]*ds;
    Xjm:=XYpt.Re-pvt.Re;
    Yjm:=XYpt.Im-pvt.Im;
    rsq:=sqr(Xjm)+sqr(Yjm);
    r:=sqrt(rsq);
    if (r/ds>2.0) then begin { r/ds changed from 1.0 to 1.5 for trial }
     oneOver2Rsq:=0.5/rsq;
     tmpVel.Re:=tmpVel.Re+segVortex*Yjm*oneOver2Rsq;
     tmpVel.Im:=tmpVel.Im-segVortex*Xjm*oneOver2Rsq;
    end
    else begin
     XYsubelem;
     {/// check if point XYpt lies within body j ///}
     if doCheck and PtInOneBody(XYpt,j,noCheck) then begin
      insideBdy:=j;
      goto 99;
     end;
    end;
   end; {for/with}
  end;
  indVel.Re:=indVel.Re+tmpVel.Re/pi+selfCorr.Re;
  indVel.Im:=indVel.Im+tmpVel.Im/pi+selfCorr.Im;
99:end;

[global]
procedure xyuvQuick( XYpt:zNumber; var indVel:zNumber;
                     var insideBdy: integer; doCheck:boolean);
begin {/// SOProc xyuv ///}
  {/// assign main-stream velocity to upt, vpt ///}
  {/// evaluate induced velocity by bluff body ///}
  indVel.Re:=uxCurt;
  indVel.Im:=vyCurt;
  bodyxyuv(XYpt,indVel,insideBdy,doCheck);
  {/// add result to upt, vpt ///}
  if ndv>modeChangeOver then begin
   if (CtoCStart=nil) or ReformCtoc then EasyAssign;
   CtocXYUV(CtocStart,XYpt,indVel)
  end
  else xyuvSum(xypt,indVel);
end; {/// Proc xyuv ///}

procedure storeVel;
var
  curPtr : dvPtr;
begin
  curPtr:=dvStart;
  while curPtr<>nil do with curPtr^ do begin
    link^.veli.Re:=link^.veli.Re+vortex.vel.Re;
    link^.veli.Im:=link^.veli.Im+vortex.vel.Im;
    curPtr:=curPtr^.next;
  end;
end;

procedure getDvRhs;
var
  theMag: double;
  curPtr  : dvPtr;
  bdy,asgn : integer;
begin
  for bdy:=1 to numActiveBbody do with bodyAry[bdy]^ do
   for asgn:=1 to mSeg do fullRhs[asgn]:=0.0;

  sumActiveDv:=0.0;

  (*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
     Evaluate influence due to active discrete vortices. Should one d.v be
     found drifted inside body, it is either relocated or absorbed depending
     on the current control mode.

     If 'Absorbed' is active, absorded dv are to be re-generated in the
     next time step. In the same time step as they were absorbed, they are
     temporary attached to the vorticity sheet of the corresponding 'closest'
     segment and become bound circulation of that body.
   \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*)

  curPtr:=dvStart;
  while curPtr<>nil do begin
    getOneInduced(curPtr);
    if curPtr^.vortex.mag<>0.0 then begin
      if curPtr^.vortex.nearBdy<>0 then theMag:=curPtr^.vortex.mag
      else theMag:=vortDecay(curPtr);
      {theMag:=curPtr^.vortex.mag;}
      for bdy:=1 to numActiveBbody do with bodyAry[bdy]^ do
       for asgn:=1 to mSeg do
        fullRhs[asgn]:=fullRhs[asgn]+dvRhs[asgn]*theMag;
      {/// sum magitude of all active vortices ///}
      sumActiveDv:=sumActiveDv+theMag;
     end;
   curPtr:=curPtr^.next;
  end;{while}
end;  {/// EOProc getDvRhs ///}

  {/////////////////////////////////////////////////////////////////////
   Procedure updateVortexSheet

   evaluates the complete RHS vector taking into account of free stream,
   vortex cloud, and snuff out vorticity strength.

   Vector is stored as fullRhs.

   caller Proc. xconvect
   /////////////////////////////////////////////////////////////////////}

procedure updateVortexSheet;
var
  i,j: integer;
  curPtr: dvPtr;
  tmpVector: PVector;
begin
  getDvRhs;
  totalBoundCirc:=0.0;
  tmpVector:=globalRhs;
  for j:=1 to numActiveBbody do with bodyAry[j]^ do
   for i:=1 to mSeg do begin
     tmpVector^.realNum:=fsRhs[i]-fullRhs[i];
     tmpVector:=tmpVector^.nextNum;
   end;
  solution;
end;

procedure getPotFlowRhs;
var
  sumbb : double;
  i,j: integer;
  tmpVector: PVector;
begin
  for j:=1 to numActiveBbody do with bodyAry[j]^ do
   for i:=1 to mSeg do fullRhs[i]:=0.0;
  sumbb:=0.0;
  sumActiveDv:=0.0;        { no active dv in potential flow analysis }
  totalBoundCirc:=0.0;
  for j:=1 to numActiveBbody do with bodyAry[j]^ do
   totalBoundCirc:=totalBoundCirc+prescribeCirc;
  tmpVector:=globalRhs;
  for j:=1 to numActiveBbody do with bodyAry[j]^ do begin
   for i:=1 to mSeg do begin
     tmpVector^.realNum:=fsRhs[i]+prescribeCirc;{+solidVel[i];}
     tmpVector:=tmpVector^.nextNum;
   end;
  end;
end;

  {////////////////////////////////////////////////////////////////////////
    procedure RandomWalk

    apply random walk to simulate vortex diffusion of real fluid:

       radial shift  := sqrt(4.0*enul*dt*ln(1.0/prand));
                        sqrt(4.0*dt/Re*ln(1/prand));

       angular shift := 2*pi*qrand;

    where prand, qrand are random numbers (0,1)
  ///////////////////////////////////////////////////////////////////////}

[global]
procedure RandomWalk;
var
  curPtr: dvPtr;
  randomAng,prand,qrand,dr,dx,dy: double;
begin
  curPtr:=dvStart;
  while curPtr<>nil do with curPtr^.Vortex do begin
   repeat prand:=KnuthRan; until (prand<>0.0);
   qrand:=KnuthRan;
   randomAng:=2.0*qrand*pi;
   dr:=rwCoeffND*sqrt(ln(1.0/prand));
   pos.Re :=pos.Re+dr*cos(randomAng);
   pos.Im:=pos.Im+dr*sin(randomAng);
   curPtr:=curPtr^.next;
  end; {while/with}
end;  {/// Proc randomWalk ///}

[global]
procedure ShedDv;
var
  i,bdy             : integer;
  tmpDv             : Tvortex;
  zz,
  outNormal,netGamma,
  xDelta,yDelta,
  xBase,yBase       : double;
  qvar,sx,sxx       : double;
  kConst            : double;

procedure shedThree;
begin
  with bodyAry[bdy]^ do begin
   tmpDv.nearSeg:=i;
   tmpDv.mag:=survor[i]*seg[i]^.ds/3.00;
   outNormal:=seg[i]^.slope+0.5*pi;
   xDelta:=seg[i]^.ds*0.333*cos(seg[i]^.slope);
   yDelta:=seg[i]^.ds*0.333*sin(seg[i]^.slope);

   xBase:=seg[i]^.pvt.Re-xDelta;
   yBase:=seg[i]^.pvt.Im-yDelta;
   tmpDv.pos.Re:=xBase+kConst*cos(outNormal);
   tmpDv.pos.Im:=yBase+kConst*sin(outNormal);
   dvAppend(dvStart,dvEnd,tmpDv,ndv);

   xBase:=seg[i]^.pvt.Re+xDelta;
   yBase:=seg[i]^.pvt.Im+yDelta;
   tmpDv.pos.Re:=xBase+kConst*cos(outNormal);
   tmpDv.pos.Im:=yBase+kConst*sin(outNormal);
   dvAppend(dvStart,dvEnd,tmpDv,ndv);

   xBase:=seg[i]^.pvt.Re;
   yBase:=seg[i]^.pvt.Im;
   tmpDv.pos.Re:=xBase+kConst*cos(outNormal);
   tmpDv.pos.Im:=yBase+kConst*sin(outNormal);
   dvAppend(dvStart,dvEnd,tmpDv,ndv);
  end;
end;

procedure shedTwo;
begin
  with bodyAry[bdy]^ do begin
   tmpDv.nearSeg:=i;
   tmpDv.mag:=survor[i]*0.5*seg[i]^.ds;
   outNormal:=seg[i]^.slope+0.5*pi;
   xDelta:=seg[i]^.ds*0.25*cos(seg[i]^.slope);
   yDelta:=seg[i]^.ds*0.25*sin(seg[i]^.slope);
   xBase:=seg[i]^.pvt.Re-xDelta;
   yBase:=seg[i]^.pvt.Im-yDelta;
   tmpDv.pos.Re:=xBase+kConst*cos(outNormal);
   tmpDv.pos.Im:=yBase+kConst*sin(outNormal);
   dvAppend(dvStart,dvEnd,tmpDv,ndv);

   xBase:=seg[i]^.pvt.Re+xDelta;
   yBase:=seg[i]^.pvt.Im+yDelta;
   tmpDv.pos.Re:=xBase+kConst*cos(outNormal);
   tmpDv.pos.Im:=yBase+kConst*sin(outNormal);
   dvAppend(dvStart,dvEnd,tmpDv,ndv);
  end;
end;

procedure shedOne;
begin
  with bodyAry[bdy]^ do begin
   tmpDv.nearSeg:=i;
   tmpDv.mag:=survor[i]*seg[i]^.ds;
   outNormal:=seg[i]^.slope+0.5*pi;
   tmpDv.pos.Re:=seg[i]^.pvt.Re+kConst*cos(outNormal);
   tmpDv.pos.Im:=seg[i]^.pvt.Im+kConst*sin(outNormal);
   dvAppend(dvStart,dvEnd,tmpDv,ndv);
  end;
end;

begin
  for bdy:=1 to numActiveBbody do
  with bodyAry[bdy]^ do begin
   tmpDv.tGen:=elapsed;
   tmpDv.nearBdy:=bdy;

   sx:=0.0;
   sxx:=0.0;
   for i:=1 to mSeg do begin
    sx:=sx+survor[i]*seg[i]^.ds;
    sxx:=sxx+sqr(survor[i]*seg[i]^.ds);
   end;
   qvar:=sqrt((sxx-sqr(sx)/mSeg)/(mSeg-1));

   if ndv<1000 then begin
    for i:=1 to mSeg do begin
     kConst:=ShedOffset;
     shedTwo;
    end;
   end
   else begin
    for i:=1 to mSeg do begin
     zz:=survor[i]*seg[i]^.ds/qvar;
     kConst:=ShedOffset;
     if seg[i]^.ds<1.2*dsAvg then begin
	if abs(zz)>3.00 then shedThree
	else if abs(zz)>2.0 then shedTwo
        else shedOne;
     end
     else shedTwo;
    end;
   end;
  end;
end;

[global]
procedure createVortexSheet;
var
  i,j: integer;
  curPtr: dvPtr;
  tmpVector: PVector;
begin
  for j:=1 to numActiveBbody do with bodyAry[j]^ do
   for i:=1 to mSeg do fullRhs[i]:=0.0;
  getDvRhs;

  totalBoundCirc:=0.0;

  tmpVector:=globalRhs;
  for j:=1 to numActiveBbody do with bodyAry[j]^ do
   for i:=1 to mSeg do begin
    tmpVector^.realNum:=fsRhs[i]-fullRhs[i];
    tmpVector:=tmpVector^.nextNum;
   end;
  {------------------------------------------------------------------
  bound circ. is to be regenerated by Martensen's operator, so reset
  totalBoundCirc to facilitate vorticity conservation equation
  -------------------------------------------------------------------}
  totalBoundCirc:=0.0;
  solution;
end;


  {//////////////////////////////////////////////////////////////////////
    Procedure vortexUV

    evaluates convection velocity of a D.V. at (x,y) under the influence
    of all the discrete vortices, return the vel vectors u,v
    (i.e. dvPtr^.u   and    dvPtr^.v  )  to the calling procedure.

    calling procedure : convect,  xconvect,  convectAB2

   //////////////////////////////////////////////////////////////////////}

procedure vortexUV;
  {/// I M P O R T A N T : reset velocity before calling this rotine ///}
var
  sinCore,cosCore,
  Xmn,Ymn,Mmag,Nmag,Mx,My,
  oneOver2Rsq,rsq,tMpUxd,tMpVyd    : double;
  mPtr,nPtr                        : dvPtr;
begin
  case vortexMode of
  directSum:
     begin
       {/// evaluate vortex to vortex induced velocity ///}
       mPtr:=dvStart;
       while mPtr<>nil do with mPtr^.Vortex do begin
       {Mmag:=mag;}
        Mmag:=vortDecay(mPtr);

        Mx:=pos.Re;
        My:=pos.Im;
        nPtr:=mPtr^.next;
        while nPtr<>nil do begin
         if nPtr<>mPtr then begin
         {Nmag:=nPtr^.Vortex.mag;}
          Nmag:=vortDecay(nPtr);
          Xmn:=Mx-nPtr^.vortex.pos.Re;
          Ymn:=My-nPtr^.vortex.pos.Im;
          rsq:=sqr(Xmn) + sqr(Ymn);
          {/// apply blob stream function to avoid singularity ///}
          if rsq>cutOffsq then begin
           oneOver2Rsq:=0.50/(pi*rsq);
           vel.Re:=vel.Re+Nmag*Ymn*oneOver2Rsq;
           vel.Im:=vel.Im-Nmag*Xmn*oneOver2Rsq;
           nPtr^.Vortex.vel.Re:=nPtr^.Vortex.vel.Re-Mmag*Ymn*oneOver2Rsq;
           nPtr^.Vortex.vel.Im:=nPtr^.Vortex.vel.Im+Mmag*Xmn*oneOver2Rsq;
          end
          else begin
          {/// using Rankine vortex model ///}
           vel.Re:=vel.Re+Nmag*Ymn*oneOver2piCSq;
           vel.Im:=vel.Im-Nmag*Xmn*oneOver2piCSq;
           nPtr^.Vortex.vel.Re:=nPtr^.Vortex.vel.Re-Mmag*Ymn*oneOver2piCSq;
           nPtr^.Vortex.vel.Im:=nPtr^.Vortex.vel.Im+Mmag*Xmn*oneOver2piCSq;
           {/// using Chorin vortex model ///}
          {sinCore:=oneOver2piC*Ymn/sqrt(rsq);
           cosCore:=oneOver2piC*Xmn/sqrt(rsq);
           vel.Re:=vel.Re+Nmag*sinCore;
           vel.Im:=vel.Im-Nmag*cosCore;
           nPtr^.Vortex.vel.Re:=nPtr^.Vortex.vel.Re-Mmag*sinCore;
           nPtr^.Vortex.vel.Im:=nPtr^.Vortex.vel.Im+Mmag*cosCore;}
          end;
         end;
         nPtr:=nPtr^.next;
        end;  {while}
        mPtr:=mPtr^.next;
       end;
     end;

  VortexInCell:
     begin
      {/// cloud-in-cell method ///}
     end;

  CellToCell:
     begin {/// cell-to-cell method ///}
       assignDVtoCtoc;
       evaluateCtocUV(CtocStart);
     end;
  end;{case}
end;

procedure newbodyUV;
var
  curPtr    : dvPtr;
  indVel    : zNumber;
  insideBdy : integer;
begin
  curPtr:=dvStart;
  while curPtr<>nil do with curPtr^.vortex do begin
   indVel.Re:=0.0;
   indVel.Im:=0.0;
   bodyxyuv(pos,indVel,insideBdy,noCheck);
   vel.Re:=vel.Re+indVel.Re;
   vel.Im:=vel.Im+indVel.Im;
   curPtr:=curPtr^.next;
  end; {while}
end; {/// EOProc. newbodyUV ///}

  {///////////////////////////////////////////////////////////////////////
    Procedure Xconvect

    deals with vortices convection under the influence of main stream,
    d.v. cloud, and vortex/body interaction, returns the new locations
    of all d.v. i.e new dvPtr^.x & dvPtr^.y to calling procedures.

   ///////////////////////////////////////////////////////////////////////}

[global]
procedure Xconvect;
var
  tmpU,tmpV    : double;
  i,nuit,count : integer;
  curPtr       : dvPtr;
begin
  {// create velRec records & store initial dv position ///}
   createVelRec;
  {/// set velocity to zero as variables are used for temp storage ///}
  initDvVelocity;
  updateVortexSheet;
  newbodyUV;
  vortexUV;
   {/// vels due to vortex cloud interaction ///}
  storeVel;
  curPtr:=dvStart;
  while curPtr<>nil do begin
   with curPtr^ do begin
    tmpU:=link^.veli.Re + UxCurt;
    tmpV:=link^.veli.Im + VyCurt;
    vortex.pos.Re:=link^.posi.Re+tmpU*dt;
    vortex.pos.Im:=link^.posi.Im+tmpV*dt;
   end; {with}
   curPtr:=curPtr^.next;
  end; {while}
  {** do iteras-1 iteration towards true central difference convection **}
   for nuit:=2 to xIteration do begin
      InitDvVelocity;
      updateVortexSheet;
      newbodyUV;
      vortexUV;
      curPtr:=dvStart;
   while curPtr<>nil do with curPtr^ do begin
    vortex.vel.Re:=0.5*(vortex.vel.Re + link^.veli.Re) + UxCurt;
    vortex.vel.Im:=0.5*(vortex.vel.Im + link^.veli.Im) + VyCurt;
    vortex.pos.Re:=link^.posi.Re+vortex.vel.Re*dt;
    vortex.pos.Im:=link^.posi.Im+vortex.vel.Im*dt;
    curPtr:=curPtr^.next;
   end; {while/with}
   end;{for}
end;   {/// EoProc xConvect ///}

[global]
procedure getGeneralVel;
var
  curPtr : dvPtr;
begin
  if ndv>modeChangeOver then easyAssign;
  curPtr:=dvStart;
  while curPtr<>nil do with curPtr^.vortex do begin
   vel.Re:=(pos.Re-curPtr^.link^.posi.Re)/dt;
   vel.Im:=(pos.Im-curPtr^.link^.posi.Im)/dt;
   curPtr:=curPtr^.next;
  end;
end;

function  autoTpSheet(thePoint:zNumber): double;
var
  normDir,timeRate,tmpStag: double;
  TpQi,angQi : quartet;
  bdy,i,j : integer;
  effpvt : zNumber;
begin
  tmpStag:=0.0;
  for bdy:=1 to numActiveBbody do begin
   for j:=1 to 4 do begin
    TpQi[j]:=0.0;
    angQi[j]:=0.0;
   end;
   for i:=1 to bodyAry[bdy]^.mSeg do
    with bodyAry[bdy]^.seg[i]^ do begin
    timeRate:=bodyAry[bdy]^.survor[i]*ds/dt;
    normDir:=BodyAry[bdy]^.seg[i]^.slope+0.5*pi;
    effpvt.Re:=pvt.Re-0.1*BodyAry[bdy]^.seg[i]^.ds*cos(normDir);
    effpvt.Im:=pvt.Im-0.1*BodyAry[bdy]^.seg[i]^.ds*sin(normDir);
    multiPotAngle(effpvt,thePoint,angQi);
    for j:=1 to 4 do TpQi[j]:=TpQi[j]+timeRate*angQi[j];
   end;
  tmpStag:=tmpStag+match4(TpQi);
  end;
  autoTpSheet:=tmpStag/pi;
end;

[global]
procedure QuickPitot ( thePoint:zNumber;var totalP,staticP: double;
                       var indVel: zNumber);
var
  theMag,xxi,yyi,theSquare,tmpStag : double;
  curPtr  : dvPtr;
  insideBdy : integer;
  oneVel,accVel : zNumber;
begin
  { calculate the stagnation pressure distribution induced by moving
    point vortices }
  tmpStag:=0.0;
  accVel.Re:=0.0;
  accVel.Im:=0.0;
  curPtr:=dvStart;
  while curPtr<>nil do with curPtr^ do begin
   if curPtr^.vortex.nearBdy<>0 then theMag:=curPtr^.vortex.mag
   else theMag:=vortDecay(curPtr);
   xxi:=thePoint.Re-vortex.pos.Re;
   yyi:=thePoint.Im-vortex.pos.Im;
   theSquare:=sqr(xxi)+sqr(yyi);
   if theSquare>=cutOffsq then begin
    oneVel.Re:= theMag*yyi/theSquare;
    oneVel.Im:=-theMag*xxi/theSquare;
   end
   else begin
    oneVel.Re:= theMag*yyi/cutOffSq;
    oneVel.Im:=-theMag*xxi/cutOffSq;
   end;
   tmpStag:=tmpStag+oneVel.Re*vortex.vel.Re+oneVel.Im*vortex.vel.Im;
   accVel.Re:=accVel.Re+oneVel.Re;
   accVel.Im:=accVel.Im+oneVel.Im;
   curPtr:=curPtr^.next;
  end;
  totalP:=tmpStag/pi+autoTpSheet(thePoint)+1.0;
  indVel.Re:=uxCurt+0.5*accVel.Re/pi;
  indVel.Im:=vyCurt+0.5*accVel.Im/pi;
  bodyxyuv(thePoint,indVel,insideBdy,noCheck);
  staticP:=totalP-(sqr(indVel.Re)+sqr(indVel.Im));
end;

procedure ProbeSample;
var
  curPoint         : PlinkPt;
  tmpTotP,tmpStatP : double;
  tmpVel           : zNumber;
  tmpFirstSet      : PProbeData;
begin
  if (elapsed>=probeOn) and (elapsed<=probeOff) then begin
  tmpFirstSet:=nil;
  curPoint:=probePoint;
  while curPoint<> nil do begin
   QuickPitot(curPoint^.ptCoord,tmpTotP,tmpStatP,tmpVel);
   probeAppend(tmpFirstSet,tmpTotP,tmpStatP,tmpVel);
   curPoint:=curPoint^.downLink;
  end;
  probeListAppend(probeStart,probeEnd,tmpFirstSet,elapsed);
  end;
end;

procedure searchStagnationPts(bdy:integer;var frontPt, rearPt: integer);
const smallvalue=1.0e-4;
var
  i : integer;
  gammaA, gammaB : double;
begin
  frontPt:=0;
  rearPt:=0;
  with bodyAry[bdy]^ do begin
   gammaA:=survor[mSeg];gammaB:=survor[1];
   if abs(gammaA-gammaB)>smallValue then begin
   if not sameSign(gammaA,gammaB) then
    if gammaA<gammaB then frontPt:=1
    else rearPt:=mSeg
   end;

   for i:= 1 to mSeg-1 do begin
    gammaA:=survor[i]; gammaB:=survor[i+1];
    if abs(gammaA-gammaB)>smallValue then begin
     if sameSign(gammaA,gammaB)= false then
      if gammaA<gammaB then frontPt:=i
       else rearPt:=i
    end;
   end;
   frontStag:=frontPt;
  end;
end;

procedure searchNewStag(bdy, oldStag : integer; var newStag : integer);
var
  i,j,segA,segB : integer;
  changeSign : boolean;
  gammaA,gammaB : double;

  procedure findStag(inA,inB:integer);
  var
    GammaRatio,totalength,lOne : double;
  begin
    GammaRatio:=abs(gammaA/gammaB);
    with bodyAry[bdy]^ do
     begin
      totalength:=(seg[inA]^.ds+seg[inB]^.ds);
      lOne:=gammaRatio*totalength/(1+gammaRatio);
      if lOne<seg[inA]^.ds then newStag:=inA
       else if lOne>seg[inA]^.ds then newStag:=inB;
     end;
  end;

begin
  i:=oldStag;j:=oldStag;
  with bodyAry[bdy]^ do begin
    Repeat
     segA:=getSegmentTag(bdy,i);
     segB:=getSegmentTag(bdy,i+1);
     gammaA:=survor[segA];gammaB:=survor[segB];
     changeSign:=sameSign(gammaA,gammaB);
     if changeSign then findStag(segA,segB)
     else begin
       i:=i+1;
       segA:=getSegmentTag(bdy,j);
       segB:=getSegmentTag(bdy,j-1);
       gammaA:=survor[segA];gammaB:=survor[segB];
       changeSign:=sameSign(gammaA,gammaB);
       if changeSign then findStag(segA,segB)
       else j:=j+1;
      end;
    until changeSign;
  end;
end;

procedure PressureDist;
var
  i,bdy,curNum         : integer;
  nforce,timeAvgP,
  tempdrag,templift,
  dPSeg,
  pcorrt,pvel,
  pcoeff,sumdp         : double;
  dPdsSeg,dPdsMean,
  Pdist,PnAvg          : vectorM;
  thisListStart        : PfcEntry;
  tmpTotP,tmpStatP     : double;
  tmpVel               : zNumber;
begin
  pcoeff:=-2.0/dt;
  thisListStart:=nil;
  for bdy:= 1 to numActiveBbody do with bodyAry[bdy]^ do begin
    for i:=1 to mSeg do begin
      dPdsSeg[i]:=0.0;
      dPdsMean[i]:=0.0;
      Pdist[i]:=0.0;
    end;

    sumdp:=0.0;
    for i:=1 to mSeg do begin
      dPdsSeg[i]:=pcoeff*survor[i];
      sumdp:=sumdp+dPdsSeg[i]*seg[i]^.ds;
      pdist[i]:=sumdp;
    end;

    {/// correct for stagnation pressure ///}
    QuickPitot(seg[frontStag]^.datPt,tmpTotP,tmpStatP,tmpVel);
    pcorrt:=tmpStatP-pdist[frontStag];

    for i:=1 to mSeg do begin
      pdist[i]:=pdist[i]+pcorrt;
     end;

    {/// calculate surface forces ///}
    PnAvg[1]:=0.5*(pdist[mSeg]+pdist[1]);
    for i:=2 to mSeg do begin
     PnAvg[i]:=0.5*(pdist[i]+pdist[i-1]);
    end;
    tempdrag:=0.0;
    templift:=0.0;
    {if publish then write(' p : ',pdist[frontStag]:9:6,' dp : ',sumdp:9:6);}

    for i:=1 to mSeg do with seg[i]^ do begin
      nforce:=PnAvg[i]*ds;
      templift:=templift-nforce*cos(slope);
      tempdrag:=tempdrag+nforce*sin(slope);
    end;

    if PonBody.NumRecord=0 then begin
     PonBody.NumRecord:=PonBody.NumRecord+1;
     for i:=1 to mSeg do begin
      PonBody.MinPress[i]:=pdist[i];
      PonBody.MaxPress[i]:=pdist[i];
      PonBody.AvgPress[i]:=pdist[i];
     end;
    end
    else begin
     curNum:=PonBody.NumRecord;
     for i:=1 to mSeg do begin
      PonBody.AvgPress[i]:=(PonBody.AvgPress[i]*curNum+pdist[i])/(curNum+1);

      PonBody.MinPress[i]:=
      sqrt((sqr(PonBody.MinPress[i])*curNum+
	    sqr(pdist[i]-PonBody.AvgPress[i]))/(curNum+1));

      if PonBody.AvgPress[i]>PonBody.MaxPress[i]
        then PonBody.MaxPress[i]:=PonBody.AvgPress[i];
     end;

     PonBody.NumRecord:=PonBody.NumRecord+1;
    end;
    {/// append individual body to force list ///}
    fcEntryAppend(thisListStart,tempDrag,tempLift);
    {/// save current surface vortice to survorp[j] ///}
  end; {for/with}
  fceListAppend(fcListStart,fcListEnd,thisListStart,elapsed);
end;

[global]
procedure PotentialFlow;
var
  bdy,a,b : integer;
  lastCirc : array[1..maxBody] of double;
begin
  getpotFlowRhs;
  solution;
  goKutta:=false;
  for bdy:=1 to numActiveBbody do
   if bodyAry[bdy]^.aerofoil then goKutta:=true;
  if goKutta then begin
    for bdy:=1 to numActiveBbody do
     lastCirc[bdy]:=bodyAry[bdy]^.prescribeCirc;
    EnforceKuttaCondition;
    getpotFlowRhs;
    solution;
    for bdy:=1 to numActiveBbody do
     bodyAry[bdy]^.prescribeCirc:=lastCirc[bdy];
  end;
  for bdy:=1 to numActiveBbody do
   searchStagnationPts(bdy,a,b);
end;

[global]
procedure ViscousFlow;
begin
  updateVortexSheet;
end;

procedure saveFileNow;
var
  strSave,saveName,cpsname : TString;
begin
  if ((istep mod saveFileInt)=0) or (istep=numTimeStep) then
   begin
     isave:=isave+1;
     OurStr(isave,strSave);
     saveName:=FileGrp+strSave+'.DVR';
     SaveResult(saveName);
     cpsName:=FileGrp+strSave+'.CPS';
     SaveCP(cpsName);
     AppendForceList;
     if doProbing then saveProbeList;
     if publish then writeln(' save files  ',saveName,' ',cpsName);
   end;
end;

procedure resetValue;
begin
  firstAccess:=true;
  snuff:=0.0;          {initial strength of snuffed vorticity :=0.0   }
  elapsed:=0.0;        {set time zero                                 }
  ndv:=0;              {initial number of free vortex :=0             }
  fceListLength:=0;
  freeDvMemory(dvStart, dvEnd);
  allGone:=true;
end;

procedure initMemCounter;
begin
  newDvMem:=0;
  newLinkMem:=0;
  newCellMem:=0;
  newChainMem:=0;
  newzCoordMem:=0;
  newDataBankMem:=0;
  newCGroupMem:=0;
  newzMmtMem:=0;
  newCtocMem:=0;
  chaindump:=0;
  chainReUse:=0;
end;

procedure reportCounter;
begin
  {write(' new DV:',newDvMem:4);}
  {write(' new Link:',newLinkMem:4);}
  {write(' new Cell:',newCellMem:4);}
  {write(' new Chain:',newChainMem:4);}
  {write(' disposed:',chaindump:4);}
  {write(' into bim :',chainReUse:4);}
  {write(' new zCoord:',newzCoordMem:4);}
  {write(' new dataBank:',newDataBankMem:4);}
  {write(' new CGroup:',newCGroupMem:4);}
  {write(' new zMomemt:',newzMmtMem:4);}
  {write(' new Ctoc:',newCtocMem:4);}
  write(' dvBin ',nTrash:6:0);
  write(' ChainBin ',nTrash:6:0);
  write(' ',ndv:8:1);
end;

[global]
procedure FreshStart;
begin
  resetValue;
  QuitGridTTT;
  initGridData;
  VortexMode:=DirectSum;
  istep:=0;isave:=0;
  potentialFlow;     { potential flow analysis to find stagnation pts }
  if goKutta then begin
    getpotFlowRhs;
    solution;
  end;
  shedDv;            { initial shedding at time zero }
  while (iStep < numTimeStep) do begin
    if (vortexMode=directSum) and (ndv>modeChangeOver) then vortexMode:=cellToCell;
    clockOn;
    if ((istep+1) mod mergStep)=0 then DVMerging;
    xConvect;
    RandomWalk;
    createVortexSheet;
    elapsed:=elapsed+dt;
    istep:=istep+1;
    if checkTimeOn and timesUp then istep:=numTimeStep;
    getGeneralVel;
    pressureDist;
    if doProbing then probeSample;
    shedDv;
    clockOff;
    saveFileNow;
    reportCounter;
    if publish then write(' ',istep,' of ',numTimeStep,' done  ', AlarmDate);
    writeln;
   end;
end;

[global]
procedure reStart;
begin
  QuitGridTTT;
  initGridData;
  vortexMode:=directSum;
  potentialFlow;              { potential flow anal to find stag. pts }
  viscousFlow;
  istep:=0;
  writeln(' decay factor : ',decayfactor);
  writeln(' ctc level    : ',numSubLevel);
  while (iStep < numTimeStep) do begin
    if (vortexMode=directSum) and (ndv>modeChangeOver)
      then vortexMode:=cellToCell;
     clockOn;
     {initMemCounter;}
     if ((istep+1) mod mergStep)=0 then DVMerging;
     xConvect;
     RandomWalk;
     createVortexSheet;
     elapsed:=elapsed+dt;
     istep:=istep+1;
     if checkTimeOn and timesUp then istep:=numTimeStep;
     getGeneralVel;
     pressureDist;
     if doProbing then probeSample;
     shedDv;
     clockOff;
     reportCounter;
     saveFileNow;
     if publish then write(' ',istep,' of ',numTimeStep,' done ', alarmDate);
     writeln;
   end;
end;

to begin do
  goKutta:=false;
end.

