[inherit('vglobal')]
Module utility(output);

const
  dayOfYear   = arrayYear(0,31,59,90,120,151,181,212,243,273,304,334);
  crossOver   = 1;
  SameSide    = 2;
  NotThisLine = 3;
  NumBin      = 3;

type
  dvLogRec = record
   IsInside,
   stillWell,
   {tooClose,}
   twin          : boolean;
   nearBody,
   segEst,segEr  : integer;
   nearDist      : double;
  end;


  TdvBin = array[1..NumBin] of integer;

var
  AlarmYear,AlarmMonth,AlarmDay,AlarmHour,AlarmMin     : integer;
  ghour,gmin,gsec: integer;
  StartClock,StopClock                : double;

  dvBin                               : TdvBin;

  KSeed                               : longInt;
  KINext,KINextp                      : integer;
  KnuthMa                             : array[1..55] of longInt;

  K2BSeed,K2BInext,K2BInextp          : integer;
  K2BMa                               : array[1..55] of double;

  PortSeed,ran2Iy        : longInt;
  ran2Ir                              : array[1..97] of longInt;
  ran2Seed                            : longInt;

{procedure KnuthRanise;
procedure Knuth2Bise;
procedure PortRandomise;}

Procedure GetSysDate (Var aYear,aMonth,aDay:Integer);
Var
   TimeVar : TimeStamp;
begin
   GetTimestamp(Timevar);
   with TimeVar do
   begin
    aDay  :=day;
    aMonth:=month;
    aYear :=year;
   end;
end;

Procedure GetSysTime (Var aHour, aMin, aSec: Integer);
Var
   TimeVar : TimeStamp;
begin
   GetTimestamp(Timevar);
   with TimeVar do
   begin
    aHour :=hour;
    aMin  :=minute;
    aSec  :=second;
   end;
end;

[global]
function StrLength(aString:TString):integer;
begin
  {StrLength:=Ord(aString[0]);}
  StrLength:=aString.length;
end;

[global]
procedure OurStr(inInt:integer;var aStr:TString);
var
  cutter,aReal:real;
  aCode:integer;
begin
  aStr:='';
  aReal:=abs(inInt);
  repeat
  cutter:=aReal*0.1;
  aCode:=round((cutter-trunc(cutter))*10.0);
  aStr:=Chr(aCode+48)+aStr;
  aReal:=trunc(cutter);
  until aReal<=0.0;
  if inInt<0 then aStr:='-'+aStr;
end;

[global]
function copyStr(masterStr:TString;index,count:integer):TString;
var
  tmpStr:TString;
  i,jj: integer;
begin
  jj:=StrLength(masterStr);
  tmpStr:='';
  if index<1 then index:=1;
  i:=index;
  while (i<=jj) and (i<index+count) do
  begin
   tmpStr:=tmpStr+masterStr[i];
   i:=i+1;
  end;
  copyStr:=tmpStr;
end;

[global]
function PosStr(subStr,masterStr:TString):integer;
var
  ii,jj,a,b : integer;
  aBingo :Boolean;
begin
  if (subStr<>'') and (masterStr<>'') then
  begin
    ii:=strLength(subStr);
    jj:=strLength(masterStr);
    aBingo:=false;
    a:=1;
    b:=1;
    repeat
     if subStr[a]=masterStr[b] then
      begin
       aBingo:=(a=ii);
       a:=a+1;
       b:=b+1;
      end
     else b:=b+1;
    until (a>ii) or (b>jj) or aBingo;
    if aBingo then PosStr:=b-ii
    else PosStr:=0;
  end
  else PosStr:=0;
end;

[global]
procedure Value(aString:TString; var outValue:double; var errCode:integer);
const validNum=['0'..'9','+','-','.'];
var
  dec,jj,i:integer;
  tmpValue:double;
  aChar:char;
begin
  tmpValue:=0.0;
  jj:=StrLength(aString);
  if jj>0 then
  begin
   dec:=0;
   errCode:=0;
   i:=jj;
   Repeat
    aChar:=aString[i];
    if aChar in validNum then
    begin
      if aChar in ['0'..'9'] then
       tmpValue:=tmpValue+(Ord(aChar)-48)*exp((jj-i+dec)*ln(10))
      else
      if aChar='.' then
        if dec=0 then
          begin
            dec:=(i-jj-1);
            tmpValue:=tmpValue*exp((i-jj)*ln(10));
          end
        else errCode:=1
      else
      if (aChar in ['+','-']) and (i=1) then
        if aChar='-' then tmpValue:=-tmpValue
        else tmpValue:=tmpValue
      else errCode:=1;
    end
    else errCode:=1;

    i:=i-1;
   until (errCode<>0) or (i<1);
  end;
  if errCode=0 then outValue:=tmpValue
  else outValue:=0.0;
end;

[global]
procedure ValInt(aString:TString; var outValue:Integer; var errCode:integer);
const validNum=['0'..'9','+','-'];
var
  jj,i:integer;
  tmpValue:double;
  aChar:char;
begin
  tmpValue:=0.0;
  jj:=StrLength(aString);
  if jj>0 then
  begin
   errCode:=0;
   i:=jj;
   Repeat
    aChar:=aString[i];
    if aChar in validNum then
    begin
      if aChar in ['0'..'9'] then
       tmpValue:=tmpValue+(Ord(aChar)-48)*exp((jj-i)*ln(10))
      else
      if (aChar in ['+','-']) and (i=1) then
        if aChar='-' then tmpValue:=-tmpValue
        else tmpValue:=tmpValue
      else errCode:=1;
    end
    else errCode:=1;
    i:=i-1;
   until (errCode<>0) or (i<1);
  end;
  if errCode=0 then outValue:=trunc(tmpValue)
  else outValue:=0;
end;

[global]
function UpperCase(aChar:Char):Char;
const
  lowerSet=['a'..'z'];
begin
  if aChar in lowerSet then UpperCase:=Chr(ord(aChar)-32)
  else UpperCase:=aChar;
end;

procedure KnuthRanise;
var aa,bb,cc:integer;
begin
  if KRandSeed<0 then KSeed:=KRandSeed
  else if KRandSeed>0 then KSeed:=-KRandSeed
  else
  begin
   getSysTime(aa,bb,cc);
   KSeed:=-(aa*3600+bb*60+cc);
   if KSeed>0 then KSeed:=-KSeed;
  end;
  KnuthRan;
end;

[global]
function KnuthRan:double;
const
  mBig=1000000000;
  mseed=161803398;
  mz=0.0;
  fac=1.0E-9;
var
  i,ii,k,mj,mk:LongInt;
begin
  if KSeed=0 then KnuthRanise;
  if KSeed<0 then
   begin
    mj:=mseed+KSeed;
    if mj>=0 then mj:=(mj mod mBig)
    else mj:=mBig-abs(mj)+mBig*trunc(abs(mj)/mBig);
    KnuthMa[55]:=mj;
    mk:=1;
    for i:=1 to 54 do
     begin
      ii:=21*i mod 55;
      KnuthMa[ii]:=mk;
      mk:=mj-mk;
      if mk<mz then mk:=mk+mBig;
      mj:=KnuthMa[ii];
     end;
    for k:=1 to 4 do
     begin
      for i:=1 to 55 do
       begin
        KnuthMa[i]:=KnuthMa[i]-KnuthMa[1+((i+30) mod 55)];
        if KnuthMa[i]<mz then KnuthMa[i]:=KnuthMa[i]+mBig;
       end;
      end;
    KINext:=0;
    KINextp:=31;
    KSeed:=1;
   end;
  KINext:=KINext+1;
  if KINext=56 then KINext:=1;
  KINextp:=KINextp+1;
  if KINextp=56 then KINextp:=1;
  mj:=KnuthMa[KINext]-KnuthMa[KINextp];
  if mj<mz then mj:=mj+mBig;
  KnuthMa[KINext]:=mj;
  KnuthRan:=mj*fac;
end;

procedure Knuth2Bise;
var aa,bb,cc : integer;
begin
  if K2BRandSeed<0 then K2BSeed:=K2BRandSeed
  else if K2BRandSeed>0 then K2BSeed:=-K2BRandSeed
  else
  begin
   getSysTime(aa,bb,cc);
   K2BSeed:=-(aa*3600+bb*60+cc);
   if K2BSeed>0 then K2BSeed:=-K2bSeed;
  end;
  Knuth2B;
end;

[global]
function Knuth2B:double;
{ return a uniform random deviate between 0.0 and 1.0. Set idum to
  any negative value to initialise or reinitialise the sequence.   }
const
  mBig=4.0E6;
  mseed=161803398;
  mz=0.0;
  fac=2.5E-7;             {1/mBig}
var
  i,ii,k:integer;
  mj,mk:double;
begin
  if K2BSeed=0 then Knuth2Bise;
  if K2BSeed<0 then
   begin
    mj:=mseed+K2BSeed;
    if mj>=0.0 then mj:=mj-mBig*trunc(mj/mBig)
    else mj:=mBig-abs(mj)+mBig*trunc(abs(mj)/mBig);
    K2BMa[55]:=mj;
    mk:=1;
    for i:=1 to 54 do
     begin
      ii:=21*i mod 55;
      K2BMa[ii]:=mk;
      mk:=mj-mk;
      if mk<mz then mk:=mk+mBig;
      mj:=K2BMa[ii];
     end;
    for k:=1 to 4 do
     begin
      for i:=1 to 55 do
       begin
        K2BMa[i]:=K2BMa[i]-K2BMa[1+((i+30) mod 55)];
        if K2BMa[i]<mz then K2BMa[i]:=K2BMa[i]+mBig;
       end;
      end;
    K2BInext:=0;
    K2BInextp:=31;
    K2BSeed:=1;
   end;
  K2BInext:=K2BInext+1;
  if K2BInext=56 then K2BInext:=1;
  K2BInextp:=K2BInextp+1;
  if K2BInextp=56 then K2BInextp:=1;
  mj:=K2BMa[K2BInext]-K2BMa[K2BInextp];
  if mj<mz then mj:=mj+mBig;
  K2BMa[K2BInext]:=mj;
  Knuth2B:=mj*fac;
end;

procedure PortRandomise;
var aa,bb,cc : integer;
begin
  if PortRandSeed<0 then PortSeed:=PortRandSeed
  else if PortRandSeed>0 then PortSeed:=-PortRandSeed
  else
  begin
   getSysTime(aa,bb,cc);
   PortSeed:=-(aa*3600+bb*60+cc);
   if PortSeed>0 then PortSeed:=-PortSeed;
  end;
  PortRandom;
end;

[global]
function PortRandom:double;
{ return a uniform random deviate between 0.0 and 1.0. Set idum to
  any negative value to initialise or reinitialise the sequence.
  idum must be of type longInt (4-byte integer) in turbo pascal    }
const
  m=714025;
  ia=1366;
  ic=150889;
  rm=1.4005112E-6;  {1/m}
var
  j: integer;
begin
  if PortSeed=0 then PortRandomise;
  if PortSeed<0 then
  begin
    PortSeed:=(ic-PortSeed) mod m;
    for j:=1 to 97 do
     begin
       PortSeed:=(ia*PortSeed+ic) mod m;
       ran2Ir[j]:=PortSeed;
     end;
    PortSeed:=(ia*PortSeed+ic) mod m;
    ran2Iy:=PortSeed;
  end;
  j:=1 + (97*Ran2Iy) div m;
  if (j>97) or (j<1) then
  begin
    writeln('pause in rotine ran2 ');
  end;
  ran2Iy:=ran2Ir[j];
  PortRandom:=ran2Iy*rm;
  PortSeed:=(ia*PortSeed+ic) mod m;
  Ran2Ir[j]:=PortSeed;
end;

{procedure wait;
var
  dump     : char;
  functKey : boolean;
begin
  repeat
   dump:=ReadKey;
   if dump <> #0 then functKey:=false else functKey:=true;
   if functKey then dump:=ReadKey;
  until (dump<>'');
end;}

[global]
procedure setAlarm(aYear,aMonth,aDay,aHour,aMin:integer);
begin
  AlarmYear:=aYear;
  AlarmMonth:=aMonth;
  AlarmDay:=aDay;
  AlarmHour:=aHour;
  AlarmMin:=aMin;
end;

[global]
procedure clockOn;
var
  ghour,gmin,gsec: integer;
begin
  getSysTime(ghour,gmin,gsec);
  StartClock:=ghour*3600+gmin*60+gsec;
end;

[global]
procedure clockOff;
var
  ghour,gmin,gsec: integer;
begin
  getSysTime(ghour,gmin,gsec);
  StopClock:=ghour*3600+gmin*60+gsec;
  if publish then writeln(' Time : ',(StopClock-StartClock):0:2,' sec');
end;

[global]
function  timesUp:Boolean;
var
  year,month,day,hour,minute,second,DayDiff,MinDiff : Integer;
begin
  GetSysTime(hour,minute,second);
  GetSysDate(year,month,day);
  DayDiff:=(alarmYear-year)*365 + dayOfYear[alarmMonth]-dayOfYear[month] + alarmDay-day;
  if (DayDiff>0) then TimesUp:=false
   else if (DayDiff<0) then TimesUp:=true
    else
      begin
       MinDiff:=(alarmHour-hour)*60+(alarmMin-minute);
       if (MinDiff>0) then TimesUp:=false
        else TimesUp:=true;
      end;
end;

{function getBigOf2(nos1, nos2: double): double;
begin
  if nos1 >= nos2 then getBigOf2:=nos1
  else getBigOf2:=nos2;
end;}

[global]
function ValidDate(aString:TString;var dayW,monthW,yearW:integer):boolean;
const
  dayInMonth = arrayYear(31,29,31,30,31,30,31,31,30,31,30,31);
  Yr='20';
var
  errCode:integer;
  century,dayStr,monthStr,yearStr: TString;
  dow,yrLength,count,ii:integer;
  lookupStr: boolean;
begin
  {GetDate(yearW,monthW,dayW,dow);
  str(yearW:0,YearStr);
  century:=copyStr(yearStr,1,2);}
  ii:=StrLength(aString);
  dayStr:='';
  monthStr:='';
  yearStr:='';
  yrLength:=0;
  count:=1;
  lookupStr:=true;
  while (count<=ii) and lookupStr do
   begin
    if (aString[count] in ['0'..'9']) then dayStr:=dayStr+aString[count]
    else if aString[count]='/' then lookupStr:=false;
    count:=count+1;
   end;
  lookupStr:=true;
  while (count<=ii) and lookupStr do
   begin
    if (aString[count] in ['0'..'9']) then monthStr:=monthStr+aString[count]
    else if aString[count]='/' then lookupStr:=false;
    count:=count+1;
   end;
  lookupStr:=true;
  while (count<=ii) and lookupStr do
   begin
    if (aString[count] in ['0'..'9']) then
     begin
      yearStr:=yearStr+aString[count];
      yrLength:=yrLength+1;
     end
    else if aString[count]='/' then lookupStr:=false;
    count:=count+1;
   end;
  if yrLength=2 then yearStr:=Yr+yearStr;{century+yearStr;}

  valInt(dayStr,dayW,errCode);
  valInt(monthStr,monthW,errCode);
  valInt(yearStr,yearW,errCode);
  validDate:=false;
  if (monthW>0) and (monthW<=12) then
   if (dayW>0) and (dayW<=dayinMonth[monthW]) then
    if (yearW>1990) and (yearW<=2099) then
     validDate:=true;
end;

[global]
function ValidTime(aString:TString; var hourW,minuteW:integer):boolean;
var
  errCode:integer;
  hourStr,minuteStr: TString;
  {hourW,minuteW,}
  count,ii:integer;
  lookupStr: boolean;
begin
  ii:=StrLength(aString);
  hourStr:='';
  minuteStr:='';
  count:=1;
  lookupStr:=true;
  while (count<=ii) and lookupStr do
   begin
    if (aString[count] in ['0'..'9']) then hourStr:=hourStr+aString[count]
    else if aString[count]=':' then lookupStr:=false;
    count:=count+1;
   end;
  lookupStr:=true;
  while (count<=ii) and lookupStr do
   begin
    if (aString[count] in ['0'..'9']) then minuteStr:=minuteStr+aString[count]
    else if aString[count]=':' then lookupStr:=false;
    count:=count+1;
   end;

  valInt(hourStr,hourW,errCode);
  valInt(minuteStr,minuteW,errCode);
  validTime:=false;
  if (hourW>=0) and (hourW<=23) then
   if (minuteW>=0) and (minuteW<=59) then
     validTime:=true;
end;

[global]
function getSegmentTag(bdy,inputTag:integer):integer;
begin
  with bodyAry[bdy]^ do
  if inputTag<=0 then getSegmentTag:=mSeg+inputTag else
    if inputTag>mSeg then getSegmentTag:=inputTag-mSeg
      else getSegmentTag:=inputTag;
end;

{/// func getCofSlope calculates the change of slope at a given segment ///}
[global]
function getCofSlope(bdy,inputElem:integer):double;
var
  upper,lower: integer;
  tmpChgSlope: double;
begin
  lower:=getSegmentTag(bdy,inputElem-1);
  upper:=getSegmentTag(bdy,inputElem+1);
  with bodyAry[bdy]^ do
   tmpChgSlope:=seg[upper]^.slope-seg[lower]^.slope;
  if (tmpChgSlope>pi) then tmpChgSlope:=tmpChgSlope-2.0*pi;
  getCofSlope:=tmpChgSlope;
end;

{/// func getSlope2Seg calculates the slope between two given segments ///}
[global]
function getSlope2Seg(bdy,Elem1,Elem2:integer):double;
var
  upper,lower: integer;
  tmpChgSlope: double;
begin
  lower:=getSegmentTag(bdy,Elem1);
  upper:=getSegmentTag(bdy,Elem2);
  with bodyAry[bdy]^ do
    tmpChgSlope:=seg[upper]^.slope-seg[lower]^.slope;
  if (tmpChgSlope>pi) then tmpChgSlope:=tmpChgSlope-2.0*pi;
  getSlope2Seg:=tmpChgSlope;
end;


(*
   func getAngleR calculates the angle (radian) of a given lines to the
   x-axis.



                           x pt(x2,y2)
                         /
                       /
                     /  angleR, in radian   ccw +ve (range: 0 to pi/2    )
                - -/------------            cw  -ve (range: 0 to -3*pi/2 )
                 /
               x  pt(x1,y1)
         _____________________________________  X-axis

*)

[global]
function getAngleR(inpt1,inpt2:zNumber):double;
var
  deltax,deltay,InvSqrr,tpSin,tpCos,absCos,tpSinCos,tangt : double;
begin
  deltaY:=sqr(inPt2.Im-inpt1.Im);
  deltaX:=sqr(inpt2.Re-inpt1.Re);
  if (deltaX<>0) or (deltaY<>0) then
   begin
    InvSqrr:=1.0/sqrt(deltaY+deltaX);
    tpSin:=(inPt2.Im-inpt1.Im)*InvSqrr;tpCos:=(inpt2.Re-inpt1.Re)*InvSqrr;
    absCos:=abs(tpCos);
    tpSinCos:=tpSin*tpCos;
    if absCos>tooSmall then tangt:=arctan(tpSin/tpCos);
    if (absCos<=tooSmall) then getAngleR:=tpSin/abs(tpSin)*pi*0.5;
    if tpCos>tooSmall then getAngleR:=tangt;
    if tpCos<(-tooSmall) then getAngleR:=tangt-pi;
   end
  else getAngleR:=0.0;
end;

{function getVelAng(inUx,inVy:double):double;
var
  InvSqrr,tpSin,tpCos,absCos,tpSinCos,tangt : double;
begin
  if (inUx<>0) or (inVy<>0) then
   begin
    InvSqrr:=1.0/sqrt(sqr(inUx)+sqr(inVy));
    tpSin:=inVy*InvSqrr;tpCos:=(inUx)*InvSqrr;
    absCos:=abs(tpCos);
    tpSinCos:=tpSin*tpCos;
    if absCos>tooSmall then tangt:=arctan(tpSin/tpCos);
    if (absCos<=tooSmall) then getVelAng:=tpSin/abs(tpSin)*pi*0.5;
    if tpCos>tooSmall then getVelAng:=tangt;
    if tpCos<(-tooSmall) then getVelAng:=tangt-pi;
   end
  else getVelAng:=0.0;
end;}

{function getDvVelAng(inDv:dvPtr):double;
var
  InvSqrr,tpSin,tpCos,absCos,tpSinCos,tangt : double;
begin
  if (inDv^.vortex.vel.Re<>0) or (inDv^.vortex.vel.Im<>0) then
   begin
    InvSqrr:=1.0/sqrt(sqr(inDv^.Vortex.vel.Re)+sqr(inDv^.Vortex.vel.Im));
    tpSin:=inDv^.Vortex.vel.Im*InvSqrr;tpCos:=(inDv^.vortex.vel.Re)*InvSqrr;
    absCos:=abs(tpCos);
    tpSinCos:=tpSin*tpCos;
    if absCos>tooSmall then tangt:=arctan(tpSin/tpCos);
    if (absCos<=tooSmall) then getDvVelAng:=tpSin/abs(tpSin)*pi*0.5;
    if tpCos>tooSmall then getDvVelAng:=tangt;
    if tpCos<(-tooSmall) then getDvVelAng:=tangt-pi;
   end
  else getDvVelAng:=0.0;
end;}

(*
   func get3ptAngle calculates the angle (radian) between lines defined
   by 3 pts.



                           x pt3 (C)
                         /
                       /
                     /  angleR, in radian   ccw +ve (range: 0 to pi/2    )
                   /                        cw  -ve (range: 0 to -3*pi/2 )
                 /
               x   - - - - - - x
             pt2 (B)           pt1 (A)
         _____________________________________  X-axis

*)

[global]
function get3ptAngle(ptA,ptB,ptC:zNumber):double;
var
  incAng,angleCB,angleAB : double;
begin
  angleCB:=getAngleR(ptB,ptC);
  angleAB:=getAngleR(ptB,ptA);
  incAng:=angleCB-angleAB;
  if incAng>pi then get3ptAngle:=incAng-2.0*pi
  else if incAng<-pi then get3PtAngle:=incAng+2.0*pi
  else if abs(abs(incAng)-pi)<tooSmall then get3ptAngle:=pi
  else get3ptAngle:=incAng;
end;

[global]
function facingAngle(angleA,angleB:double):boolean;
var
  prod : double;
begin
  prod:=angleA*angleB;
  if prod<=0.0 then facingAngle:=true
  else if (abs(abs(angleA)-pi)<tooSmall) or (abs(abs(angleB)-pi)<tooSmall) then
   facingAngle:=true
  else facingAngle:=false;
end;

[global]
function  crossLine(lineA,lineB,ptA,ptB:zNumber):integer;
var
  angleCB,angleAB : double;
begin
  angleCB:=get3ptAngle(ptA,lineA,ptB);
  angleAB:=get3ptAngle(ptA,lineB,ptB);
  if facingAngle(angleCB,angleAB) then
   begin
     angleCB:=get3ptAngle(lineA,ptA,lineB);
     angleAB:=get3ptAngle(lineA,ptB,lineB);
     if facingAngle(angleCB,angleAB) then crossLine:=CrossOver
     else crossline:=SameSide;
   end
  else crossLine:=NotThisLine;
end;

[global]
function crossSegment(theBdy,theSeg:integer;theDv:dvPtr):integer;
var
  lineA,lineB: zNumber;
begin
  endpoints(theBdy,theSeg,lineA,lineB);
  CrossSegment:=crossLine(lineA,lineB,theDv^.link^.posi,theDv^.vortex.pos);
end;

[global]
function getCrossOver(var bdy,suspect:integer;theDv:dvPtr):boolean;
var
  lineAx,lineAy,lineBx,lineBy : double;
  count,thisMaxSeg,curSeg : integer;
  done: boolean;
begin
  with bodyAry[bdy]^ do
   begin
    thisMaxSeg:=round(mSeg/2+1);
    curSeg:=suspect;
    count:=1;
    repeat
     done:=(crossSegment(bdy,curSeg,theDv)=crossOver);
     if done then suspect:=curSeg
     else
      begin
       curSeg:=getSegmentTag(bdy,suspect+count);
       done:=(crossSegment(bdy,curSeg,theDv)=crossOver);
       if done then suspect:=curSeg
       else
        begin
         curSeg:=getSegmentTag(bdy,suspect-count);
         done:=(crossSegment(bdy,curSeg,theDv)=crossOver);
         if done then suspect:=curSeg
         else count:=count+1;
        end;
      end;
    until done or (count>thisMaxSeg);
   end;
   getCrossOver:=done;
end;

[global]
procedure endPoints (bdy,i:integer; var xy1,xy2: zNumber);
var
  upper: integer;
begin
  with bodyAry[bdy]^ do
   begin
    upper:=getSegmentTag(bdy,i+1);
    xy1.Re:=seg[i]^.datPt.Re;
    xy1.Im:=seg[i]^.datPt.Im;
    xy2.Re:=seg[upper]^.datPt.Re;
    xy2.Im:=seg[upper]^.datPt.Im;
   end;
end;

  {////////////////////////////////////////////////////////////////////////
    function getDist

    get the normal distance from the input point xi,yi to the input
    segment
   ////////////////////////////////////////////////////////////////////////}
[global]
function getDist(xyi:zNumber;var bdy,elem:integer):double;
var
  xe1,ye1,
  xe2,ye2,
  pla,nga,aba : double;
  upper       : integer;
begin
  with bodyAry[bdy]^ do
  begin
    upper:=getSegmentTag(bdy,elem+1);
    xe1:=seg[elem]^.datPt.Re;
    ye1:=seg[elem]^.datPt.Im;
    xe2:=seg[upper]^.datPt.Re;
    ye2:=seg[upper]^.datPt.Im;
    pla:=(xe1*ye2+xe2*xyi.Im+xyi.Re*ye1);
    nga:=(xe1*xyi.Im+xyi.Re*ye2+xe2*ye1);
    aba:=abs(pla-nga);
    getdist:=aba/seg[elem]^.ds;
  end;{with}
end;

[global]
function getGap2Seg(xyi:zNumber;var bdy,elem:integer;var theGap:double):double;
var
  xe1,ye1,xe2,ye2,pla,nga : double;
  upper   : integer;
begin
  with bodyAry[bdy]^ do
  begin
    upper:=getSegmentTag(bdy,elem+1);
    xe1:=seg[elem]^.datPt.Re;
    ye1:=seg[elem]^.datPt.Im;
    xe2:=seg[upper]^.datPt.Re;
    ye2:=seg[upper]^.datPt.Im;
    pla:=(xe1*ye2+xe2*xyi.Im+xyi.Re*ye1);
    nga:=(xe1*xyi.Im+xyi.Re*ye2+xe2*ye1);
    theGap:=abs(pla-nga)/seg[elem]^.ds;
    getGap2Seg:=theGap/seg[elem]^.ds;
  end;{with}
end;

  {/// procedure to get the closest element from the input point, and ///*
   /// find the corresponding NORMAL distance from that element.      ///*}
[global]
procedure getElemD(inpt:zNumber;bdy:integer;var which:integer;var distance:double);
var
  minr,rsq : double;
  i,minm   : integer;
begin
  minr:=1.0E+05;
  with bodyAry[bdy]^ do
  for i:=1 to mSeg do
   begin
    rsq:=sqr(inpt.Re-seg[i]^.pvt.Re)+sqr(inpt.Im-seg[i]^.pvt.Im);
    if rsq<minr then
     begin
      minr:=rsq;
      minm:=i;
     end;
   end;
  which:=minm;
  distance:=getdist(inpt,bdy,minm);
end;     {/// EOProc getelemd ///}

  {/////////////////////////////////////////////////////////////////////////
   procedure getElemR

   find the nearest element from the input point, and finds the
   corresponding distance.

   called by Proc. bodyUV
   /////////////////////////////////////////////////////////////////////////}
[global]
procedure getElemR(inpt:zNumber;bdy:integer;var which:integer;var rdist:double);
var
  minEst,rsq    : double;
  i,tmpSegEst   : integer;
begin
  minEst:=1.0E+05;
  with bodyAry[bdy]^ do
  begin
    for i:=1 to mSeg do
     begin
       rsq:=sqr(inpt.Re-seg[i]^.pvt.Re)+sqr(inpt.Im-seg[i]^.pvt.Im);
       if rsq<minEst then
        begin
          minEst:=rsq;
          tmpSegEst:=i;
        end;
     end;
    which:=tmpSegEst;
    rsq:= sqr(inpt.Re-seg[tmpSegEst]^.pvt.Re)
         +sqr(inpt.Im-seg[tmpSegEst]^.pvt.Im);
    rdist:=sqrt(rsq);
  end; {with}
end;

procedure Zconj(z1: zNumber; var z2: zNumber);
begin
  z2.Re:=z1.Re;
  z2.Im:=-1.0*z1.Im;
end;

[global]
procedure ZSum(z1,z2: zNumber; var z3: zNumber);
begin
  z3.Re:=z1.Re+z2.Re;
  z3.Im:=z1.Im+z2.Im;
end;

[global]
procedure ZProd(z1, z2: zNumber; var z3: zNumber);
{ z3:=z1*z2,  z3 must NOT be same variable as z1 and z2 }
begin
  z3.Re:=z1.Re*z2.Re - z1.Im*z2.Im;
  z3.Im:=z1.Re*z2.Im + z1.Im*z2.Re;
end;

[global]
procedure ZDiv(z1, z2: zNumber; var z3: zNumber);
{ z3:=z1/z2, z3 returns 0 + 0 i if z2=0 }
var
  oneOverz2mod: double;
begin
  if (z2.Re=0) and (z2.Im=0) then
   begin
     z3.Re:=0.0;z3.Im:=0.0;
   end
  else
   begin
     oneOverz2mod:=1.0/(sqr(z2.Re)+sqr(z2.Im));
     z3.Re:=(z1.Re*z2.Re + z1.Im*z2.Im)*oneOverz2mod;
     z3.Im:=(z1.Im*z2.Re - z1.Re*z2.Im)*oneOverz2mod;
   end;
end;

[global]
procedure zInv(z1: zNumber; var z2: zNumber);
var
  oneOverz2mod: double;
begin
  if (z1.Re=0) and (z1.Im=0) then
   begin
     z2.Re:=0.0;z2.Im:=0.0;
   end
  else
   begin
     oneOverz2mod:=1.0/(sqr(z1.Re)+sqr(z1.Im));
     z2.Re:=z1.Re*oneOverz2mod;
     z2.Im:=-z1.Im*oneOverz2mod;
   end;
end;

[global]
function factorial(seed: integer):double;
var dseed: double;
begin
  dseed:=seed*1.0;
  if seed<>0 then
   factorial:=dseed*factorial(seed-1)
  else factorial:=1.0;
end;

[global]
procedure splitString(var inString, outString: TString; var numPart:integer);
var
  extPos,value,tmpVal2,cc,i,j,errCode: integer;
  UPSTR,strAny,tmpStr2 : TString;

begin
  cc:=StrLength(inString);
  if cc>0 then
  begin
    UPSTR:='';
    for i:=1 to cc do UPSTR:=UPSTR+upperCase(inString[i]);
    extPos:=posStr('.DVR',UPSTR);
    if extPos>1 then
    begin
      i:=1;
      repeat
        j:=extpos-i;
        strAny:=copyStr(inString,j,i);
        valInt(strAny,value,errCode);
        if errCode=0 then
         begin
           i:=i+1;
           numPart:=value;
         end;
      until errCode<>0;
      tmpStr2:=copyStr(inString,1,j);
      outString:=tmpStr2;
    end
    else
    begin
      numPart:=0;
      outString:=inString;
    end;
  end;
end;

[global]
function DeNumberNACA4D(inValue:double;var camber,pValue,thick:double):boolean;
var
  cutter,aReal:double;
begin
  aReal:=trunc(inValue);
  cutter:=aReal*0.0100;
  thick:=round((cutter-trunc(cutter))*100.00);
  aReal:=trunc(cutter);
  cutter:=aReal*0.1000;
  pValue:=round((cutter-trunc(cutter))*10.00);
  aReal:=trunc(cutter);
  cutter:=aReal*0.1000;
  camber:=round((cutter-trunc(cutter))*10.00);
  DeNumberNACA4D:=trunc(cutter)=9;
end;

[global]
procedure swapPtr(var A, B: dvPtr);
var
  Temp : dvPtr;
begin
  Temp := A;
  A    := B;
  B    := Temp;
  Temp := nil;
end;

[global]
procedure swapInt(var A, B: integer);
var
  Temp : integer;
begin
  Temp := A;
  A    := B;
  B    := Temp;
end;

[global]
procedure swapReal(var A, B: double);
var
  Temp : double;
begin
  Temp := A;
  A    := B;
  B    := Temp;
end;

[global]
procedure swapBool(var A, B: Boolean);
var
  Temp : Boolean;
begin
  Temp := A;
  A    := B;
  B    := Temp;
end;

 {///////////////////////////////////////////////////////////////////////////
  The following two procedures xQSort, yQSort arrange DV list into ascending
  order of x-coordinate and of y-coordinate respectively. The indexed lists
  are stored as xOrder and yOrder respectively. They are used in vortex
  merging procedure to determine the nearby vortice to a particular vortex.
  \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\}

[global]
procedure xQSort (Var Values: ptrList; Start, EndList: integer);
var
  Low, High       : integer;
  ListSeparator  : double;

begin
  {write(' begin xQsort ');}
  Low :=Start;
  High :=EndList;
  ListSeparator := Values[(Start+EndList) Div 2]^.vortex.pos.Re;

  {/// Make the two lists ///}
  repeat
   while (Values[Low]^.vortex.pos.Re < ListSeparator) do
       Low := Low + 1;
   while (Values[High]^.vortex.pos.Re > ListSeparator) do
       High := High - 1;

   if (Low <= High) Then
    begin
     swapPtr(Values[Low], Values[High]);
     {switch(Values[Low], Values[High],sizeof(values[Low]));}
     Low := Low + 1;
     High := High - 1;
    end;
  until (Low > High);

  {/// Recursively sort the lower list ///}
  if (Start < High) then
    xQSort(Values, Start, High);

  {/// Recursively sort the higher list ///}
  if (Low < EndList) Then
    xQSort(Values, Low, EndList);
  {write(' end xQsort ');}
end;  {/// EOProc xQSort ///}

[global]
procedure yQSort (Var Values: ptrList; Start, EndList: Integer);
var
  Low, High      : Integer;
  ListSeparator : double;
begin
  {write(' begin yQsort ');}
  Low :=Start;
  High :=EndList;
  ListSeparator := Values[(Start+EndList) Div 2]^.vortex.pos.Im;

  {/// Make the two lists ///}
  repeat
   while (Values[Low]^.vortex.pos.Im < ListSeparator) do
       Low := Low + 1;
   while (Values[High]^.vortex.pos.Im > ListSeparator) do
       High := High - 1;

   if (Low <= High) Then
    begin
     swapPtr(Values[Low], Values[High]);
     {switch(Values[Low], Values[High],sizeof(values[Low]));}
     Low := Low + 1;
     High := High - 1;
    end;
  until (Low > High);

  {/// Recursively sort the lower list ///}
  if (Start < High) then
    yQSort(Values, Start, High);

  {/// Recursively sort the higher list ///}
  if (Low < EndList) Then
    yQSort(Values, Low, EndList);
  {write(' end yQsort ');}
end;   {/// EOProc yQSort ///}

{function minFromList(theVector:vectorM;endList:integer):double;
var
  i:integer;
  tempMin:double;
begin
  tempMin:=theVector[1];
  for i:=1 to endList do
   if theVector[i]<tempMin then tempMin:=theVector[i];
  minFromList:=tempMin;
end;}

[global]
function minTagOfList(theVector:vectorM;endList:integer):integer;
var
  minTag,i:integer;
  tempMin:double;
begin
  minTag:=1;
  tempMin:=theVector[minTag];
  for i:=1 to endList do
   if theVector[i]<tempMin then
    begin
     minTag:=i;
     tempMin:=theVector[i];
    end;
  minTagOfList:=minTag;
end;

[global]
function sameSign(realA,realB:double):boolean;
var
  prod : double;
begin
  prod:=realA*realB;
  if prod>=0.0 then sameSign:=true
  else sameSign:=false;
end;

{function diffSign(realA,realB:double):boolean;
var
  prod : double;
begin
  prod:=realA*realB;
  if prod<=0.0 then diffSign:=true
  else diffSign:=false;
end;}

procedure CtocStatistic(theStart:PCell);
var
  curNode,curRow: PCell;
  dvChain: Pchain;
  indv: integer;
begin
  curRow:=theStart;
  while curRow<>nil do
   begin
    curNode:=CurRow;
    while curNode<>nil do
     begin
      {/// survey only the inner-most cell ///}
      if (curNode^.dvInCell<>nil) then
       begin
        if (curNode^.NodeStart=nil) then
         begin
          indv:=0;
          dvChain:=curNode^.dvInCell;
          while dvChain<>nil do {with dvChain^.dvElem^.Vortex do}
           begin
            indv:=indv+1;
            dvChain:=dvChain^.nextElem;
           end;
          if indv<12 then dvBin[1]:=dvBin[1]+1
          else if indv<50 then dvBin[2]:=dvBin[2]+1
          else dvBin[3]:=dvBin[3]+1;
         end;
        CtocStatistic(curNode^.NodeStart);
       end;
      curNode:=curNode^.UpC;
     end;
    curRow:=curRow^.RtC;
  end;
end;

[global]
procedure getCtocStatistic;
var
  j,totalBin:integer;
  dump,xFactor,
  PropSmall,PropGood,PropBig:double;
begin
  for j:=1 to NumBin do dvBin[j]:=0;
  CtocStatistic(CtocStart);
  totalBin:=dvBin[1]+dvBin[2]+dvBin[3];
  PropSmall:=dvBin[1]/totalBin;
  PropGood:=dvBin[2]/totalBin;
  PropBig:=dvBin[3]/totalBin;

  if PropSmall>0.6 then
   begin
    if PropSmall<0.65 then xFactor:=1.05
     else if PropSmall<0.7 then xFactor:=1.2
      else xFactor:=1.35;
    reformCtoc:=true;
    basicMeshFactor:=basicMeshFactor*xFactor;
    {write(' mesh too small ');}
   end;
  if PropBig>0.35 then
   begin
    if PropBig<0.4 then xFactor:=0.95
     else if PropBig<0.5 then xFactor:=0.8
      else xFactor:=0.6;
    reformCtoc:=true;
    basicMeshFactor:=basicMeshFactor*xFactor;
    {write(' mesh too big ');}
   end;
  if PropGood>0.5 then write(' good mesh size ');
end;

{function NumActiveCtc(theStart:PCell):integer;
var
  curNode,curRow : PCell;
  mElem,dvChain : Pchain;
  i,indv,count,iCol,jRow : integer;
begin
  count:=0;
  curRow:=theStart;
  while curRow<>nil do
   begin
    curNode:=CurRow;
    while curNode<>nil do
     begin
      if curNode^.dvInCell<>nil then
      begin
        count:=count+1;
      end;
      curNode:=curNode^.UpC;
    end;
    curRow:=curRow^.RtC;
  end;
  numActiveCtc:=count;
end;}

[global]
procedure SearchHiLoDV(starter: dvPtr; var LLo,HHi: zNumber);
var
  curPtr : dvPtr;
begin
  curPtr:=Starter;
  if curPtr<>nil then begin
    LLo:=curPtr^.Vortex.pos;
    HHi:=curPtr^.Vortex.pos;
  end;
  while curPtr<>nil do with curPtr^.Vortex.pos do begin
    if Re<LLo.Re then LLo.Re:=Re
    else if Re>HHi.Re then HHi.Re:=Re;
    if Im<LLo.Im then LLo.Im:=Im
    else if Im>HHi.Im then HHi.Im:=Im;
    curPtr:=curPtr^.next;
  end;
end;

[global]
function ValidDv(inPutDv: dvPtr): boolean;
begin
  ValidDv:=true;
  with inPutDv^.Vortex do
  begin
   if mag=0.0 then ValidDV:=false
   else if pos.Re < LLFlowField.Re then ValidDV:=false
    else if pos.Re > URFlowField.Re then ValidDV:=false
     else if pos.Im < LLFlowField.Im then ValidDV:=false
      else if pos.Im > URFlowField.Im then ValidDV:=false;
  end;
end;

{[global]
function FileExist(FileName:TString):boolean;
var F: file;
begin
  Assign(F, fileName);
  FileMode := 0;
  Reset(F);
  Close(F);
  FileExist := (IOResult = 0) and (FileName <> '');
  FileMode := 2;
end;}

[global]
function bodyLapped:boolean;
label 1;
var
  p,k,j,i: integer;
  xdv,ydv,r,rsq,oneOver2Rsq,
  uInduce,vInduce,circul : double;

procedure subelement;
var
  isub,nsubs,upper                : integer;
  xo,yo,xn,yn,xsub,ysub,acu,acv   : double;
begin
  acu:=0.0;acv:=0.0;
  with bodyAry[i]^ do
  begin
  nsubs:=(1+round(seg[p]^.ds/(2*r)));
  if nsubs>10 then nsubs:=10;

  upper:=getSegmentTag(i,p+1);
  xo:=seg[p]^.datPt.Re;
  yo:=seg[p]^.datPt.Im;
  xn:=seg[upper]^.datPt.Re;
  yn:=seg[upper]^.datPt.Im;

  for isub:=1 to nsubs do
   begin
    xsub:=xo+(xn-xo)/nsubs*(isub-0.5);
    ysub:=yo+(yn-yo)/nsubs*(isub-0.5);
    rsq:=sqr(xdv-xsub)+sqr(ydv-ysub);
    oneOver2Rsq:=0.5/rsq;
    acu:=acu+(ysub-ydv)*oneOver2Rsq;
    acv:=acv+(xsub-xdv)*oneOver2Rsq;
   end;
  end;{with}
  uInduce:=acu/nsubs;vInduce:=-acv/nsubs;
end; {/// EOProc subelement ///}

begin
  if numActiveBbody>1 then
   begin
    j:=1;
    while j<numActiveBbody do
     begin
      i:=j+1;
      while i<=numActiveBbody do with bodyAry[i]^ do
       begin
        for k:=1 to bodyAry[j]^.mSeg do
         begin
          xdv:=bodyAry[j]^.seg[k]^.pvt.Re;
          ydv:=bodyAry[j]^.seg[k]^.pvt.Im;
          circul:=0.0;
          for p:=1 to mSeg do with seg[p]^ do
           begin
            rsq:=sqr(xdv-pvt.Re)+sqr(ydv-pvt.Im);
            r:=sqrt(rsq);
            if (r>0.1*ds) then
             begin
              oneOver2Rsq:=0.5/rsq;
              uInduce:=(pvt.Im-ydv)*oneOver2Rsq;
              vInduce:=-(pvt.Re-xdv)*oneOver2Rsq;
             end
            else subElement;
            circul:=circul+(uInduce*cosine+vInduce*sine)*ds/pi;
          end;{for p}
          if (circul>0.5000) then
           begin
            bodyLapped:=true;
            goto 1;
           end;
        end;{for k}
        i:=i+1;
      end;{while i}
      j:=j+1;
    end;{while j}
    bodyLapped:=false;
  end
  else bodyLapped:=false;
1:end;

[global]
function checkPointInBody(theChain:PlinkPt):boolean;
label 1;
var
  p,k,j,i: integer;
  xdv,ydv,r,rsq,oneOver2Rsq,
  uInduce,vInduce,circul : double;
  curLinkPt : PlinkPt;

procedure subelement;
var
  isub,nsubs,upper                : integer;
  xo,yo,xn,yn,xsub,ysub,acu,acv   : double;
begin
  acu:=0.0;acv:=0.0;
  with bodyAry[i]^ do
  begin
  nsubs:=(1+round(seg[p]^.ds/(2*r)));
  if nsubs>10 then nsubs:=10;

  upper:=getSegmentTag(i,p+1);
  xo:=seg[p]^.datPt.Re;
  yo:=seg[p]^.datPt.Im;
  xn:=seg[upper]^.datPt.Re;
  yn:=seg[upper]^.datPt.Im;

  for isub:=1 to nsubs do
   begin
    xsub:=xo+(xn-xo)/nsubs*(isub-0.5);
    ysub:=yo+(yn-yo)/nsubs*(isub-0.5);
    rsq:=sqr(xdv-xsub)+sqr(ydv-ysub);
    oneOver2Rsq:=0.5/rsq;
    acu:=acu+(ysub-ydv)*oneOver2Rsq;
    acv:=acv+(xsub-xdv)*oneOver2Rsq;
   end;
  end;{with}
  uInduce:=acu/nsubs;vInduce:=-acv/nsubs;
end; {/// EOProc subelement ///}

begin
  for i:=1 to numActiveBbody do with bodyAry[i]^ do
  begin
    curLinkPt:=theChain;
    while curLinkPt<>nil do
     begin
      xdv:=curLinkPt^.ptCoord.Re;
      ydv:=curLinkPt^.ptCoord.Im;
      circul:=0.0;
      for p:=1 to mSeg do with seg[p]^ do
       begin
         rsq:=sqr(xdv-pvt.Re)+sqr(ydv-pvt.Im);
         r:=sqrt(rsq);
         if (r>0.1*ds) then
          begin
           oneOver2Rsq:=0.5/rsq;
           uInduce:=(pvt.Im-ydv)*oneOver2Rsq;
           vInduce:=-(pvt.Re-xdv)*oneOver2Rsq;
          end
         else subElement;
         circul:=circul+(uInduce*cosine+vInduce*sine)*ds/pi;
       end;{for p}
      if (circul>0.5000) then
       begin
        checkPointInBody:=true;
        goto 1;
       end;
     curLinkPt:=curLinkPt^.downLink;
     end; { while curLinkPt }
   end; { for i }
  checkPointInBody:=false;
1:end;

[global]
function PtInOneBody( thePoint:zNumber; abody:integer;
                      doCheck:boolean):boolean;
{  this procedure checks if a point "thePoint" lies within the contour of a
   given body "abody",
   "doCheck" check if the point lies a bit outside the contour.
   USE :
   for normal use set doCheck=noCheck; then only points inside contour will
   render PtInOneBody true;
   for check mesh points set doCheck=yesCheck, then those points
   which are a bit outside the contour (with abs(circul)>1.0E-03) will also
   render PtInOneBody true
}
label 1;

var
  i: integer;
  Xmj,Ymj,r,rsq,oneOver2Rsq,
  uInduce,vInduce,circul: double;

procedure subelement;
var
  isub,nsubs          : integer;
  xyo,xyn : zNumber;
  xsub,ysub,acu,acv   : double;
begin
  acu:=0.0;acv:=0.0;
  with bodyAry[aBody]^ do begin
   if r>0.0 then begin
    nsubs:=12*(1+round(2*seg[i]^.ds/r));
    if abs(nsubs)>50 then nsubs:=50;
   end
   else nsubs:=50;
   endpoints(aBody,i,xyo,xyn);
   for isub:=1 to nsubs do begin
    xsub:=xyo.Re+(xyn.Re-xyo.Re)/nSubs*(isub-0.5);
    ysub:=xyo.Im+(xyn.Im-xyo.Im)/nSubs*(isub-0.5);
    Xmj:=xsub-thePoint.Re;
    Ymj:=ysub-thepoint.Im;
    rsq:=sqr(Xmj)+sqr(Ymj);
    if rsq>cutOffsq then begin
     oneOver2Rsq:=0.50/rsq;
     acu:=acu+Ymj*oneOver2Rsq;
     acv:=acv+Xmj*oneOver2Rsq;
    end
    else begin
     acu:=acu+Ymj*oneOver2CSq;
     acv:=acv+Xmj*oneOver2CSq;
    end;
   end;
  end;{with}
  uInduce:=acu/nSubs;
  vInduce:=-acv/nSubs;
end; {/// Proc subelement ///}

begin
  with bodyAry[aBody]^ do begin
   circul:=0.0;             { reset induced circulation }
   for i:=1 to mSeg do begin
    Xmj:=seg[i]^.pvt.Re-thePoint.Re;
    Ymj:=seg[i]^.pvt.Im-thePoint.Im;
    rsq:=sqr(Xmj)+sqr(Ymj);
    if rsq<1e-10 then begin
     circul:=0.5;
     goto 1;
    end;

    r:=sqrt(rsq);
    {/// record the 'nearest' and 'nearer' segment ///}
    if (r<2.0*seg[i]^.ds) then subelement
    else begin
     oneOver2Rsq:=0.50/rsq;
     uInduce:=Ymj*oneOver2Rsq;
     vInduce:=-Xmj*oneOver2Rsq
    end;
     circul:=circul+(uInduce*seg[i]^.cosine+vInduce*seg[i]^.sine)*seg[i]^.ds;
   end;{for}
   circul:=circul/pi
  end; { while/with }
1:if doCheck then
   PtInOneBody:=(circul>0.500) or (abs(circul)>1.0E-4)
  else
   PtInOneBody:=(circul>0.500);
end;

[global]
function IsPtInBody(thePoint:zNumber;doCheck:boolean):boolean;
label 1;
var
  i: integer;
begin
  for i:=1 to numActiveBbody do with bodyAry[i]^ do begin
   if PtInOneBody(thePoint,i,doCheck) then begin
    IsPtInBody:=true;
    goto 1;
   end;
  end;{ for i}
  IsPtInBody:=false;
1:end;

[global]
function chkCellinBody(aCell:PCell):integer;
var
  chkPt:zNumber;
  i: integer;
begin
  with aCell^ do
  begin
   chkPt.Re:=Ctr.Re-PData^.xWidth*0.5;
   chkPt.Im:=ctr.Im-PData^.yHeight*0.5;
   if IsPtInBody(chkPt,yesCheck) then chkCellinBody:=cellNotClear
   else
    begin
     chkPt.Re:=chkPt.Re+PData^.xWidth;
     if IsPtInBody(chkPt,yesCheck) then chkCellinBody:=cellNotClear
     else
     begin
      chkPt.Im:=chkPt.Im+PData^.yHeight;
      if IsPtInBody(chkPt,yesCheck) then chkCellinBody:=cellNotClear
      else
      begin
       chkPt.Re:=chkPt.Re-PData^.xWidth;
       if IsPtInBody(chkPt,yesCheck) then chkCellinBody:=cellNotClear
       else chkCellinBody:=CellClear;
      end;
     end;
    end;
  end; {with}
end;

[global]
procedure publishCircStatus;
var
  bodyAbsorb,
  totalSum,
  sumDVDV  : double;
  curPtr   : dvPtr;
  bdy,m      : integer;
begin
  totalSum:=0.0;
  sumDVDV:=0.0;
  curPtr:=dvStart;
  while curPtr<>nil do
   begin
    sumDVDV:=sumDVDV+vortDecay(curPtr);{^.vortex.mag;}
    curPtr:=curPtr^.next;
   end;
  {for bdy:=1 to numActiveBbody do with BodyAry[bdy]^ do
   begin
    bodyAbsorb:=0.0;
    for m:=1 to mSeg do
     bodyAbsorb:=bodyAbsorb+dvAbsorbed[m]*seg[m]^.ds;
    writeln(' body ',bdy,' absorbed  : ',bodyAbsorb);
    totalSum:=totalSum+bodyAbsorb;
   end;}
  totalSum:=totalSum+sumDVDV+snuff;
  writeln(' total active dv strength  : ',sumDVDV);
  writeln(' total snuffed dv strength : ',snuff);
  writeln(' circulation in flow field : ',totalSum);
end;

[global]
procedure InitDvVelocity;
var
  mPtr: dvPtr;
begin
  {/// initialise velocity of d.v. to zero ///}
  mPtr:=dvStart;
  while mPtr<>nil do
   with mPtr^.vortex.vel do
   begin
     Re:=0.00;Im:=0.00;
     mPtr:=mPtr^.next;
   end;
end;

  {////////////////////////////////////////////////////////////////////////
    procedure relocate

    should a vortex drifts into body, it is to be displaced outwards along
    a normal to the surface.

    called by : getDvRhs;     randomWalk;     xConvect;
   ///////////////////////////////////////////////////////////////////////}

procedure relocate(inDvPtr: dvPtr; inputLog: dvLogRec);
var
  oldX,oldY,origDist,
  outNormal,dispmt,radisp: double;
begin
  oldX:=inDvPtr^.Vortex.pos.Re;
  oldY:=inDvPtr^.Vortex.pos.Im;
  with inputLog do begin
    if (inDvPtr^.link=nil) then dispmt:=2.0*nearDist
    else begin
      origDist:=getDist(inDvPtr^.link^.posi,nearBody,SegEst);
      dispmt:=2.0*(nearDist+origDist);
      radisp:=sqrt(sqr(inDvPtr^.Vortex.pos.Re-inDvPtr^.link^.posi.Re)+
                   sqr(inDvPtr^.Vortex.pos.Im-inDvPtr^.link^.posi.Im));
      if dispmt<radisp then dispmt:=radisp;
    end;
      {/// move d.v. out in outward normal direction ///}
    outNormal:=bodyAry[nearBody]^.seg[inputLog.SegEst]^.slope+0.5*pi;
    inDvPtr^.Vortex.pos.Re:=oldX+dispmt*cos(outNormal);
    inDvPtr^.Vortex.pos.Im:=oldY+dispmt*sin(outNormal);
  end;{with}
end;

[global]
Procedure xyuvSum(Point:zNumber;var inVel:zNumber);
var
  {sinCore,cosCore,}
  theMag,
  Xjm,Yjm,oneOver2Rsq,rsq  : double;
  curPtr                   : dvPtr;
begin {/// SOProc xyuv ///}
  {/// vel must be set zero before calling this procedure ///}
  {/// evaluate vortex-induced velocities ///}
  curPtr:=dvStart;
  while curPtr<>nil do begin
   theMag:=vortDecay(curPtr);
   Xjm:=Point.Re-curPtr^.Vortex.pos.Re;
   Yjm:=Point.Im-curPtr^.Vortex.pos.Im;
   rsq:=sqr(Xjm)+sqr(Yjm);
   if rsq>=cutOffsq then begin
    oneOver2Rsq:=0.5/(pi*rsq);
    inVel.Re:=inVel.Re+theMag*Yjm*oneOver2Rsq;
    inVel.Im:=inVel.Im-theMag*Xjm*oneOver2Rsq;
   end
   else begin
    inVel.Re:=inVel.Re+theMag*Yjm*oneOver2piCSq;
    inVel.Im:=inVel.Im-theMag*Xjm*oneOver2piCSq;

    {sinCore:=oneOver2piC*Yjm/sqrt(rsq);
    cosCore:=oneOver2piC*Xjm/sqrt(rsq);
    inVel.Re:=inVel.Re+theMag*sinCore;
    inVel.Im:=inVel.Im-theMag*cosCore;}
   end;
   curPtr:=curPtr^.next;
  end; {while/with}
end; {/// Proc xyuv ///}

  {////////////////////////////////////////////////////////////////////////
    Procedure getOneInduced

    (i)  calculates the induced surface velocity by a given UNIT discrete
         vortex,

    (ii) determines if the vortex concerned goes inside/too-close to body.
         If this happen, the dv concerned is temporary absorbed to the
         closest segment and will be regenerated in the next time step;

    calling procedure : getDvRhs

    Nomenclature :-

    unitRhs   : induced surface vorticity on every segment by a unit vortex
    circul    : induced circulation on body surface by a unit vortex
   ////////////////////////////////////////////////////////////////////////}

[global]
procedure getOneInduced(inDvPtr: dvPtr);
var
  redo: boolean;
  logBk: dvLogRec;
  bodyEst,i,bdy,tmpSegEst,bodySeg: integer;
  Xmj,Ymj,xdv,ydv,{dvMag,}r,rsq,oneOver2Rsq,
  uInduce,vInduce,circul,minEst: double;
  effCore,dmpDist : double;

procedure subelement;
var
  isub,nsubs          : integer;
  {xo,yo,xn,yn,}
  invNsubs,xsub,ysub,acu,acv   : double;
  xyo,xyn: zNumber;
begin
  acu:=0.0;acv:=0.0;
  with bodyAry[bdy]^ do
  begin
  nsubs:=12*(1+round(2*seg[i]^.ds/r));
  if nsubs>50 then nsubs:=50;
  invNsubs:=1.0/nsubs;
  endpoints(bdy,i,xyo,xyn);
  for isub:=1 to nsubs do
   begin
    xsub:=xyo.Re+(xyn.Re-xyo.Re)*InvNsubs*(isub-0.5);
    ysub:=xyo.Im+(xyn.Im-xyo.Im)*InvNsubs*(isub-0.5);
    Xmj:=xsub-xdv;
    Ymj:=ysub-ydv;
    rsq:=sqr(Xmj)+sqr(Ymj);
    if rsq>cutOffsq then
     begin
      oneOver2Rsq:=0.50/rsq;
      acu:=acu+Ymj*oneOver2Rsq;
      acv:=acv+Xmj*oneOver2Rsq;
     end
    else
     begin
      acu:=acu+Ymj*oneOver2CSq;
      acv:=acv+Xmj*oneOver2CSq;
     end;
   end;
  end;{with}
  uInduce:=acu*InvNsubs;vInduce:=-acv*InvNsubs;
end; {/// Proc subelement ///}

{procedure suppressDV;
var
  SvAdded: double;
begin
  with bodyAry[bdy]^ do
  begin
    getCrossOver(bodyEst,tmpSegEst,inDvPtr);
    SvAdded:=dvMag/seg[tmpSegEst]^.ds;
    dvAbsorbed[tmpSegEst]:=dvAbsorbed[tmpSegEst]+SvAdded;
    ndvAbsorbed:=ndvAbsorbed+1;
    inDvPtr^.Vortex.mag:=0.0;
  end;
  someGone:=true;
end;}

begin {/// Proc getOneInduced ///}
  bdy:=1;
  logBk.Stillwell:=true;
  while (bdy<=numActiveBbody) and logBk.stillwell
    do with bodyAry[bdy]^ do  begin
    minEst:=1.0E+05;         { search for closest segment for eac body }
    circul:=0.0;             { reset induced circulation }
    xdv:=inDvPtr^.Vortex.pos.Re;  { assign dv location inside 'while' loop }
    ydv:=inDvPtr^.Vortex.pos.Im;  { as values may change due to relocation }
    {dvMag:=inDvPtr^.Vortex.mag;}
    {/// effect vorticity decay ///}
    {dvMag:=vortDecay(inDvPtr);}

    redo:=false;

    for i:=1 to mSeg do begin
     Xmj:=seg[i]^.pvt.Re-xdv;
     Ymj:=seg[i]^.pvt.Im-ydv;
     rsq:=sqr(Xmj)+sqr(Ymj);
     r:=sqrt(rsq);
     if r<minEst then begin
      minEst:=r;
      bodyEst:=bdy;
      tmpSegEst:=i;
     end;

     if (r<2.0*seg[i]^.ds) then subelement
     else begin
      oneOver2Rsq:=0.50/rsq;
      uInduce:=Ymj*oneOver2Rsq;
      vInduce:=-Xmj*oneOver2Rsq;
     end;

     dvRhs[i]:=(uInduce*seg[i]^.cosine+vInduce*seg[i]^.sine)/pi;
     circul:=circul+dvRhs[i]*seg[i]^.ds;
    end;{for}

    logBk.nearDist:=minEst;
    if MinEst>bufferRad then begin
      inDvPtr^.Vortex.nearBdy:=0;
      inDvPtr^.Vortex.nearSeg:=0;
    end
    else begin
    if (MinEst<0.98*bufferRad) then begin
      inDvPtr^.Vortex.nearBdy:=bdy;
      inDvPtr^.Vortex.nearSeg:=tmpSegEst;
    end
    else begin
      inDvPtr^.Vortex.nearBdy:=0;
      inDvPtr^.Vortex.nearSeg:=0;
    end;
    end;

    logBk.IsInside:=(circul>0.5000) and (MinEst<2.0*bufferRad);
    if logBk.IsInside then begin
      getCrossOver(bodyEst,tmpSegEst,inDvPtr);
      logBk.nearDist:=getDist(inDvPtr^.Vortex.pos,bdy,tmpSegEst);
     end;
    logBk.segEst:=tmpSegEst;
    logBk.nearBody:=bodyEst;

    {//// determine if the current vortex is active or not ////}
    {if (not LogBk.IsInside) then LogBk.stillWell:=true
     else redo:=true;}

    if LogBk.IsInside then redo:=true
     else LogBk.stillWell:=true;

    if redo then begin
       relocate(inDvPtr,LogBk);
       bdy:=1;
      end
     else begin{ if stillwell and not Redo }
       if abs(circul)>zeroLimit then
        dvRhs[tmpSegEst]:=-Circul/seg[tmpSegEst]^.ds+dvRhs[tmpSegEst];
       bdy:=bdy+1;
     end;
  end; { while/with }
end;

[global]
function VortDecay(inDv:dvPtr): double;
begin
  with inDv^.vortex do
   VortDecay:=mag*exp(decayFactor*(elapsed-tGen));
end;

[global]
procedure getFluctuateFS;
var
  XRandom,YRandom : double;
  i,j             : integer;
begin
  XRandom:=KnuthRan;
  YRandom:=KnuthRan;
  uxCurt:=UxMean+fluctLevel*(-1.0+2.0*XRandom);
  vyCurt:=VyMean+fluctLevel*(-1.0+2.0*YRandom);
  for j:=1 to numActiveBbody do with bodyAry[j]^ do
   for i:=1 to mSeg do
    fsRhs[i]:=-(uxCurt*seg[i]^.cosine+vyCurt*seg[i]^.sine);
end;

[global]
procedure multiPotAngle( inpt1,inpt2:zNumber;var angQ : quartet);
var
  i:integer;
  ang:double;
begin
  ang:=getAngleR(inpt1,inpt2);
  angQ[1]:=ang;
  angQ[2]:=ang-0.5*pi;
  angQ[3]:=ang+pi;
  angQ[4]:=ang+0.5*pi;
  for i:=1 to 4 do
   if angQ[i]<0.0 then angQ[i]:=angQ[i]+2.0*pi;
end;

[global]
function  match4(numAry: quartet): double;
var
  i : integer;
  reject,sx,sxx : double;
  qmean,qstdd,qvar : double;
begin
  reject:=0.0;
  sx:=0.0;
  sxx:=0.0;
  for i:=1 to 4 do begin
   sx:=sx+numAry[i];
   sxx:=sxx+sqr(numAry[i]);
  end;
  qmean:=0.25*sx;
  qvar:=0.25*(sxx-0.25*sqr(sx));
  if qvar>0.0 then begin
   qstdd:=sqrt(qvar);
   for i:=1 to 4 do
    if abs(numAry[i]-qmean)>qstdd then reject:=numAry[i];
   if reject<>0.0 then match4:=(sx-reject)/3.0
   else match4:=qmean;
  end
  else match4:=qmean;
end;

to begin do
  begin
   KRandSeed:=0;
   K2BRandSeed:=0;
   PortRandSeed:=0;
   KSeed:=0;
   K2BSeed:=0;
   PortSeed:=0;
  end;
End.

