[environment ('vglobal')]
Module vGlobal;

const
  pi	    = 3.141592658;
  maxSeg    = 300;      { 101 max number of segment on each body }
  maxDv	    = 2000;     { 2000 max number of discrete point vortices }
  maxLine   = 20;       { max number of streamline }
  maxStrmPt = 3000;   { 1000 max number of point on each streamline }
  slantXY   = 0.2;      { 'width'/'height' ratio to define a slant body }
  ZeroLimit = 1.0E-9; { Limit for zero induced circulation }
  tooSmall  = 1.0E-06;
  tooBig    = 1.0E+06;
  maxBody   = 5;  { max number of body in one configuration }
  maxMutual = maxBody-1;  { mutual interacting bodies must be maxBody-1}

  maxCtocData = 25;   { matrix size of storing c-to-c method cell data }
  errlimit = 0.02;    { error limit for cell-to-cell algorithm }
  ctcOrderLimit = 11; { 11 for errlimit=0.02, 15 for errlimit=0.005 & 0.01 }
  maxCellevel=3;
  NumSubCell = 4;
  minDvPerCell = 20;
  CellUnknown = 0;
  CellClear = 13;
  CellNotClear = 17;

  AllSmpBin     = 3;
  vobin         = 1;
  pobin         = 2;
  psbin         = 3;

  mergStep      = 3;
  mergReform    = 15;  { merReform = multiple of mergStep }


  modeChangeOver=1000;
  NACA4D = 9;         { GeoShape Code for NACA 4-digit aerofoil }
  userDefine = 6;
  noError = 0;
  calPotential = 1;
  calViscousFlow = 2;
  calFreshRun  = 3;
  calRestart   = 4;
  calStreamline = 5;
  overLapping  = 1;

  invalidTimeStep = 11;
  nullActiveTracer = 12;
  dvtracerNotMatch = 13;
  forceFileAlreadyExist = 14;
  forceFileNotFound = 15;
  bodyNotAtColl = 16;
  nullNdv = 17;

  directSum = 0;
  VortexInCell = 1;
  CellToCell = 2;

  noCheck = false;
  yesCheck = true;
  {calculateMode  : integer = 0;
  vortexFlowMode : integer = 0;
  DataError      : integer = noError;
  vortexMode     : integer = directSum;
  basicMeshFactor: double=25.0;
  xIteration     : integer=2;
  DomainX        : double=25;
  DomainY        : double=6;
  decayFactor    : double=-0.03;
  numSubLevel    : integer=0;}

type
  {string definition for DEC pascal compiler}
  TString = string(72);
  LongInt = integer32;
  ArrayYear = array[1..12] of integer;
  zNumber = record       { size 16 }
   Re,Im: double;
  end;

  quartet  = array [1..4] of double;

  PVector = ^TVector;
  TVector = record                  { define column matrix }
    RealNum : double;               { value of entry       }
    nextNum : PVector;              { point to next entry  }
  end;

  PlinkPt = ^TlinkPt;
  TlinkPt = record                  { linked point data }
    ptCoord : zNumber;              { complex coord of point }
    downLink : PlinkPt;             { point to next point }
  end;

  PpointArry = ^TpointArry;
  TpointArry = record               { a list of linked point }
    PtRuler,
    LkPointFirst,
    LkPointLast  : PlinkPt;
    prevIndex,
    nextIndex    : PpointArry;
  end;

  PFcEntry = ^TFcEntry;
  TFcEntry = record
    Cdrag,
    CLift    : double;
    nextFce  : PFcEntry;
  end;

  PfcList = ^TfcList;
  TfcList = record
    timeDue  : double;
    fcEntry  : PfcEntry;
    upFcList,
    dnFcList : PfcList;
  end;

  PProbeData = ^TProbeData;
  TProbeData = record
    PTotal,PStatic : double;
    LocalVel       : zNumber;
    nextSet        : PProbeData;
  end;

  PProbeList = ^TProbeList;
  TProbeList = record
    sampleTime : double;
    probeData  : PProbeData;
    upPbList,
    dnPbList   : PProbeList;
  end;

  PvelRec  = ^TvelRec;
  TvelRec  = record                 { define velocity record of vortex }
    veli,posi:znumber;
  end;

  TVortex = record
    mag,tGen:double;
    pos,vel:zNumber;
    nearBdy,nearSeg  : integer;
  end;

  PCell = ^TCell;

  PzMoment = ^TzMoment;
  TzMoment = record
   ctcPter : PCell;                  { cell to which dv belongs }
   zzj     : array[1..ctcOrderLimit] of zNumber; { moment of dv about cell center }
  end;

  PzCoord = ^TzCoord;    { size 4 }
  TzCoord = record       { size 74 }
   OrderTerm : integer;
   velCoeff  : zNumber;
   IZZab     : array [1..ctcOrderLimit] of zNumber;
   next      : PzCoord;
  end;

  PctcDataBank = ^ctcDataBank;
  ctcDataBank = array[-maxCtocData..maxCtocData,-maxCtocData..maxCtocData] of PzCoord;  { size 40000 }

  PWeight = ^TWeight;
  TWeight = record
   WeighLL,WeighLR,              { weighing factors }
   WeighUL,WeighUR   : double;
   VicMesh           : PCell;
  end;

  dvPtr = ^dvRec;
  dvRec = record
    Vortex           : TVortex;  { data of vortex }
    next,last        : dvPtr;    { pointers of doubly linked list }
    link             : PvelRec;  { point to velocity record }
    zMoment          : array[0..maxCellevel] of PzMoment; { point to cell-to-cell data }
  end;

  Pchain = ^Tchain;
  Tchain = record      { single linked pointer list for dvPtr }
   dvElem : dvPtr;
   nextElem : Pchain;  { point to next element down the chain }
  end;

  PVinc = ^TVinc;
  TVinc = record
   Vortcic,UXcic,VYcic : double;  { vorticity in cell, x-vel,y-vel of cell }
   NodeMmt             : integer     { array index for Victc method }
  end;

  PCtoc = ^TCtoc;   { Tcell define later }
  TCtoc = record        { cell-to-cell method record }
   Zm0      : double;   { sum zero moment in cell }
   Zmmt     : array[1..ctcOrderLimit] of zNumber;
  end;

  PGridData=^TGridData; {// contains details of cell-to-cell calculation //}
  TGridData=record
   level: integer;
   xWidth,yHeight,InvWidth,InvHeight: double;
   DataBank: PctcDataBank;
  end;

  TCell = record                 { basic cell data  }
   Ctr             : zNumber;    { coord of cell center }
   Clear,Zdv       : integer;
   dvInCell        : PChain;     { chain of dvPtr in cell }
   NodeStart,
   NodeEnd         : PCell;
   UpC,DnC,LfC,RtC : PCell;      { pointer to neighbouring cells }
   CtoC            : PCtoc;      { pointer to cell-to-cell data, if any }
   PData           : PGridData;
  end;

  PSmpCell =^TSmpCell;
  TSmpCell = record              { simple cell for data sampling }
   Ctr               : zNumber;  { coord of cell center }
   Nth,Sth,East,West : PSmpCell; { pointer to neighbouring cells }
   sample            : array [1..AllSmpBin] of double;   { sampling date storage }
   freePt            : boolean;
  end;

  geoPtr  = ^geoRec;
  geoRec  = record
   datPt,                        { data point }
   pvt,                          { collocation point }
   CenRad          : zNumber;
   selfCoup,
   radOfCurv,                    { radius/center of curvature }
   ds,slope,                     { length/ slope of segment }
   sine,cosine     : double;     { sine /cosine of segment }
   IsSharp,
   IsConvex        : boolean;    { define sharp edge/convex curvature }
  end;

  BbodyRec = record
    ShapeCode            : integer;   { 9 for NACA 4-digit aerofoil }
    BodyCX,BodyCY,
    BodyTilt,
    xDim,                          { Chord }
    yDim,                          { 9xxxx NACA 4-digit code }
    zDim                 : double;
    Nx,Ny,Nnose          : integer;
    BoundCirc            : double;
    UserInputFile        : TString;
  end;

  timeRec = record
    flowRey,
    flowAngle,
    tStart,
    tLast,
    tStep       : double;
  end;

  PCoeffRow = ^TCoeffRow;
  TCoeffRow = record              { keeps one row of matrix }
    RowStart: PVector;            { point to first entry of a row }
    upRow,                        { point to first entry of next upper row }
    downRow : PCoeffRow;          { point to frist entry of next lower row }
  end;

  ptrList = array[1..maxDv] of dvPtr;  { size 40000 }
  geoList = array[1..maxSeg] of geoPtr;
  vectorM = array[1..maxSeg] of double;
  vectorSS = array[1..maxStrmPt] of zNumber;
  bodyList = array[1..maxBody] of BbodyRec;
  ctcCoeff = array[0..ctcOrderLimit,0..ctcOrderLimit] of double;
  ctcArray = array[0..ctcOrderLimit] of double;

  PSortList = ^TSortList;
  TSortList = record
   theList   : ptrList;
  end;

  PoneLine = ^ToneLine;
  ToneLine = record
   numPoint : integer;
   theLine  : vectorSS;
   nextLine : PoneLine;
  end;

  TBodyPressure = record  { record the pressure fluctuation on segment }
    NumRecord : integer;
    MaxPress,
    MinPress,
    AvgPress  : vectorM;
  end;

  PBodyData = ^TBodyData;
  TBodyData = record
    Ubody,Vbody,AngV,xRotate,yRotate,
    PrescribeCirc,ctrlVolCirc,dsAvg,dsSum : double;
    AeroFoil,slant                        : boolean;
    SegTe,{lastStag,}frontStag,mSeg         : integer;
    seg         : geoList;      { array of pointer for segment data }
    PonBody     : TBodyPressure;
    {SolidVel,
    staticVort,}
    survor,
    fsRhs,
    dvrhs,
    fullRhs,
    {dvAbsorbed,}
    { adhoc is used in Gridttt for calculating the dvRhs by CTC }
    adhoc    : vectorM;
    selfMatx : PCoeffRow;
    mutlMatx : array[1..maxMutual] of PCoeffRow;
  end;

var
  numProbePoint,numStrmLines,
  matrixSize,NumActiveBbody                      : integer;
  fceListLength{,ndvAbsorbed}                    : integer;
  istep,isave,NumTimeStep,saveFileInt            : integer;
  nPzCoordTrash,nTrash,nPCTrash,nPCellTrash,
  ndv,ReyNum,tFirst,tLast,dt,snuff,alpha,elapsed : double;
  FluctLevel                                     : double;
  strmRlimit,strmLlimit,strmTime                 : double;
  probeOn,probeOff                               : double;

  sumActiveDv,totalBoundCirc                     : double;
  BodyLowerX,BodyLowerY,Xspan,Yspan,
  bufferRad,coreRad,alphaRad,Rcri{,magCri}       : double;
  UxMean,VyMean,UxCurt,VyCurt                    : double;
  oneOver2piC,cutOff,cutOffSq,oneOver2CSq,
  oneOver2piCSq,shedOffset,rwCoeffND             : double;
  LLFlowField,URFlowField                        : zNumber;

  lapped,ReformCtoC,allGone,{snuffActive,}
  {bfBdyActive,offsetActive,}newForceFile,
  firstAccess,VortexStrmLine,tracerActive,
  ReUseMem,Publish,fluctFlow,reformMerg          : boolean;

  ActiveBbody                                    : bodyList;
  Trash,dvStart,dvEnd                            : dvPtr;
  PCellTrash,
  CtocStart,CtocEnd,MCStart,MCEnd                : PCell;
  DynaMatx                                       : PCoeffRow;
  globalRhs                                      : PVector;
  fcListStart,fcListEnd                          : PfcList;
  probeStart,probeEnd                            : PProbeList;
  ChainTrash,dvOffMerg                           : Pchain;
  xOrder,yOrder                                  : PSortList;
  probePoint,StrmStartPts                        : PlinkPt;
  SmpStart,SmpEnd                                : PSmpCell;
  PzCoordTrash                                   : PzCoord;

  { file related }
  WriteSLName,WriteProbeName,
  WriteForceName,ReadForceName                   : TString;
  pathToReStartFile,pathToDVfile                 : TString;

  AlarmDate,fileGrp                              : TString;
  BodyAry                                        : array[1..maxBody] of PBodyData;
  thisTotal                                      : array[1..maxBody] of double;

  {utility}
  doProbing,checkTimeOn                          : boolean;
  PortRandSeed,KRandSeed                         : longInt;
  K2BRandSeed                                    : integer;
  calculateMode  : integer;
  vortexFlowMode : integer;
  DataError      : integer;
  vortexMode     : integer;
  basicMeshFactor: double;
  xIteration     : integer;
  DomainX        : double;
  DomainY        : double;
  decayFactor    : double;
  numSubLevel    : integer;
  chaindump,chainReUse,
  newzMmtMem,newCtocMem,
  newDvMem,newLinkMem,newDataBankMem,newCGroupMem,
  newCellMem,newChainMem,newzCoordMem : integer;

{gridttt}
procedure QuitGridTTT; external;
procedure assignDVtoCtoc; external;
procedure evaluateCtocUV(CellGroup:PCell); external;
function aPtInCtocMesh(thePt:zNumber;theStart:PCell):PCell; external;
procedure RoundtheWorld(GroupStart:PCell); external;
{procedure formMergMesh; external;}
procedure AssignMergCell;external;
procedure initGridData; external;
procedure CtocXYUV( GroupStart:PCell;
                    aPt:zNumber;var inducedVel:zNumber); external;
procedure easyAssign; external;
Procedure formSmpCell(var instart,inend :PsmpCell;
                          LoBound,UpBound :zNumber;
                          gridSize:double); external;
procedure CheckSmpCell(theStart:PSmpCell); external;
procedure SmearDVonSmp( whichSmp:PSmpCell;
                        celength:double; bin:integer);external;
{procedure saveSample( theStart:PSmpCell;bin,avgOf:integer;
                      filename:Tstring);external;}
{merging}
procedure DVMerging; external;
{ptvortex}
procedure ViscousFlow; external;
procedure PotentialFlow; external;
procedure FreshStart; external;
procedure Restart; external;
procedure xyuvQuick( XYpt:zNumber; var indVel:zNumber;
                     var insideBdy: integer; doCheck:boolean); external;
procedure xystm( xypt:zNumber; var stmValue:double;
                var insideBdy:integer; doCheck:boolean);external;
procedure QuickPitot ( thePoint:zNumber;var totalP,staticP: double;
                       var indVel: zNumber); external;

{added on 5 Apr for xsample.p}
procedure getGeneralVel; external;
procedure xconvect; external;
procedure randomwalk; external;
procedure createVortexSheet; external;
procedure shedDv; external;

{manager}
function GoEasyRun:boolean; external;
function prepareData:boolean; external;
{mathttt}
procedure solution; external;
procedure EnforceKuttaCondition; external;
procedure copyMatx(souMatx:PcoeffRow;var tarMatx:PcoeffRow); external;
{dpmifile}
procedure SaveResult(fileName: TString); external;
procedure ReadDvforRestart (fileName: TString; var inStart, inEnd: dvPtr;
                    var itemRead:double;var inTime: timeRec); external;
procedure freeDvMemory (var inStart,inEnd : dvPtr); external;
procedure getVectorMemory(var theVector: PVector; theSize:integer); external;
procedure freeVectorMemory(var theVector:PVector); external;
procedure freeMatrixMemory(var theMatrix:PCoeffRow); external;
procedure DvAppend (var ListStart, ListEnd: dvPtr; NewRec: TVortex;
                    var listLength: double); external;
procedure DvPack (var ListStart, ListEnd: dvPtr;var ListEmpty: boolean); external;

procedure FceListAppend (var gbFceStart,gbFceEnd: PFcList;
                         ListBegin: PFcEntry; inTime: double); external;

procedure FcEntryAppend (var oneListStart: PfcEntry; newDrag,newLift: double); external;
procedure ReadForceList; external;
procedure AppendForceList; external;
procedure saveCP(filename:Tstring); external;
procedure freeForceListMemory; external;

procedure freeProbeMemory; external;
procedure probeListAppend (var theStart,theEnd:PProbeList;firstSet:PProbeData;
                          inTime:double); external;
procedure probeAppend (var firstSet:PProbeData;newTotP,newStatP:double;
                       newVel: zNumber); external;
procedure saveProbeList; external;

procedure getGlobalRhsMemory; external;
procedure getGeoMemory(theBody:integer); external;
procedure getXYptrList(ThisDvStart : dvPtr; var dvOnList: integer); external;
procedure getCellXYlist(chainStart : PChain; var NumDVinCell: integer); external;

procedure createVelRec; external;
procedure freeMemoryAtEnd; external;
procedure freeBodyAryMemory; external;
procedure freeChainOfPoint(var chainStart:PlinkPt); external;
procedure addPointToChain(var chainStart:PlinkPt;newPoint:zNumber); external;

{/// RomBios procedure ///}
{Procedure GetSysDate (Var Year,month,Day:integer); external;}
{Procedure GetSysTime (Var Hours, Minutes, Seconds, Hundredths : integer); external;}
{/// string handling functions ///}
procedure OurStr(inInt:integer;var aStr:TString); external;
function copyStr(masterStr:TString;index,count:integer):TString; external;
function PosStr(subStr,masterStr:TString):integer; external;
function StrLength(aString:TString):integer; external;
procedure Value(aString:TString; var outValue:double; var errCode:integer); external;
procedure ValInt(aString:TString; var outValue:Integer; var errCode:integer); external;
function UpperCase(aChar:Char):Char; external;
{/// random number generators ///}

function  KnuthRan:double; external;
function  Knuth2B:double; external;
function  PortRandom:double; external;

procedure setAlarm(aYear,aMonth,aDay,aHour,aMin:integer); external;
procedure ClockOn; external;
procedure ClockOff; external;
function  timesUp:Boolean; external;
function  ValidDate(aString:TString;var dayW,monthW,yearW:integer):boolean; external;
function  ValidTime(aString:TString; var hourW,minuteW:integer):boolean; external;

function  getSegmentTag(bdy,inputTag: integer):integer; external;
function  getCofSlope(bdy,inputElem: integer):double; external;
function  getSlope2Seg(bdy,Elem1, Elem2:integer):double; external;
procedure endPoints (bdy,i:integer; var xy1,xy2: zNumber); external;
function  getDist(xyi:zNumber;var bdy,elem:integer):double; external;
function  getGap2Seg(xyi:zNumber;var bdy,elem:integer;var theGap:double):double; external;
procedure getElemD(inpt:zNumber;bdy:integer;var which:integer;var distance:double); external;
procedure getElemR(inpt:zNumber;bdy:integer;var which:integer;var rdist:double); external;
function  getAngleR(inpt1,inpt2:zNumber):double; external;
function  get3ptAngle(ptA,ptB,ptC:zNumber):double; external;
function  crossLine(lineA,lineB,ptA,ptB:zNumber):integer; external;
function  getCrossOver(var bdy,suspect:integer;theDv:dvPtr):boolean; external;

procedure ZSum(z1,z2: zNumber;var z3: zNumber); external;
procedure ZProd(z1,z2: zNumber;var z3: zNumber); external;
procedure ZDiv(z1,z2: zNumber; var z3: zNumber); external;
procedure zInv(z1: zNumber; var z2: zNumber); external;
function  factorial(seed: integer):double; external;

procedure splitString(var inString, outString: TString; var numPart:integer); external;

function DeNumberNACA4D(inValue:double;var camber,pValue,thick:double):boolean; external;

procedure swapPtr(var A, B: dvPtr); external;
procedure swapInt(var A, B: integer); external;
procedure swapReal(var A, B: double); external;
procedure swapBool(var A, B: Boolean); external;
procedure xQSort (Var Values: ptrList; Start, EndList: integer); external;
procedure yQSort (Var Values: ptrList; Start, EndList: Integer); external;
function  minTagOfList(theVector:vectorM;endList:integer):integer; external;

function  sameSign(realA,realB:double):boolean; external;
function  facingAngle(angleA,angleB:double):boolean; external;
procedure multiPotAngle( inpt1,inpt2:zNumber;var angQ : quartet); external;
function  match4(numAry: quartet): double; external;

procedure getCtocStatistic; external;

function  ValidDv(inPutDv: dvPtr): boolean; external;
{function  FileExist(FileName:string):boolean; external;}
function  bodyLapped:boolean; external;
function  checkPointInBody(theChain:PlinkPt):boolean; external;
function IsPtInBody(thePoint:zNumber;doCheck:boolean):boolean; external;
function  PtInOneBody( thePoint:zNumber; abody:integer;
                       doCheck:boolean):boolean; external;
procedure publishCircStatus; external;
procedure InitDvVelocity; external;
function  chkCellinBody(aCell:PCell):integer; external;
procedure SearchHiLoDV(starter:dvPtr; var LLo,HHi: zNumber); external;

Procedure xyuvSum(Point:zNumber;var inVel:zNumber); external;
procedure getOneInduced(inDvPtr: dvPtr); external;
function  VortDecay(inDv:dvPtr): double; external;
procedure getFluctuateFS; external;

procedure InitData;
var bdy: integer;
begin
  snuff:=0.0;           { initial strength of snuffed vorticity :=0.0   }
  elapsed:=0.0;         { set time zero                                 }
  ndv:=0;               { initial number of free vortex :=0             }
  nTrash:=0;
  nPCTrash:=0;
  nPCellTrash:=0;
  nPzCoordTrash:=0;

  fceListLength:=0;
  dvStart:=nil;
  dvEnd:=nil;
  smpStart:=nil;
  smpEnd:=nil;
  allGone:=true;

  Trash:=nil;
  ChainTrash:=nil;
  PCellTrash:=nil;
  PzCoordTrash:=nil;

  {bfBdyActive:=true;}
  {offsetActive:=true;}
  {snuffActive:=false;}
  vortexMode:=0;          { default vortexMode = 0, i.e. direct summation }
  strmStartPts:=nil;
  probePoint:=nil;
  dynaMatx:=nil;
  globalRhs:=nil;
  fcListStart:=nil;
  fcListEnd:=nil;
  probeStart:=nil;
  probeEnd:=nil;
  xOrder:=nil;
  yOrder:=nil;
  for bdy:=1 to maxBody do bodyAry[bdy]:=nil;
  fileGrp:='';
  isave:=0;
  Publish:=false;
  ReUseMem:=false;
  doProbing:=false;
  fluctFlow:=false;
  firstAccess:=true;
  calculateMode  :=0;
  vortexFlowMode := 0;
  DataError      := noError;
  vortexMode     := directSum;
  basicMeshFactor:=25.0;
  xIteration     :=2;
  DomainX        :=25;
  DomainY        :=6;
  decayFactor    :=-0.03;
  numSubLevel    :=0;
  FluctLevel:=0.0;
  probeOn:=0.0;
  probeOff:=0.0;
  strmTime:=0.025;
  newDvMem:=0;
  newLinkMem:=0;
  newCellMem:=0;
  newChainMem:=0;
  newzCoordMem:=0;
  newDataBankMem:=0;
  newCGroupMem:=0;
  newzMmtMem:=0;
  newCtocMem:=0;
  chaindump:=0;
  chainReUse:=0;
end;

to begin do
  initData;
End.




