[inherit('vglobal')]
program sample(input,output,logbook);

const
  validChar=['0'..'9','a'..'z','A'..'Z','.','/','\',':','+','-'];
  limiter=[',','(',')',';','='];
  maxFLine=10;

type
  FlowLine = record
   FirstLink : plinkPt;
   numUnit   : integer;
   netLength,NetFlow : double;
   Lopt,Uppt,meanVel,
   uniNorm           : zNumber;
  end;

var
  logbook                     : text;
  Heading,CodeName,curString  : Tstring;
  NumFlowLine,curInteger,
  IntError,RealError          : integer;
  stepCorr                    : integer;
  hitEoln                     : boolean;
  curReal,tmpDecay            : double;
  FlowBook                    : array [1..maxFLine] of Flowline;
  CurNumBdy                                 : integer;
  CurBody                                   : bodyList;
  curMatx                                   : PCoeffRow;

procedure initCurBody;
var i : integer;
begin
  CurNumBdy:=0;
  for i:=1 to maxBody do with CurBody[i] do begin
    ShapeCode:=0;
    BodyCX:=0.0;
    BodyCY:=0.0;
    BodyTilt:=0.0;
    xDim:=0.0;
    yDim:=0.0;
    zDim:=0.0;
    Nx:=0;
    Ny:=0;
    Nnose:=0;
  end;
end;

procedure updateBody;
var i: integer;
begin
  CurNumBdy:=numActiveBbody;
  for i:=1 to numActiveBbody do begin
   CurBody[i].ShapeCode:=ActiveBbody[i].ShapeCode;
   CurBody[i].BodyCX:=ActiveBbody[i].BodyCX;
   CurBody[i].BodyCY:=ActiveBbody[i].BodyCY;
   CurBody[i].BodyTilt:=ActiveBbody[i].BodyTilt;
   CurBody[i].xDim:=ActiveBbody[i].xDim;
   CurBody[i].yDim:=ActiveBbody[i].yDim;
   CurBody[i].zDim:=ActiveBbody[i].zDim;
   CurBody[i].Nx:=ActiveBbody[i].Nx;
   CurBody[i].Ny:=ActiveBbody[i].Ny;
   CurBody[i].Nnose:=ActiveBbody[i].Nnose;
  end;
end;

function matchBody:boolean;
label 99;
var i,j: integer;
begin
  matchBody:=false;
  if CurNumBdy=numActiveBbody then begin
    i:=1;
    repeat
      j:=i;
      if CurBody[i].ShapeCode=ActiveBbody[i].ShapeCode then
       if CurBody[i].BodyCX=ActiveBbody[i].BodyCX then
        if CurBody[i].BodyCY=ActiveBbody[i].BodyCY then
         if CurBody[i].BodyTilt=ActiveBbody[i].BodyTilt then
          if CurBody[i].xDim=ActiveBbody[i].xDim then
           if CurBody[i].yDim=ActiveBbody[i].yDim then
            if CurBody[i].zDim=ActiveBbody[i].zDim then
             if CurBody[i].Nx=ActiveBbody[i].Nx then
              if CurBody[i].Ny=ActiveBbody[i].Ny then
               if CurBody[i].Nnose=ActiveBbody[i].Nnose then
                if i=numActiveBbody then matchBody:=true
                else i:=i+1;
      if j=i then goto 99;
    until i>numActiveBbody;
  end;

99: updateBody;
end;


procedure SmpResetValue;
begin
  firstAccess:=true;
  snuff:=0.0;          {initial strength of snuffed vorticity :=0.0   }
  elapsed:=0.0;        {set time zero                                 }
  ndv:=0;              {initial number of free vortex :=0             }
  allGone:=true;
  initGridData;
  vortexMode:=directSum;
end;

procedure getSmpData( theStart,theEnd:PSmpCell;var col,row:double;
                      var Loxy,Upxy: zNumber;
                      var Celsize : double);
var
  curRow: PSmpCell;
begin
  Col:=0.0;
  Loxy:=theStart^.ctr;
  Upxy:=theEnd^.ctr;
  curRow:=theStart;
  while curRow<>nil do begin
    Col:=Col+1.0;
    curRow:=curRow^.East;
  end;
  Row:=0.0;
  curRow:=theStart;
  while curRow<>nil do begin
    Row:=Row+1.0;
    curRow:=curRow^.Nth;
  end;
  celSize:=abs((Upxy.Re-Loxy.Re))/(col-1);
end;

function totalGrid( theStart:PSmpCell):double;
var
  col,row: double;
  curRow: PSmpCell;
begin
  Col:=0.0;
  curRow:=theStart;
  while curRow<>nil do begin
    Col:=Col+1.0;
    curRow:=curRow^.East;
  end;
  Row:=0.0;
  curRow:=theStart;
  while curRow<>nil do begin
    Row:=Row+1.0;
    curRow:=curRow^.Nth;
  end;
  totalGrid:=col*row;
end;

procedure saveSampleD(theStart,theEnd:PSmpCell;bin,avgOf:integer;filename:Tstring);
var
  Loxy,upxy                 : zNumber;
  celsize,col,row,divid,dmp : double;
  writeFile                 : file of double;
  curNode,curRow            : PSmpCell;
begin
  rewrite(writeFile,fileName);
  if avgOf>0 then divid:=1/avgOf else divid:=1.0;
  getSmpData(theStart,theEnd,col,row,Loxy,Upxy,celsize);
  write(writeFile,col,row,Loxy.Re,Loxy.Im,Upxy.Re,Upxy.Im,celSize);
  curRow:=theStart;
  while curRow<>nil do begin
    curNode:=CurRow;
    while curNode<>nil do with curNode^ do begin
      dmp:=sample[bin]*divid;
      write(writeFile,dmp);
      curNode:=curNode^.East;
    end;
    curRow:=curRow^.Nth;
  end;
  close(writeFile);
end;

procedure SaveFlowBook(filename : Tstring);
var
  i         : integer;
  tmpPt     : PlinkPt;
  writeFile : file of double;
begin
  rewrite(writeFile,fileName);
  for i:=1 to maxFLine do with FlowBook[i] do begin
   if FirstLink<>nil then begin
    write(writefile,Lopt.Re,Lopt.Im,Uppt.Re,Uppt.Im,netLength);
    write(writefile,netFlow);
   end;
  end;
  close(writeFile);
  writeln(' file save ', filename);
end;

procedure ReadHeading;
var
  OneChar: Char;
  TmpStg : Tstring;
begin
  TmpStg:='';
  repeat
   read(logbook,OneChar);
   if OneChar in ValidChar then
   TmpStg:=TmpStg+UpperCase(OneChar);
  until OneChar in Limiter;

  Heading:=TmpStg;
end;

Procedure readItem;
var
  OneChar : Char;
  tmpStg  : Tstring;
begin
  TmpStg:='';
  repeat
   read(logbook,OneChar);
   if OneChar in validChar then TmpStg:=TmpStg+UpperCase(OneChar);
  until OneChar in Limiter;
  hitEoln:=OneChar=')';
  CodeName:=TmpStg;

  CurString:='';
  repeat
   read(logbook,OneChar);
   if OneChar in ValidChar then CurString:=CurString+UpperCase(OneChar);
  until (OneChar in Limiter);
  hitEoln:=OneChar=')';
  if CurString<>'' then Value(curString,curReal,RealError);
end;

function retrieveFile(readN:string):boolean;
var
  reTime : timeRec;
begin
  retrieveFile:=false;
  if readN <>'' then begin
   ReadDvforRestart(readN,dvStart,dvEnd,ndv,reTime);
   dt:=reTime.tStep;
   elapsed:=reTime.tLast;
   if ndv>0 then retrieveFile:=true;
  end;
end;

procedure ReadData( var readN,saveN:Tstring;
                    var LoSmp,UpSmp:zNumber;var celength: double);
begin
  LoSmp.Re:=0.0;
  LoSmp.Im:=0.0;
  UpSmp.Re:=0.0;
  UpSmp.Im:=0.0;
  celength:=0.0;
  readN:='';
  saveN:='';

  Repeat
  ReadItem;

  if CodeName='FILE' then begin
    if CurString<>'' then readN:=CurString;
  end else
  if CodeName='SAVE' then begin
    if CurString<>'' then saveN:=CurString;
  end else
  if CodeName='LX' then begin
    if RealError=0 then LoSmp.Re:=CurReal;
   end else
  if CodeName='LY' then begin
    if RealError=0 then LoSmp.Im:=CurReal;
   end else
  if CodeName='HX' then begin
    if RealError=0 then UpSmp.Re:=CurReal;
   end else
  if CodeName='HY' then begin
    if RealError=0 then UpSmp.Im:=CurReal;
   end else
  if CodeName='DXY' then begin
    if RealError=0 then celength:=abs(CurReal);
   end else
  if CodeName='DECAY' then begin
    if RealError=0 then tmpDecay:=-abs(CurReal);
  end;
  until hitEoln;
end;

function prepareSmpCell(LoSmp,UpSmp : zNumber;celength: double):boolean;
begin
  prepareSmpCell:=false;
  if ((LoSmp.Re<>UpSmp.Re) and
     (LoSmp.Im<>UpSmp.Im)) and (celength>0.0) then begin
   formSmpCell(smpStart,smpEnd,LoSmp,UpSmp,Celength);
   prepareSmpCell:=true;
  end;
end;

function correcttime:boolean;
begin
  if ReyNum<150.0 then begin
    dt:=dt*4.0;
    stepCorr:=2;
  end
   else stepCorr:=1;
  correcttime:=dt>0.0;
end;

function formOneLine(var numLine:integer;LoSmp,UpSmp:zNumber;npoint:double):boolean;
var
  lineSlope : double;
  tmpPt,zProp : zNumber;
  tmpCounter,i,npnt : integer;
begin
  formOneLine:=false;
  tmpCounter:=numLine+1;
  if tmpCounter<=maxFLine then begin
   if ((LoSmp.Re<>UpSmp.Re) or (LoSmp.Im<>UpSmp.Im)) and
     (round(npoint)>0) then begin

     with FlowBook[tmpCounter] do begin

      netLength:=sqrt(sqr(LoSmp.Re-UpSmp.Re)+sqr(LoSmp.Im-UpSmp.Im));
      numUnit:=round(nPoint);
      Lopt:=LoSmp;
      Uppt:=UpSmp;
      lineSlope:=getAngleR(LoSmp,UpSmp);
      uniNorm.Re:=-sin(lineSlope);
      uniNorm.Im:=cos(lineSlope);
      if numUnit>=1 then begin
       zProp.Re:=(UpSmp.Re-LoSmp.Re)/numUnit;
       zProp.Im:=(UpSmp.Im-LoSmp.Im)/numUnit;
       for i:=1 to numUnit do begin
        tmpPt.Re:=LoSmp.Re+zProp.Re*(i-0.5);
        tmpPt.Im:=LoSmp.Im+zProp.Im*(i-0.5);
        addPointToChain(firstLink,tmpPt);
       end;
      end;
     end; {with}
     formOneLine:=true;
     numLine:=tmpCounter;
   end;{if}
  end;
end;

procedure sampleVort(whichSmp:PSmpCell;typside:double;var numDone : integer);
begin
  SmearDVonSmp(whichSmp,typside,vobin);
  numDone:=numDone+1;
end;

procedure samplePressure(whichSmp:PSmpCell;var numDone:integer);
var
  tmpPo,tmpPs    : double;
  tmpVel         : zNumber;
  curNode,curRow : PSmpCell;
begin
  curRow:=whichSmp;
  while curRow<>nil do begin
    curNode:=CurRow;
    while (curNode<>nil) do with curNode^ do begin
      {/// survey only the inner-most cell ///}
      if curNode^.freept then QuickPitot(ctr,tmpPo,tmpPs,tmpVel)
      else begin
       tmpPo:=0.0;
       tmpPs:=0.0;
      end;
      sample[pobin]:=sample[pobin]+tmpPo*100;
      sample[psbin]:=sample[psbin]+tmpPs*100;
      curNode:=curNode^.East;
    end;
    curRow:=curRow^.Nth;
  end;
  numDone:=numDone+1;
end;

procedure sampleStreamFn(whichSmp:PSmpCell;var numDone:integer);
var
  insideBdy    : integer;
  tmpStream    : double;
  tmpVel         : zNumber;
  curNode,curRow : PSmpCell;
begin
  curRow:=whichSmp;
  while curRow<>nil do begin
    curNode:=CurRow;
    while curNode<>nil do with curNode^ do begin
      {/// survey only the inner-most cell ///}
      xystm(ctr,tmpStream,insideBdy,noCheck);
      sample[vobin]:=sample[vobin]+tmpStream;
      curNode:=curNode^.East;
    end;
    curRow:=curRow^.Nth;
  end;
  numDone:=numDone+1;
end;

procedure genStreamLine(readN,saveN: Tstring;LoSmp,UpSmp : zNumber; celength: double);
var
  streamDone : integer;
begin
  SmpResetValue;
  if retrieveFile(readN) then
   if prepareData then
    if prepareSmpCell(LoSmp,UpSmp,celength) then begin
     if matchBody then begin
      firstAccess:=false;
      copyMatx(curMatx,dynaMatx);
     end;

     writeln;writeln(' generate streamlines... from file : ',readN);
     writeln(' domain : ','(',LoSmp.Re:6:3,',',LoSmp.Im:6:3,')',
                          '(',UpSmp.Re:6:3,',',UpSmp.Im:6:3,')',
                          'x ',celength:6:3);
     decayFactor:=tmpDecay;
     writeln(' ctc level  ....... : ',numSubLevel);
     writeln(' decay factor... : ',decayFactor);
     viscousFlow;
     streamDone:=0;
     {numTimeStep:=2;
     istep:=0;
     while (istep<numTimeStep) do begin
       if (vortexMode=directSum) and (ndv>modeChangeOver) then vortexMode:=cellToCell;
       xconvect;
       randomwalk;
       elapsed:=elapsed+dt;
       istep:=istep+1;
       createVortexSheet;
       if istep>1 then sampleStreamFn(smpStart,streamDone);
       shedDv;
     end;}
     sampleStreamFn(smpStart,streamDone);
     saveSampleD(smpStart,smpEnd,vobin,streamDone,saveN);
     writeln(' file saved : ',saveN);
     copyMatx(dynaMatx,curMatx);
    end;

  freeMemoryAtEnd;
end;

procedure genPressure(readN,saveN: Tstring;LoSmp,UpSmp : zNumber; celength: double);
var
  aint,pressDone : integer;
begin
  SmpResetValue;
  if retrieveFile(readN) then
   if correcttime then
    if prepareData then
     if prepareSmpCell(LoSmp,UpSmp,celength) then begin
      checkSmpCell(smpStart);
      basicMeshFactor:=basicMeshFactor*1.5;
      decayFactor:=tmpDecay;
      writeln;
      writeln(' generate pressure field ... from file : ',readN);
      writeln(' domain : ','(',LoSmp.Re:6:3,',',LoSmp.Im:6:3,')',
                           '(',UpSmp.Re:6:3,',',UpSmp.Im:6:3,')',
                           'x ',celength:6:3);
      writeln(' decay factor... : ',decayFactor);
      writeln(' timestep    ... : ',dt);
      writeln(' MeshFactor  ... : ',basicMeshFactor);
      writeln(' main stream ... : ',uxCurt:6:3,' ',vyCurt:6:3);
      writeln(' number ndv  ... : ',ndv:9:6);

      pressDone:=0;
      numTimeStep:=4;
      istep:=0;
      viscousFlow;


      while (iStep < numTimeStep) do begin
       if (vortexMode=directSum) and (ndv>modeChangeOver) then
        vortexMode:=cellToCell;

       xConvect;
       RandomWalk;

       elapsed:=elapsed+dt;
       istep:=istep+1;

       createVortexSheet;
       if istep>1 then begin
        getGeneralVel;
        samplePressure(smpStart,pressDone);
       end;
      shedDv;
     end;
     saveSampleD(smpStart,smpEnd,pobin,pressDone,saveN+'.PO');
     saveSampleD(smpStart,smpEnd,psbin,pressDone,saveN+'.PS');
     writeln(' file saved : ',saveN);
    end;
  freeMemoryAtEnd;
end;

procedure readUserinput(var readN,saveN : Tstring;
                        var LoSmp,UpSmp : zNumber;var celength: double);
begin
  write('  Sampling file name    : ');
  readln(readN);
  write('  output file name      : ');
  readln(saveN);

  write('  Lower left domain,  x : ');
  readln(LoSmp.Re);
  write('  Lower left domain,  y : ');
  readln(LoSmp.Im);
  write('  Upper right domain, x : ');
  readln(UpSmp.Re);
  write('  Upper right domain, y : ');
  readln(UpSmp.Im);
  write('  grid size  : ');
  readln(celength);
  write(' decay factor : ');
  readln(decayFactor);
  decayFactor:=-abs(decayFactor);
end;

procedure goStrmLine;
var
  LoSmp,UpSmp : zNumber;
  celength: double;
  ReadName,saveName: Tstring;
begin
  readData(ReadName,saveName,LoSmp,UpSmp,celength);
  genStreamLine(ReadName,saveName,LoSmp,UpSmp,celength);
end;

procedure goPressure;
var
  LoSmp,UpSmp : zNumber;
  celength: double;
  ReadName,saveName: Tstring;
begin
  readData(ReadName,saveName,LoSmp,UpSmp,celength);
  genPressure(ReadName,saveName,LoSmp,UpSmp,celength);
end;

procedure goFlowRate;
var
  LoSmp,UpSmp,tmpVel : zNumber;
  celength           : double;
  ReadName,saveName  : Tstring;
  i,insideBdy        : integer;
  tmpPt              : PlinkPt;
begin
  smpResetValue;
  readData(ReadName,saveName,LoSmp,UpSmp,celength);
  if retrieveFile(readName) then
    if prepareData then begin
     writeln;
     writeln(' find flowrate... from file : ',readName);
     viscousFlow;
     for i:=1 to maxFLine do with FlowBook[i] do begin
      if FirstLink<>nil then begin
       netFlow:=0.0;
       meanVel.Re:=0.0;
       meanVel.Im:=0.0;

       tmpPt:=FirstLink;
       while tmpPt<>nil do begin
        xyuvQuick(tmpPt^.ptCoord,tmpVel,insideBdy,noCheck);
        meanVel.Re:=meanVel.Re+tmpVel.Re;
        meanVel.Im:=meanVel.Im+tmpVel.Im;
        tmpPt:=tmpPt^.downLink;
       end;
       netFlow:=netLength*(meanVel.Re*uniNorm.Re +
                           meanVel.Im*uniNorm.Im)/numUnit;
      end;
     end;
     { save result to file }
     saveFlowBook(saveName);
    end;
  freeMemoryAtEnd;
end;

procedure initFlowBook;
var
  i : integer;
begin
  for i:=1 to maxFLine do with FlowBook[i] do begin
    firstLink:=nil;
    Lopt.Re:=0.0;
    Lopt.Im:=0.0;
    Uppt.Re:=0.0;
    Uppt.Im:=0.0;
    uniNorm.Re:=0.0;
    uniNorm.Im:=0.0;
    meanVel.Re:=0.0;
    meanVel.Im:=0.0;
    NetFlow:=0.0;
  end;
  numFlowLine:=0;
end;

procedure formFlowNet;
var
  LoSmp,UpSmp : zNumber;
  celength    : double;
  ReadName,saveName : Tstring;
begin
  readData(ReadName,saveName,LoSmp,UpSmp,celength);
  formOneLine(numFlowLine,LoSmp,UpSmp,celength);
end;

procedure freeFlowBook;
var i: integer;
    tmpPt : PlinkPt;
begin
  for i:=1 to maxFLine do
   if FlowBook[i].FirstLink<>nil then freeChainOfPoint(FlowBook[i].FirstLink);
end;

begin
  reset(logbook,'SAMPLE.LOG');
  numSubLevel:=2;
  InitGridData;
  InitFlowBook;
  initCurBody;
  curMatx:=nil;

  Repeat

    ReadHeading;

    if Heading='STRM' then goStrmLine else
    if Heading='PRES' then goPressure else
    if Heading='FNET' then formFlowNet else
    if Heading='FLOW' then goFlowRate;

  until Heading='END';

  close(logbook);
  freeFlowBook;
  freeMatrixMemory(curMatx);

end.
